
#pragma once

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <vector>
#include <list>
#include <map>
#include <string>
#include <boost/any.hpp>

#include <boost/shared_ptr.hpp>

//Compile Time configuration
//#define ENABLE_PARALLEL_SELECTION
#define ENABLE_GPU_ACCELERATION
#define ENABLE_TWO_PHASE_PHYSICAL_OPTIMIZATION
//#define PRINT_QUERY_STARTINGTIME
//#define FORCE_USE_UNSTABLE_SORT_FOR_BENCHMARK

//enables use of CDK primitives
#define ENABLE_CDK_USAGE

#define ENABLE_GPU_ACCELERATED_VARCHAR_SCAN
//#define ENABLE_GPU_JOIN 
#define ENABLE_GPU_PK_FK_JOIN  
#ifndef HYPE_ENABLE_PARALLEL_QUERY_PLAN_EVALUATION
        #define HYPE_ENABLE_PARALLEL_QUERY_PLAN_EVALUATION
#endif

#define COGADB_WARNING(X, Y)  std::cout << "WARNING: " << X << ": In " <<  __PRETTY_FUNCTION__ << std::endl;\
std::cout << "In File: " << __FILE__ << " Line: " << __LINE__ << std::endl;

#define COGADB_ERROR(X, Y)  std::cout << "ERROR: " << X << ": In " << __PRETTY_FUNCTION__ << std::endl;\
std::cout << "In File: " << __FILE__ << " Line: " << __LINE__ << std::endl;

#define COGADB_FATAL_ERROR(X, Y)  std::cout << "FATAL ERROR: " << X << ": In " << __PRETTY_FUNCTION__ << std::endl;\
std::cout << "In File: " << __FILE__ << " Line: " << __LINE__ << std::endl;\
std::cout << "Aborting Execution..." << std::endl;\
exit(-1);

#define COGADB_INSTANTIATE_CLASS_TEMPLATE_FOR_SUPPORTED_TYPES(TEMPLATE) template class TEMPLATE<int>; \
        template class TEMPLATE<float>; \
        template class TEMPLATE<std::string>;

namespace shared_pointer_namespace = boost; //std::tr1

namespace CoGaDB{

enum AttributeType{INT,FLOAT,VARCHAR,BOOLEAN};

enum ComputeDevice{CPU,GPU};

enum AggregationMethod{SUM,MIN,MAX};

enum ValueComparator{LESSER,GREATER,EQUAL,LESSER_EQUAL,GREATER_EQUAL};

enum PredicateType {ValueValuePredicate, ValueConstantPredicate};

enum SortOrder{ASCENDING,DESCENDING};

enum Operation{SELECTION,PROJECTION,JOIN,GROUPBY,SORT,COPY,AGGREGATION,FULL_SCAN,INDEX_SCAN};

enum JoinAlgorithm{SORT_MERGE_JOIN,SORT_MERGE_JOIN_2,NESTED_LOOP_JOIN,HASH_JOIN, PARALLEL_HASH_JOIN, RADIX_JOIN};

enum MaterializationStatus{MATERIALIZE,LOOKUP};

enum ParallelizationMode{SERIAL,PARALLEL};

enum ColumnAlgebraOperation{ADD,SUB,MUL,DIV};

enum PositionListOperation{POSITIONLIST_INTERSECTION,POSITIONLIST_UNION};

enum BitmapOperation{BITMAP_AND,BITMAP_OR};

enum DebugMode{quiet=1,
					verbose=0,
					debug=0,
					print_time_measurement=0};

/*
enum DebugMode{quiet=0,
					verbose=1,
					debug=0,
					print_time_measurement=0};
*/
typedef unsigned int TID;

typedef std::pair<TID,TID> TID_Pair;

typedef std::pair<AttributeType,std::string> Attribut;

typedef std::list<Attribut> TableSchema;

typedef std::vector<boost::any> Tuple;

typedef std::pair<std::string,AggregationMethod> ColumnAggregation;

//for Rename Operator
typedef std::pair<std::string,std::string> RenameEntry;
typedef std::list<RenameEntry> RenameList;

//struct Attribut {

//	AttributeType type_;
//	std::string name_;
//	ColumnPtr column_;

//	AttributeType& first;
//	std::string& second;

//}
}; //end namespace CogaDB

