
#pragma once

#include <core/global_definitions.hpp>
#include <core/base_column.hpp>

namespace CoGaDB{

#ifdef ENABLE_SIMD_ACCELERATION
    void simd_selection_int_thread(int* array, unsigned int array_size,  int comparison_value, const ValueComparator comp, unsigned int* result_array, unsigned int* result_array_size, unsigned int result_tid_offset);
    
    void simd_selection_float_thread(float* array, unsigned int array_size, float comparison_value, const ValueComparator comp, unsigned int* result_array, unsigned int* result_array_size, unsigned int result_tid_offset);
    /* \brief  scans an integer column with SIMD isntructions
     * \details result_tid_offset is a fixed value which is added to all matching 
     *          tid values, allowing to use the SIMD Scan in Multithreaded algorithms
    */
    const PositionListPtr simd_selection_int(int* array, unsigned int array_size,  int comparison_value, const ValueComparator comp, unsigned int result_tid_offset=0);
    /* \brief  scans a float column with SIMD isntructions
     * \details result_tid_offset is a fixed value which is added to all matching 
     *          tid values, allowing to use the SIMD Scan in Multithreaded algorithms
    */
    const PositionListPtr simd_selection_float(float* array, unsigned int array_size,  float comparison_value, const ValueComparator comp, unsigned int result_tid_offset=0);
#endif
    
    namespace CDK{
        namespace selection{
            namespace variants{
                void simd_selection_int_thread(int* array, size_t column_size, unsigned int thread_id, unsigned int number_of_threads, const int value, const ValueComparator comp, unsigned int* result_tids, unsigned int* result_size);
            };
        };
        
    };
    
}; //end namespace CoGaDB

