#pragma once

#include <string>
#include <fstream>
#include <queue>

#include <core/table.hpp>

#include<boost/tokenizer.hpp>

namespace CoGaDB {

    class CommandLineInterpreter {
    public:
        CommandLineInterpreter();

        bool parallelExecution(const std::string& input, std::ifstream *in_stream);
        void threadedExecution();
        bool execute(const std::string& command);
        bool getline(const std::string &prompt, std::string &result);

    private:
        typedef bool (*SimpleCommandHandlerPtr)();
        typedef bool (*ParameterizedCommandHandlerPtr)(const std::string&);
        typedef std::map<std::string, SimpleCommandHandlerPtr> SimpleCommandMap;
        typedef std::map<std::string, ParameterizedCommandHandlerPtr> ParameterizedCommandMap;
        SimpleCommandMap simple_command_map_;
        ParameterizedCommandMap command_map_;
        std::queue<std::string> commandQueue;
        std::string prompt_;
    };




} //end namespace CogaDB


