#pragma once

#include <query_processing/definitions.hpp>
#include <query_processing/column_processing/definitions.hpp>
#include <util/getname.hpp>
#include <core/lookup_array.hpp>
#include <core/bitmap.hpp>
#include <query_processing/operator_extensions.hpp>

#include <hardware_optimizations/primitives.hpp>

namespace CoGaDB {
    namespace query_processing {
        
        namespace physical_operator {

            class CPU_Bitmap_Operator : public hype::queryprocessing::BinaryOperator<ColumnPtr, ColumnPtr, ColumnPtr>, public BitmapOperator {
            public:
                //typedef hype::queryprocessing::OperatorMapper_Helper_Template<ColumnPtr>::TypedOperatorPtr ColumnWise_TypedOperatorPtr;
                typedef column_processing::cpu::TypedOperatorPtr TypedOperatorPtr;

                CPU_Bitmap_Operator(const hype::SchedulingDecision& sched_dec,
                        TypedOperatorPtr left_child,
                        TypedOperatorPtr right_child,
                        BitmapOperation op,
                        MaterializationStatus mat_stat = MATERIALIZE);

                virtual bool execute();

                virtual ~CPU_Bitmap_Operator();
            private:
                BitmapOperation op_;
                MaterializationStatus mat_stat_;
            };

            class GPU_Bitmap_Operator : public hype::queryprocessing::BinaryOperator<ColumnPtr, ColumnPtr, ColumnPtr>, public BitmapOperator {
            public:
                //typedef hype::queryprocessing::OperatorMapper_Helper_Template<ColumnPtr>::TypedOperatorPtr ColumnWise_TypedOperatorPtr;
                typedef column_processing::cpu::TypedOperatorPtr TypedOperatorPtr;

                GPU_Bitmap_Operator(const hype::SchedulingDecision& sched_dec,
                        TypedOperatorPtr left_child,
                        TypedOperatorPtr right_child,
                        BitmapOperation op,
                        MaterializationStatus mat_stat = MATERIALIZE);

                virtual bool execute();

                virtual ~GPU_Bitmap_Operator();
            private:
                BitmapOperation op_;
                MaterializationStatus mat_stat_;
            };
            
            column_processing::cpu::Physical_Operator_Map_Ptr map_init_function_cpu_bitmap_operator();
            column_processing::cpu::TypedOperatorPtr create_CPU_Bitmap_Operator(column_processing::cpu::TypedLogicalNode& logical_node, const hype::SchedulingDecision&, column_processing::cpu::TypedOperatorPtr left_child, column_processing::cpu::TypedOperatorPtr right_child);
            column_processing::cpu::TypedOperatorPtr create_GPU_Bitmap_Operator(column_processing::cpu::TypedLogicalNode& logical_node, const hype::SchedulingDecision&, column_processing::cpu::TypedOperatorPtr left_child, column_processing::cpu::TypedOperatorPtr right_child);

            
            //			GPU_ColumnWise_TypedOperatorPtr create_GPU_ColumnAlgebraOperator(GPU_ColumnWise_TypedLogicalNode& logical_node, const hype::SchedulingDecision&, GPU_ColumnWise_TypedOperatorPtr left_child, GPU_ColumnWise_TypedOperatorPtr right_child);

        }//end namespace physical_operator

        namespace logical_operator {

            class Logical_Bitmap_Operator : public hype::queryprocessing::TypedNode_Impl<ColumnPtr, physical_operator::map_init_function_cpu_bitmap_operator> //init_function_Join_operator>
            {
            public:

                Logical_Bitmap_Operator(BitmapOperation op,
                        MaterializationStatus mat_stat = MATERIALIZE,
                        hype::DeviceConstraint dev_constr = hype::DeviceConstraint(hype::CPU_ONLY))
                : TypedNode_Impl<ColumnPtr, physical_operator::map_init_function_cpu_bitmap_operator>(false, dev_constr),
                op_(op),
                mat_stat_(mat_stat) {
                }

                virtual unsigned int getOutputResultSize() const {
                    return 0;
                }

                virtual double getCalculatedSelectivity() const {
                    return 1;
                }

                virtual std::string getOperationName() const {
                    return "Bitmap_Operator"; //util::getName(op_);
                }

                virtual BitmapOperation getBitmapOperation() const {
                    return op_;
                }

                const MaterializationStatus& getMaterializationStatus() const {
                    return mat_stat_;
                }
                std::string toString(bool verbose) const {
                    std::string result="Bitmap_Operator";
                    if(verbose){
                        result+=" (";
                        result+= util::getName(op_);
                        result+=")";
                    }
                    return result;
                }
            private:
                BitmapOperation op_;
                MaterializationStatus mat_stat_;
            };

        }//end namespace logical_operator

    }//end namespace query_processing

}; //end namespace CogaDB

