#ifndef SQL_DRIVER_HPP
#define SQL_DRIVER_HPP

#include "sql/server/sql_parsetree.hpp"
#include "sql_parser.hpp"

#include <iostream>
#include <string>
#include <exception>

namespace CoGaDB {
namespace SQL {

namespace Scanner {
	typedef void *buffer;

	Parser::token_type lex(Parser::semantic_type *yylval_param,
			       type yyscanner);
}

class Driver {
	Scanner::buffer buffer;
	/* FIXME: should not be public */
public:
	std::istream *istream;
private:
	Scanner::type scanner;

	Parser parser;

	ParseTree::SequencePtr result;

public:
	class ParseError : public std::exception {
	public:
		const char *
		what() const throw()
		{
			return "Error during parsing";
		}
	};

	Driver();
	~Driver();

	ParseTree::SequencePtr parse(std::istream &is);
	ParseTree::SequencePtr parse(const std::string &src);

private:
	/*
	 * defined in sql-scanner.lpp
	 */
	void init_scan();

	void set_input(std::istream &is);
	void set_input(const std::string &src);

	void destroy_scan();

	void error(const std::string& m);

	inline void
	unsupported(void)
	{
		error("Unsupported language construct");
	}

	/*
	 * Friends
	 */
	friend Parser::token_type Scanner::lex(Parser::semantic_type *,
					       Scanner::type);
	friend class Parser;
};

bool commandlineExec(const std::string &input);
bool commandlineExplain(const std::string &input);
//bool commandlineExplainStatements(const std::string &input);
bool commandlineExplainStatementsWithoutOptimization(const std::string &input);
bool commandlineExplainStatementsWithOptimization(const std::string &input);

} /* namespace SQL */
} /* namespace CoGaDB */

#endif
