#pragma once

#include <core/column.hpp>
#include <iostream>

#include <utility>
#include <functional>
#include <algorithm>

namespace CoGaDB{

template <class T>
struct min_obj : std::binary_function <T,T,T> {
	T operator() (const T& x, const T& y) const
	{
		return std::min(x,y);
	}
};

template <class T>
struct max_obj : std::binary_function <T,T,T> {
	T operator() (const T& x, const T& y) const
	{
		return std::max(x,y);
	}
};

template <typename T, typename U, typename BinaryOperator>
std::pair<ColumnPtr,ColumnPtr> reduce_by_keys(shared_pointer_namespace::shared_ptr<ColumnBaseTyped<U> > keys, shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > values, BinaryOperator binary_op){
 
 assert(keys->size()==values->size());

 Column<U>* new_keys = new Column<U>(keys->getName(),keys->getType());
 Column<T>* aggregated_values = new Column<T>(values->getName(),values->getType()); 

 //if there is nothign to do, return valid but empty colums
 if(keys->size()==0){
  return std::pair<ColumnPtr,ColumnPtr>(ColumnPtr(new_keys), ColumnPtr(aggregated_values));
 }

 T val=0;
 for(unsigned int i=0;i<values->size();i++){
 //unsigned int i=0;
 //while(i<values->size())
   
  if(i==0){
   val=(*values)[i];
   continue;
  }


 if(  (*keys)[i-1]==(*keys)[i] ){
	//std::cout << i <<  "SUM up values: " << val << "+" << (*values)[i] << "=" << binary_op(val,(*values)[i]) << std::endl;
   val = binary_op(val,(*values)[i]);  
 }else{
	//std::cout <<  "found new key: " << (*keys)[i]  << " old: " << (*keys)[i-1] << " computed value: " << val << std::endl;
	T tmp_key=(*keys)[i-1];
  	new_keys->insert(tmp_key);
  	aggregated_values->insert(val);
  	val=(*values)[i]; 
 }


 }
	//write result of last line
	T tmp_key=(*keys)[keys->size()-1];
  	new_keys->insert(tmp_key);
  	aggregated_values->insert(val);

 return std::pair<ColumnPtr,ColumnPtr>(ColumnPtr(new_keys), ColumnPtr(aggregated_values));
//return std::pair<ColumnPtr,ColumnPtr>(ColumnPtr(), ColumnPtr());
}

template <typename T, typename U>
std::pair<ColumnPtr,ColumnPtr> reduce_by_keys(shared_pointer_namespace::shared_ptr<ColumnBaseTyped<U> > keys, shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > values, AggregationMethod agg_meth){
    std::pair<ColumnPtr,ColumnPtr> result;
    if(agg_meth==SUM){
        result = reduce_by_keys(keys, values, std::plus<T>());
    }else if(agg_meth==MIN){
        result = reduce_by_keys(keys, values, min_obj<T>());
    }else if(agg_meth==MAX){
        result = reduce_by_keys(keys, values, max_obj<T>());
    }else{
      std::cerr << "FATAL Error! Unknown Aggregation Method!" << agg_meth << std::endl;

    }

    return result;
}
 



}; //end namespace CogaDB


