#include <thrust/version.h>

//thrust includes
#include <thrust/device_vector.h>
#include <thrust/copy.h>
#include <thrust/sort.h>
#include <thrust/partition.h>
#include <thrust/functional.h>

#include <gpu/gpu_algorithms.hpp>

namespace CoGaDB{

const std::vector<unsigned int> gpu_sort5(std::vector<unsigned int> column){

// initialize a device_vector with the list
  thrust::device_vector<unsigned int> D(column.begin(), column.end());

  thrust::sort(D.begin(),D.end());//A, A + N, thrust::greater<int>());

  // copy a device_vector into an STL vector
  std::vector<unsigned int> stl_vector(D.size());
  thrust::copy(D.begin(), D.end(), stl_vector.begin());

  return stl_vector;

}

const std::vector<unsigned int> gpu_sort_old(std::vector<unsigned int> values_in_column){
//template <typename T> 
//const std::vector<T> gpu_sort(std::vector<T> column){

// initialize a device_vector with the list
  thrust::device_vector<unsigned int> D(values_in_column.begin(), values_in_column.end());
  thrust::device_vector<unsigned int> device_tids;
  device_tids.resize(D.size());
  // set the elements of device_tids to 0, 1, 2, 3, ...
  thrust::sequence(device_tids.begin(), device_tids.end());

  //thrust::sort(D.begin(),D.end());//A, A + N, thrust::greater<int>());

  //note, that the keys we want to sort are the values in the column -> the order of the "values" is the desired order of tids that describe the the new order of tuples
thrust::stable_sort_by_key(D.begin(),D.end(), device_tids.begin());

  // copy a device_vector into an STL vector
  std::vector<unsigned int> tids(device_tids.size());
  thrust::copy(device_tids.begin(), device_tids.end(), tids.begin()); //copy back tids

  return tids;

}
/*
template <typename T> 
const std::vector<T> gpu_sort(std::vector<T> column);
*/

template <typename T> 
const std::vector<unsigned int> gpu_sort(std::vector<T> values_in_column){

// initialize a device_vector with the list
  thrust::device_vector<T> D(values_in_column.begin(), values_in_column.end());
  thrust::device_vector<unsigned int> device_tids;
  device_tids.resize(D.size());
  // set the elements of device_tids to 0, 1, 2, 3, ...
  thrust::sequence(device_tids.begin(), device_tids.end());

  //thrust::sort(D.begin(),D.end());//A, A + N, thrust::greater<int>());

  //note, that the keys we want to sort are the values in the column -> the order of the "values" is the desired order of tids that describe the the new order of tuples
   thrust::stable_sort_by_key(D.begin(),D.end(), device_tids.begin());

  // copy a device_vector into an STL vector
  std::vector<unsigned int> tids(device_tids.size());
  thrust::copy(device_tids.begin(), device_tids.end(), tids.begin()); //copy back tids

  return tids;


	//return std::vector<T>();
}

//empty specialization, because Thrust cannot handle bool and string
template <>
const std::vector<unsigned int> gpu_sort(std::vector<bool> column){
	std::cout << "Fatal error!!! Cannot sort bool columns on GPU!!!" << std::endl;
	std::exit(-1);
	return std::vector<unsigned int>();
}

template <>
const std::vector<unsigned int> gpu_sort(std::vector<std::string> column){
	std::cout << "Fatal error!!! Cannot sort string columns on GPU!!!" << std::endl;
	std::exit(-1);
	return std::vector<unsigned int>();
}



//__global__ void simple_join_kernel(int* join_column1, size_t size_join_column1, int* join_column2, size_t size_join_column2, int* join_column1_tids, int* join_column2_tids, size_t max_size_result) {


//assumption: only joins between primary key and foreign key columns
__global__ void nested_loop_join_kernel(int* primary_key_column, size_t size_primary_key_column, int* foreign_key_column, size_t size_foreign_key_column, int* primary_key_column_tids, int* foreign_key_column_tids, size_t max_size_result) {

int x = blockIdx.x;
int y = threadIdx.x;

if(primary_key_column[x]==foreign_key_column[y]){ 
	primary_key_column_tids[y]=x;   //add join pair
	foreign_key_column_tids[y]=y;
}

}

//int x = threadIdx.x + blockIdx.x * blockDim.x;
//int y = threadIdx.y + blockIdx.y * blockDim.y;


void HANDLE_ERROR(cudaError err){
	if(cudaSuccess==err){
		return;
	}else if(cudaErrorMemoryAllocation==err){
		std::cout << "cudaErrorMemoryAllocation! Exiting..." << std::endl;
	}
}

//only for primary key foreign key columns!
const std::vector<std::pair<unsigned int,unsigned int> > gpu_join_tids(std::vector<int>& primary_key_join_column, std::vector<int>& foreign_key_join_column){
	std::vector<std::pair<unsigned int,unsigned int> > result;

	//int* i,j,k,l;

	//simple_join_kernel<<<>>>();
//const int N = 10;

//int a[N], b[N], c[N];


int *dev_primary_key_column, *dev_foreign_key_column, *dev_primary_key_column_tids, *dev_foreign_key_column_tids;
// allocate the memory on the GPU
//HANDLE_ERROR( cudaMalloc( (void**)&dev_a, N * sizeof(int) ) );
//HANDLE_ERROR( cudaMalloc( (void**)&dev_b, N * sizeof(int) ) );
//HANDLE_ERROR( cudaMalloc( (void**)&dev_c, N * sizeof(int) ) );

//input arrays
cudaMalloc( (void**)&dev_primary_key_column, primary_key_join_column.size() * sizeof(int) );
cudaMalloc( (void**)&dev_foreign_key_column, foreign_key_join_column.size() * sizeof(int) );
//result arrays
cudaMalloc( (void**)&dev_primary_key_column_tids, foreign_key_join_column.size() * sizeof(int) );
cudaMalloc( (void**)&dev_foreign_key_column_tids, foreign_key_join_column.size() * sizeof(int) );

typedef std::vector<int>::iterator iterator;

iterator it=primary_key_join_column.begin();
iterator it2=foreign_key_join_column.begin();
int *primary_key_column_ptr = &(*it);
int *foreign_key_column_ptr = &(*it2);

	std::cout << "primary key column: " << std::endl;
for(int i=0;i<primary_key_join_column.size() ;i++){
	std::cout << primary_key_column_ptr[i] << std::endl;
}
	std::cout << "foreign key column: " << std::endl;
for(int i=0;i<foreign_key_join_column.size() ;i++){
	std::cout << foreign_key_column_ptr[i] << std::endl;
}

std::cout << "copy data" << std::endl;

cudaMemcpy( dev_primary_key_column, primary_key_column_ptr, primary_key_join_column.size() * sizeof(int) , cudaMemcpyHostToDevice );
cudaMemcpy( dev_foreign_key_column, foreign_key_column_ptr, foreign_key_join_column.size() * sizeof(int) , cudaMemcpyHostToDevice );

	dim3 grid(primary_key_join_column.size(),foreign_key_join_column.size());



//dim3 gridSize(gridWidth, gridHeight);
//dim3 blockSize(blockWidth, blockHeight);
//kernel<<<gridSize, blockSize>>>


//simple_join_kernel<<<grid,1>>>(dev_primary_key_column, primary_key_join_column.size(), dev_foreign_key_column, foreign_key_join_column.size(), 
//								 dev_primary_key_column_tids, dev_foreign_key_column_tids, foreign_key_join_column.size());


unsigned int number_of_blocks = primary_key_join_column.size();
unsigned int number_of_threads = foreign_key_join_column.size();

nested_loop_join_kernel<<<number_of_blocks,number_of_threads>>>(dev_primary_key_column, primary_key_join_column.size(), dev_foreign_key_column, foreign_key_join_column.size(), 
								 dev_primary_key_column_tids, dev_foreign_key_column_tids, foreign_key_join_column.size());

	cudaError err = cudaGetLastError();
    if ( cudaSuccess != err )
    {
        std::cout << "cudaCheckError() failed! Error during call to simple_join_kernel" << std::endl << "CUDA Error String: '" <<  cudaGetErrorString( err ) << "'" << std::endl;
 		std::cout << "GPU Nested Loop Join failed!" << std::endl;
		//cleanup
		cudaFree( dev_primary_key_column );
		cudaFree( dev_foreign_key_column );
		cudaFree( dev_primary_key_column_tids );
		cudaFree( dev_foreign_key_column_tids );
		return result;
	
        //exit( -1 );
    }

// copy the arrays 'a' and 'b' to the GPU
//HANDLE_ERROR( cudaMemcpy( dev_a, a, N * sizeof(int),
//cudaMemcpyHostToDevice ) );
//HANDLE_ERROR( cudaMemcpy( dev_b, b, N * sizeof(int),
//cudaMemcpyHostToDevice ) );

//add<<<N,1>>>( dev_a, dev_b, dev_c );



//shared_pointer_namespace::shared_ptr<int> primary_key_tids_ptr (new int [foreign_key_join_column.size()] ); // reseult of primary key foreign key join is always the size of the foreign key column (number of rows)
//shared_pointer_namespace::shared_ptr<int> foreign_key_tids_ptr ( new int [foreign_key_join_column.size()] );

int* primary_key_tids_ptr = new int [foreign_key_join_column.size()];
int* foreign_key_tids_ptr = new int [foreign_key_join_column.size()];

cudaMemcpy( primary_key_tids_ptr, dev_primary_key_column_tids, foreign_key_join_column.size() * sizeof(int) , cudaMemcpyDeviceToHost );
cudaMemcpy( foreign_key_tids_ptr, dev_foreign_key_column_tids, foreign_key_join_column.size() * sizeof(int) , cudaMemcpyDeviceToHost );

	std::cout << "primary key column: " << std::endl;
for(int i=0;i<foreign_key_join_column.size() ;i++){
	std::cout << primary_key_tids_ptr[i] << std::endl;
}
	std::cout << "foreign key column: " << std::endl;
for(int i=0;i<foreign_key_join_column.size() ;i++){
	std::cout << primary_key_tids_ptr[i] << std::endl;
}

// copy the array 'c' back from the GPU to the CPU
//HANDLE_ERROR( cudaMemcpy( c, dev_c, N * sizeof(int),
//cudaMemcpyDeviceToHost ) );



// free the memory allocated on the GPU
//cudaFree( dev_a );
//cudaFree( dev_b );
//cudaFree( dev_c );

// free the memory allocated on the GPU
cudaFree( dev_primary_key_column );
cudaFree( dev_foreign_key_column );
cudaFree( dev_primary_key_column_tids );
cudaFree( dev_foreign_key_column_tids );




for(int i=0;i<foreign_key_join_column.size();i++){
	result.push_back( std::pair<unsigned int, unsigned int>( primary_key_tids_ptr[i],foreign_key_tids_ptr[i]) );

}

delete[] primary_key_tids_ptr;
delete[] foreign_key_tids_ptr;

	return result;
}

/*
const std::vector<unsigned int> gpu_join_tids(const std::vector<int>& join_column1, const std::vector<int>& join_column2){

  thrust::device_vector<T> device_join_column1(join_column1.begin(), join_column1.end());
  thrust::device_vector<T> device_join_column2(join_column2.begin(), join_column2.end());

  thrust::device_vector<unsigned int> device_join_column1_tids;
  thrust::device_vector<unsigned int> device_join_column2_tids;
  
  device_join_column1_tids.resize(join_column1.size());
  device_join_column2_tids.resize(join_column2.size());
  //init lookup tables for columns
  thrust::sequence(device_join_column1_tids.begin(), device_join_column1_tids.end());
  thrust::sequence(device_join_column2_tids.begin(), device_join_column2_tids.end());
  //sort columns according to values, swap tids as well
  thrust::stable_sort_by_key(device_join_column1.begin(),device_join_column1.end(), device_join_column1_tids.begin());
  thrust::stable_sort_by_key(device_join_column2.begin(),device_join_column2.end(), device_join_column2_tids.begin());
  
  
  
  
  //thrust::sort(D.begin(),D.end());//A, A + N, thrust::greater<int>());

  //note, that the keys we want to sort are the values in the column -> the order of the "values" is the desired order of tids that describe the the new order of tuples
   //thrust::stable_sort_by_key(D.begin(),D.end(), device_tids.begin());

	return std::vector<unsigned int>();

}

*/

 // selector_equal(value);

//struct Selector_Equal {
//  __host__ __device__  Selector_Equal(int value) : value_(value),counter(0){}
//  __host__ __device__
//  bool operator() (const int& i) {bool b = (i==value_); return b; } // if(b) device_tids.push_back(counter); counter++; return b;} 
//  int value_;
//  unsigned int counter;
//  thrust::device_vector<int> device_tids;
//};


/*

__host__ __device__
const thrust::device_vector<int> gpu_selec(thrust::device_vector<int>& device_column, int value){
	//thrust::device_vector<int> d;

  struct is_equal
  {
    __host__ __device__
    bool operator()(const int &x)
    {
      return (x % 2) == 0;
    }
  } i_eq;

	return device_column; //d;
}
*/

struct saxpy_functor
{
  const float a;

  saxpy_functor(float _a) : a(_a) {}

  __host__ __device__
  float operator()(const float& x, const float& y) const
  { 
    return a * x + y;
  }
};

void saxpy_fast(float A, thrust::device_vector<float>& X, thrust::device_vector<float>& Y)
{
  // Y <- A * X + Y
  thrust::transform(X.begin(), X.end(), Y.begin(), Y.begin(), saxpy_functor(A));
}

/*
struct select_equal_functor
{
  const int a;

  select_equal_functor(int _a) : a(_a) {}

  __host__ __device__
  bool operator()(const int& x) const
  { 
    return (x==a);
  }
};
//*/

struct select_equal_functor
{
  const int a;
  

  select_equal_functor(int _a) : a(_a) {}

  

  __host__ __device__
  bool operator()(const int& x) const
  { 
    return (x==a);
  }
};

struct is_even
  {
    __host__ __device__
    bool operator()(const int x)
    {
      return (x % 2) == 0;
    }
  };

template <typename T>
struct select_equal_pair_functor
{
  const T a;
  

  select_equal_pair_functor(T _a) : a(_a) {}

  
  __host__ __device__
  bool operator()(const thrust::pair<T,unsigned int>& x) const
  { 
    return (x.first==a);
  }
};

template <typename T>
struct select_lesser_pair_functor
{
  const T a;
  

  select_lesser_pair_functor(T _a) : a(_a) {}

  
  __host__ __device__
  bool operator()(const thrust::pair<T,unsigned int>& x) const
  { 
    return (x.first<a);
  }
};

template <typename T>
struct select_greater_pair_functor
{
  const T a;
  

  select_greater_pair_functor(T _a) : a(_a) {}

  
  __host__ __device__
  bool operator()(const thrust::pair<T,unsigned int>& x) const
  { 
    return (x.first>a);
  }
};

/*
struct select_equal_pair_functor
{
  const int a;
  

  select_equal_pair_functor(int _a) : a(_a) {}

  
  __host__ __device__
  bool operator()(const thrust::pair<int,unsigned int>& x) const
  { 
    return (x.first==a);
  }
};
*/


template <typename T>
const std::vector<unsigned int> gpu_selection_tids(const std::vector<T>& column, T value, ValueComparator comp){

	std::vector<thrust::pair<T,unsigned int> > col;
	for(unsigned int i=0;i<column.size();i++){
		col.push_back(thrust::pair<T,unsigned int>(column[i],i));
	}

  	thrust::device_vector<thrust::pair<T,unsigned int> > dev_col(col.begin(),col.end());

	typedef thrust::device_vector<thrust::pair<T,unsigned int> > device_vector;

 	typename device_vector::iterator bound2;

	if(comp==EQUAL){
	bound2 = thrust::stable_partition(dev_col.begin(), dev_col.end(), select_equal_pair_functor<T>(value));
	}else if(comp==LESSER){
	bound2 = thrust::stable_partition(dev_col.begin(), dev_col.end(), select_lesser_pair_functor<T>(value));
	}else if(comp==GREATER){
	bound2 = thrust::stable_partition(dev_col.begin(), dev_col.end(), select_greater_pair_functor<T>(value));
	}else{
		std::cout << "FATAL Error! GPU selection: invalid ValueComparator: " << comp << std::endl;
	}

  std::cout << "result element number: " << bound2-dev_col.begin() << std::endl;
  std::vector<unsigned int> result_tids;
  result_tids.resize(bound2-dev_col.begin()); //set size of result_tids to number of matched tids (difference of iterators bound2-dev_col.begin())

  thrust::copy(dev_col.begin(), bound2 , col.begin()); //copy back tids

  	for(unsigned int i=0;i<bound2-dev_col.begin();i++){
		result_tids[i]=col[i].second;
	}

	return result_tids;
}


template <>
const std::vector<unsigned int> gpu_selection_tids(const std::vector<std::string>& column, std::string value, ValueComparator comp){
	std::cout << "Fatal error!!! Cannot perform selection on string columns on GPU!!!" << std::endl;
	std::exit(-1);
	return std::vector<unsigned int>();
}

/*
const std::vector<unsigned int> gpu_selection_tids(const std::vector<int>& column, int value, ValueComparator comp){

	std::vector<thrust::pair<int,unsigned int> > col;
	for(unsigned int i=0;i<column.size();i++){
		col.push_back(thrust::pair<int,unsigned int>(column[i],i));
	}

  	thrust::device_vector<thrust::pair<int,unsigned int> > dev_col(col.begin(),col.end());

 	thrust::device_vector<thrust::pair<int,unsigned int> >::iterator bound2;

	if(comp==EQUAL){
	bound2 = thrust::stable_partition(dev_col.begin(), dev_col.end(), select_equal_pair_functor<int>(value));
	}else if(comp==LESSER){
	bound2 = thrust::stable_partition(dev_col.begin(), dev_col.end(), select_lesser_pair_functor<int>(value));
	}else if(comp==GREATER){
	bound2 = thrust::stable_partition(dev_col.begin(), dev_col.end(), select_greater_pair_functor<int>(value));
	}else{
		std::cout << "FATAL Error! GPU selection: invalid ValueComparator: " << comp << std::endl;
	}

  std::cout << "result element number: " << bound2-dev_col.begin() << std::endl;
  std::vector<unsigned int> result_tids;
  result_tids.resize(bound2-dev_col.begin()); //set size of result_tids to number of matched tids (difference of iterators bound2-dev_col.begin())

  thrust::copy(dev_col.begin(), bound2 , col.begin()); //copy back tids

  	for(unsigned int i=0;i<bound2-dev_col.begin();i++){
		result_tids[i]=col[i].second;
	}

	return result_tids;
}

*/

/*
	//write functor that can be used with partition
	//apply partition algorithm with functor, first part of the dequence is result
	//copy data back

	


  //data CPU -> GPU
  thrust::device_vector<int> device_column(column.begin(), column.end());

  thrust::device_vector<int> d;




  thrust::device_vector<int>::iterator bound = thrust::stable_partition(device_column.begin(), device_column.end(), select_equal_functor(value));

//  thrust::device_vector<int>::iterator bound = thrust::stable_partition(device_column.begin(), device_column.end(),is_equal()); //selector_equal);


 // copy a device_vector into an STL vector
  std::vector<unsigned int> tids;
  std::cout << "result element number: " << bound-device_column.begin() << std::endl;
  result_tids.resize(bound-device_column.begin());

  thrust::copy(device_column.begin(), bound , result_tids.begin()); //copy back tids


	return result_tids;
}
*/

/*
const std::vector<unsigned int> gpu_selection_tids(const std::vector<int>& column, int value){


	std::vector<thrust::pair<int,unsigned int> > col;
	for(unsigned int i=0;i<column.size();i++){
		col.push_back(thrust::pair<int,unsigned int>(column[i],i));
	}

	//write functor that can be used with partition
	//apply partition algorithm with functor, first part of the dequence is result
	//copy data back




	std::vector<unsigned int> result_tids;

	//Selector_Equal selector_equal(value);

	const int N = 6;
  int V[N] = {-2, 0, -1, 0, 1, 2};
  int result[4];

  thrust::copy_if(V, V + N, result, is_even());






  //data CPU -> GPU
  thrust::device_vector<int> device_column(column.begin(), column.end());

  thrust::device_vector<int> d;

  //thrust::copy_if(device_column.begin(), device_column.end(), d.begin(), is_even());

  thrust::copy_if(device_column.begin(), device_column.end(), d.begin(), select_equal_functor(value));	

  thrust::for_each(device_column.begin(), device_column.end(), select_equal_functor(value));


  thrust::stable_partition(device_column.begin(), device_column.end(), select_equal_functor(value));

  //thrust::device_vector<int> d;
//  thrust::device_vector<int>::iterator bound = thrust::stable_partition(device_column.begin(), device_column.end(),is_equal()); //selector_equal);

//int A[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
//thrust::stable_partition(device_column.begin(), device_column.end(), d.begin() ,i_eq);

	//thrust::for_each(device_column.begin(), device_column.end(), i_eq);

	//i_eq(1);

 // copy a device_vector into an STL vector
  std::vector<unsigned int> tids;

//thrust::copy(device_tids.begin(), device_tids.end(), result_tids.begin()); //copy back tids


//(selector_equal.device_tids.size());


  //thrust::copy(selector_equal.device_tids.begin(), selector_equal.device_tids.end(), result_tids.begin()); //copy back tids

//  int A[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
//  const int N = sizeof(A)/sizeof(int);
//  thrust::stable_partition(A, A + N,
//                            is_even());

	return result_tids;
}
*/

/*
const std::vector<TID_Pair> merge_sorted_columns(vector<TID> tids_sorted_join_column1, vector<TID> tids_sorted_join_column2){

//				vector<TID> tids_sorted_join_column1 = join_column1->sort(CPU);
//				vector<TID> tids_sorted_join_column2 = join_column2->sort(CPU);

				vector<TID_Pair> join_tids;

				TID i=0;
				TID j=0;
				bool end=false;
				while(!end){
				//for(int i=0;i<join_column1->size();){
					//for(int j=0;j<join_column2->size();j++){

						boost::any value = join_column1->get(tids_sorted_join_column1[i]);
						boost::any value2 = join_column2->get(tids_sorted_join_column2[j]); //j);
						T val = any_cast<T>(value);
						T val2 = any_cast<T>(value2);		
						if(val>val2){
							j++;
						}else if (val<val2){
							i++;
						}else if(val==val2){
							//cout << "JOIN: " << val << " TID: " << i << " , " << val2 << " TID: " << j << endl;
								T tmp_val=val;
								int index=0;
								while(tmp_val==val){
//									boost::any tmp_value = join_column1->get(tids_sorted_join_column1[i+index]);
//									tmp_val = any_cast<T>(tmp_value);		
//									if(i+index==join_column1->size()) break; 
//									index++;

									if(i+index==join_column1->size()-1){ index++; break;} 
									index++;
									boost::any tmp_value = join_column1->get(tids_sorted_join_column1[i+index]);
									tmp_val = any_cast<T>(tmp_value);
								}
								//cout << "index 1..." << endl;
								tmp_val=val2;
								int index2=0;
								while(tmp_val==val2){
//									boost::any tmp_value = join_column2->get(tids_sorted_join_column2[j+index2]);
//									tmp_val = any_cast<T>(tmp_value);
//									if(j+index2==join_column2->size()) break; 
//									index2++;
				
			
									if(j+index2==join_column2->size()-1){ index2++; break;} 
									index2++;
									boost::any tmp_value = join_column2->get(tids_sorted_join_column2[j+index2]);
									tmp_val = any_cast<T>(tmp_value);
								}								
								//cout << "index 2..." << endl;
								//cout << "range: column1: " << i << "," << i+index << " column2: " << j << "," << j+index2 << endl;

//								if(index==0 && index2==0){
//										cout << "JOIN: "  << " TID: " << tids_sorted_join_column1[i] << " , " << "TID: " << tids_sorted_join_column2[j] << endl;
//										join_tids.push_back(TID_Pair(tids_sorted_join_column1[i],tids_sorted_join_column2[j]));
//								}else if(index==0){
//									for(int l=j;l<j+index2;l++){
//									//	once=false;
//										cout << "JOIN: "  << " TID: " << tids_sorted_join_column1[i] << " , " << "TID: " << tids_sorted_join_column2[l] << endl;
//										join_tids.push_back(TID_Pair(tids_sorted_join_column1[i],tids_sorted_join_column2[l]));
//									}
//								}else if(index2==0){
//									for(int k=i;k<i+index;k++){
//										cout << "JOIN: "  << " TID: " << tids_sorted_join_column1[k] << " , " << "TID: " << tids_sorted_join_column2[j] << endl;
//										join_tids.push_back(TID_Pair(tids_sorted_join_column1[k],tids_sorted_join_column2[j]));
//									}
//								}else
								for(int k=i;k<i+index;k++){
									for(int l=j;l<j+index2;l++){
									//	once=false;
										//cout << "JOIN: "  << " TID: " << tids_sorted_join_column1[k] << " , " << "TID: " << tids_sorted_join_column2[l] << endl;
										join_tids.push_back(TID_Pair(tids_sorted_join_column1[k],tids_sorted_join_column2[l]));
									}
								}
//								i=i+index+1;
//								j=j+index2+1;
								i=i+index;
								j=j+index2;
								//cout << "new i: " << i << " new j: " << j << endl;
								if(i>=join_column1->size()-1 && j>=join_column2->size()-1){
									end=true;
								}

						}
					//}
				//}
				}
				return join_tids;

}
*/

  int  check_for_CUDA_devices(){
         int device_count=0; 
     
         // get the number of non-emulation devices  detected 
         cudaGetDeviceCount( &device_count); 
			if(device_count<=0){
				std::cout << "No CUDA capable GPU found!" << std::endl;
				return -1;
			}else{
				int return_val=0; 
				for(int i=0;i<device_count;i++){
					if(CUDA_device_check(i)==-1) return_val=-1;
				}
				return return_val;
			}

  } 

  int  CUDA_device_check(int gpudevice) 
     { 
         int device_count=0; 
          int device;  // used with  cudaGetDevice() to verify cudaSetDevice() 
     
         // get the number of non-emulation devices  detected 
         cudaGetDeviceCount( &device_count); 
         if (gpudevice > device_count) 
         { 
             printf("gpudevice >=  device_count ... exiting\n"); 
             exit(1); 
         } 
         cudaError_t cudareturn; 
         cudaDeviceProp deviceProp; 
     
         // cudaGetDeviceProperties() is also  demonstrated in the deviceQuery/ example
         // of the sdk projects directory 
        cudaGetDeviceProperties(&deviceProp,  gpudevice); 
          printf("[deviceProp.major.deviceProp.minor] = [%d.%d]\n", 
            deviceProp.major, deviceProp.minor); 
     
        if (deviceProp.major > 999) 
         { 
//             printf("warning, CUDA Device  Emulation (CPU) detected, exiting\n"); 
//                    exit(1); 
					std::cout << "No CUDA capable GPU found!" << std::endl;
					std::cout << "CUDA Device  Emulation (CPU) detected!" << std::endl;
         } 
     
         // choose a cuda device for kernel  execution 
        cudareturn=cudaSetDevice(gpudevice); 
         if (cudareturn == cudaErrorInvalidDevice) 
         { 
            perror("cudaSetDevice returned  cudaErrorInvalidDevice"); 
        } 
        else 
         { 
            // double check that device was  properly selected 
            cudaGetDevice(&device); 
            printf("cudaGetDevice()=%d\n",device); 
       } 
		return 0;
} 



void init(){
	//only purpose of this function is to instantiate the needed template functions
	gpu_sort(std::vector<int>());
	gpu_sort(std::vector<unsigned int>());
	gpu_sort(std::vector<float>());
	//gpu_sort2(std::vector<double>());
	gpu_sort(std::vector<char>());	

	//not working!!!	
	gpu_sort(std::vector<bool>());
	gpu_sort(std::vector<std::string>());

	gpu_selection_tids(std::vector<int>(),0,EQUAL);
	unsigned int i=0;
	gpu_selection_tids(std::vector<unsigned int>(),i,EQUAL);
	gpu_selection_tids(std::vector<float>(),float(0.0f),EQUAL);
	gpu_selection_tids(std::vector<char>(),'c',EQUAL);

	gpu_selection_tids(std::vector<std::string>(),std::string(""),EQUAL);
	gpu_selection_tids(std::vector<bool>(),true,EQUAL);

}

}; //end namespace CogaDB



