#include <gpu/gpu_algorithms.hpp>
#include <gpu/gpu_typed_base_column.hpp>
#include <gpu/gpu_positionlist.hpp>
#include <core/column.hpp>
#include <compression/dictionary_compressed_column.hpp>
#include <thrust/scan.h>

using namespace std;

namespace CoGaDB {
	namespace gpu {

#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }
inline void gpuAssert(cudaError_t code, char *file, int line, bool abort=true)
{
   if (code != cudaSuccess) 
   {
      fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
      if (abort) exit(code);
   }
}

//kernel1: build flag array
/*
        PositionListPtr fk_column_matching_tids(new PositionList());
        PositionListPtr pk_column_tids = join_index->first->getPositionList();
        PositionListPtr fk_column_tids = join_index->second->getPositionList();

//input:
        unsigned int* matching_tids
        unsigned int* pk_column_tids;
        unsigned int* fk_column_tids;
//output: 
        char* flags;

        //intersect(pk_column_matching_tids, matching_tids) -> matching fks
        unsigned int left_pos=0;
        unsigned int right_pos=0;  
        unsigned int number_of_join_pairs=join_index->first->getPositionList()->size();
        //assume that matching tids contains no duplicates and is sorted
        while(left_pos < matching_tids->size() && right_pos < number_of_join_pairs){
            //std::cout << "Left: " << (*matching_tids)[left_pos] << "   Right: " << (*pk_column_tids)[right_pos] << std::endl;
            TID left_value = (*matching_tids)[left_pos];
            TID right_value = (*pk_column_tids)[right_pos];
            if((*matching_tids)[left_pos] == (*pk_column_tids)[right_pos]){
                fk_column_matching_tids->push_back((*fk_column_tids)[right_pos]);
                right_pos++;
            }else if((*matching_tids)[left_pos] < (*pk_column_tids)[right_pos]){
                left_pos++;
            }else if((*matching_tids)[left_pos] > (*pk_column_tids)[right_pos]){
                right_pos++;
            }
        }
        return fk_column_matching_tids;
*/

//assumes that matching tids is unique!
__device__ int binary_search(unsigned int* matching_tids, unsigned int number_of_matching_tids, unsigned int search_val){


   int low = 0;
   int high = number_of_matching_tids - 1;
   int mid = low + ((high - low) / 2);

   while (low <= high ){ //&& !(matching_tids[mid - 1] <= search_val && matching_tids[mid] > search_val)) 
       if(mid<0 || mid >= number_of_matching_tids) return number_of_matching_tids;
       //unsigned int mid = low + ((high - low) / 2);

       if (matching_tids[mid] > search_val){
           high = mid - 1;
       }else if (matching_tids[mid] < search_val){
           low = mid + 1;
       }else{
           return (TID) mid; // found
       }
       mid = low + ((high - low) / 2);
   }
 
   return number_of_matching_tids; // not found
}


			template<typename T> __device__ inline TID my_binary_index_search(T * in, T elem, long startIdx, long endIdx){
				long from = startIdx;
				long to = endIdx;
				long mid = (long) (from + to) / 2;
				
				while (from <= to && !(in[mid - 1] <= elem && in[mid] > elem))
				{
					if (in[mid] > elem)
						to = mid - 1;
					else
						from = mid + 1;
					mid = (long) (from + to) / 2;
				}
				return (mid == endIdx && in[endIdx] <= elem) ? mid + 1 : mid;
			}

                       //fetch_join_kernel<<<128,128>>>(matching_tids, number_of_matching_tids, pk_column_tids, input_array_size, flag_array.get());

__global__ void fetch_join_kernel(unsigned int* matching_tids, unsigned int number_of_matching_tids, unsigned int* pk_column_tids, unsigned int input_array_size, char* flags){
//     unsigned int i=matching_tids[number_of_matching_tids-1];
//     unsigned int j=pk_column_tids[input_array_size-1];
//     unsigned int k=flags[input_array_size-1];

//     matching_tids[number_of_matching_tids-1]=-1;
//     pk_column_tids[input_array_size-1]=-1;
//     flags[input_array_size-1]=-1;
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    while(tid<input_array_size){
int index = binary_search(matching_tids, number_of_matching_tids, pk_column_tids[tid]);
//int index = my_binary_index_search<TID>(matching_tids, pk_column_tids[tid], 0, number_of_matching_tids-1);
 //       pk_column_tids[tid]=0;
 //       flags[tid]=1;
        flags[tid]=(index<number_of_matching_tids);
        tid+=blockDim.x * gridDim.x;
    }

/*
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    while(tid<input_array_size){
        int index = binary_search(matching_tids, number_of_matching_tids, pk_column_tids[tid]);
        if(index>0){
                flags[tid]=1;
        }else{
                flags[tid]=0;
        }
        //c[tid]=a[tid]+b[tid];
        tid+=blockDim.x * gridDim.x;
    }
*/
}

//kernel2: fetch matching row ids to result buffer
__global__ void fetch_marked_tids_in_output_buffer_kernel(unsigned int* fk_column_tids, unsigned int input_array_size, char* flags, unsigned int* write_positions_array, unsigned int* result_buffer){
    //unsigned int i=matching_tids[number_of_matching_tids-1];
//    unsigned int j=fk_column_tids[input_array_size-1];
//    unsigned int k=flags[input_array_size-1];
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    while(tid<input_array_size){
        if(flags[tid]){
            unsigned int write_index=write_positions_array[tid];
            result_buffer[write_index]=fk_column_tids[tid];
        }
        tid+=blockDim.x * gridDim.x;
    }
}

                TID* getRawPointerToGPU_Positionlist(GPU_PositionlistPtr pos_list){
                       assert(pos_list!=NULL);
                       Impl_GPU_Positionlist* pos_list_matching_tids = dynamic_cast<Impl_GPU_Positionlist*>(pos_list.get());
                       assert(pos_list_matching_tids!=NULL);

                       thrust::device_vector<TID>& tmp_matching_tids = pos_list_matching_tids->getContent();
                       assert(!tmp_matching_tids.empty());

                       unsigned int* matching_tids=thrust::raw_pointer_cast(&tmp_matching_tids[0]);
                       assert(matching_tids!=NULL);
                       return matching_tids;
                }




//		const GPU_PositionlistPtr GPU_Operators::fetchMatchingTIDsFromJoinIndex(GPU_JoinIndexPtr join_index, GPU_PositionlistPtr pk_table_tids){
///*
//                       Impl_GPU_Positionlist* pos_list_matching_tids = dynamic_cast<Impl_GPU_Positionlist>(pk_table_tids.get());
//                       assert(pos_list_matching_tids!=NULL);

//                       thrust::device_vector<TID>& tmp_matching_tids = pos_list_matching_tids->getContent();
//                       assert(!tmp_matching_tids.empty());

//                       unsigned int* matching_tids=thrust::raw_pointer_cast(&tmp_matching_tids[0]);
//*/
//                       unsigned int* matching_tids=getRawPointerToGPU_Positionlist(pk_table_tids);
//                       unsigned int number_of_matching_tids=pk_table_tids->size(); 
//                       unsigned int* pk_column_tids=getRawPointerToGPU_Positionlist(join_index->pos_list_pair->first);
//                       unsigned int* fk_column_tids=getRawPointerToGPU_Positionlist(join_index->pos_list_pair->second);
//                       unsigned int input_array_size=join_index->pos_list_pair->first->size(); 
//                       //char* flags;

//std::cout << "matching_tids: " << (void*) matching_tids << std::endl;
//std::cout << "#PK_INPUT_TIDS: " << number_of_matching_tids << std::endl;
//std::cout << "pk_column_tids: " << (void*) pk_column_tids << std::endl;
//std::cout << "fk_column_tids: " << (void*) fk_column_tids << std::endl;
//std::cout << "#Rows in Join Index: " << join_index->pos_list_pair->first->size() << std::endl;

//                       thrust::device_ptr<char> flag_array = thrust::device_malloc<char>(input_array_size);
//                       assert(flag_array.get()!=NULL);
//std::cout << "flag array: " << (void*) flag_array.get() << std::endl;
//                       //thrust::device_ptr<TID> write_positions_array = thrust::device_malloc<TID>(input_array_size);
//                       thrust::device_vector<TID> write_positions_array(input_array_size);

//                       fetch_join_kernel<<<128,128>>>(matching_tids, number_of_matching_tids, pk_column_tids, input_array_size, flag_array.get());
//gpuErrchk( cudaPeekAtLastError() );
//gpuErrchk( cudaDeviceSynchronize() );

//                       thrust::exclusive_scan(flag_array, flag_array+input_array_size, write_positions_array.begin());
//              
//                       unsigned int result_size; // = write_positions_array[input_array_size-1];
///*
//unsigned int * tmp =thrust::raw_pointer_cast(write_positions_array);
//                       cudaError_t err = cudaMemcpy(&result_size, (void*)&tmp[input_array_size-1], sizeof(unsigned int), cudaMemcpyDeviceToHost);
//                       //assert(err==cudaSuccess);
//gpuErrchk(err);
//*/
//                       std::cout << "Array Size:" << write_positions_array.size() << std::endl;
//                       //result_size = write_positions_array.back();
//                       thrust::host_vector<TID> tmp (write_positions_array.begin(),write_positions_array.end());
//                       result_size = tmp.back();  
//                       std::cout << "Result size GPU Fetch Join:" <<  result_size << std::endl;
//                       thrust::device_vector<unsigned int> result_tids(result_size); //reserve space for result
//                       assert(result_tids.size()>0);

//                       fetch_marked_tids_in_output_buffer_kernel<<<128,128>>>(fk_column_tids, input_array_size, flag_array.get(), thrust::raw_pointer_cast(&write_positions_array[0]), thrust::raw_pointer_cast(&result_tids[0]));
//gpuErrchk( cudaPeekAtLastError() );
//gpuErrchk( cudaDeviceSynchronize() );
//                       // deallocate with device_free
//                       //thrust::device_free(write_positions_array);
//                       thrust::device_free(flag_array);

//                       GPU_PositionlistPtr result(new Impl_GPU_Positionlist(result_tids));
//                       return result;
//		}

		const GPU_PositionlistPtr fetchMatchingTIDsFromJoinIndex_Param(GPU_JoinIndexPtr join_index, GPU_PositionlistPtr pk_table_tids, int number_of_blocks, int number_of_threads_per_block){
                       
                       //const int number_of_blocks=512;
                       //const int number_of_threads_per_block=512;

                       unsigned int* matching_tids=getRawPointerToGPU_Positionlist(pk_table_tids);
                       unsigned int number_of_matching_tids=pk_table_tids->size(); 
                       unsigned int* pk_column_tids=getRawPointerToGPU_Positionlist(join_index->pos_list_pair->first);
                       unsigned int* fk_column_tids=getRawPointerToGPU_Positionlist(join_index->pos_list_pair->second);
                       unsigned int input_array_size=join_index->pos_list_pair->first->size(); 
                       //char* flags;

                       std::cout << "matching_tids: " << (void*) matching_tids << std::endl;
                       std::cout << "#PK_INPUT_TIDS: " << number_of_matching_tids << std::endl;
                       std::cout << "pk_column_tids: " << (void*) pk_column_tids << std::endl;
                       std::cout << "fk_column_tids: " << (void*) fk_column_tids << std::endl;
                       std::cout << "#Rows in Join Index: " << join_index->pos_list_pair->first->size() << std::endl;

                       thrust::device_ptr<char> flag_array = thrust::device_malloc<char>(input_array_size);
                       assert(flag_array.get()!=NULL);
                       std::cout << "flag array: " << (void*) flag_array.get() << std::endl;
                       thrust::device_ptr<TID> write_positions_array = thrust::device_malloc<TID>(input_array_size);
                       assert(write_positions_array.get()!=NULL);
                       //thrust::device_vector<TID> write_positions_array(input_array_size);

                       fetch_join_kernel<<<number_of_blocks,number_of_threads_per_block>>>(matching_tids, number_of_matching_tids, pk_column_tids, input_array_size, flag_array.get());
                       gpuErrchk( cudaPeekAtLastError() );
                       gpuErrchk( cudaDeviceSynchronize() );

                       //thrust::exclusive_scan(flag_array, flag_array+input_array_size, write_positions_array.begin());
                       thrust::exclusive_scan(flag_array, flag_array+input_array_size, write_positions_array);
              
                       unsigned int result_size; // = write_positions_array[input_array_size-1];
                       char last_flag;

                       unsigned int * tmp =thrust::raw_pointer_cast(write_positions_array);
                       cudaError_t err = cudaMemcpy(&result_size, (void*)&tmp[input_array_size-1], sizeof(unsigned int), cudaMemcpyDeviceToHost);
//                       gpuErrchk(err);
                       gpuErrchk( cudaPeekAtLastError() );
                       gpuErrchk( cudaDeviceSynchronize() )

                       char* tmp2 = thrust::raw_pointer_cast(flag_array);
                       err = cudaMemcpy(&last_flag, (void*)&tmp2[input_array_size-1], sizeof(char), cudaMemcpyDeviceToHost);
//                       gpuErrchk(err);
                       gpuErrchk( cudaPeekAtLastError() );
                       gpuErrchk( cudaDeviceSynchronize() );

                       //since we perform an exclusive scan, to get the write positions, 
                       //we have to check whether the last element in the flag array was a 
                       //match and adjust the result size accordingly
                       if(last_flag) result_size++;

                       std::cout << "Array Size:" <<input_array_size << std::endl;
                       //result_size = write_positions_array.back();
                       //thrust::host_vector<TID> tmp (write_positions_array.begin(),write_positions_array.end());
                       //result_size = tmp.back();  
                       std::cout << "Result size GPU Fetch Join:" <<  result_size << std::endl;
                       thrust::device_vector<unsigned int> result_tids(result_size); //reserve space for result
                       assert(result_tids.size()>0);

//                       fetch_marked_tids_in_output_buffer_kernel<<<128,128>>>(fk_column_tids, input_array_size, flag_array.get(), thrust::raw_pointer_cast(&write_positions_array[0]), thrust::raw_pointer_cast(&result_tids[0]));
                       fetch_marked_tids_in_output_buffer_kernel<<<number_of_blocks,number_of_threads_per_block>>>(fk_column_tids, input_array_size, flag_array.get(), write_positions_array.get(), thrust::raw_pointer_cast(&result_tids[0]));
                       gpuErrchk( cudaPeekAtLastError() );
                       gpuErrchk( cudaDeviceSynchronize() );
                       // deallocate with device_free
                       thrust::device_free(write_positions_array);
                       thrust::device_free(flag_array);

                       GPU_PositionlistPtr result(new Impl_GPU_Positionlist(result_tids));
                       return result;
		}

	
		const GPU_PositionlistPtr GPU_Operators::fetchMatchingTIDsFromJoinIndex(GPU_JoinIndexPtr join_index, GPU_PositionlistPtr pk_table_tids){
                       GPU_PositionlistPtr result;
                       result = fetchMatchingTIDsFromJoinIndex_Param(join_index, pk_table_tids,512,1024);
//  The following Code is currently used to derive the optimal number of blocks and blocks per threads for a GPU
//                       for(unsigned int num_blocks=64;num_blocks<64000;num_blocks*=2){
//                          for(unsigned int num_thread_per_block=64;num_thread_per_block<=1024;num_thread_per_block*=2){

//                              std::cout << "Blocks: " << num_blocks << "\t" << "Threads per Block: " << num_thread_per_block << std::endl;
//                              Timestamp begin = getTimestamp();
//                              result = fetchMatchingTIDsFromJoinIndex_Param(join_index, pk_table_tids,num_blocks,num_thread_per_block);
//                              Timestamp end = getTimestamp();
//                              std::cout << "Execution Time: " << double(end-begin)/(1000*1000) << "ms" << std::endl;

//                          }
//                       }
                       return result;
		}

	}; //end namespace gpu
}; //end namespace CogaDB



