
#include <thrust/copy.h>
#include <thrust/gather.h>
#include <gpu/gpu_base_column.hpp>
#include <gpu/gpu_positionlist.hpp>
#include <gpu/gpu_algorithms.hpp>
#include <cstdlib>
#include <assert.h>
#include <thrust/set_operations.h>
#include <thrust/sort.h>

namespace CoGaDB{
	namespace gpu{	

                       const GPU_PositionlistPtr GPU_Operators::computePositionListUnion(GPU_PositionlistPtr tids1, GPU_PositionlistPtr tids2){
                        
                           assert(tids1!=NULL);
                           assert(tids2!=NULL);
                           Impl_GPU_Positionlist* gpu_tids1 = dynamic_cast<Impl_GPU_Positionlist*>(tids1.get());
                           Impl_GPU_Positionlist* gpu_tids2 = dynamic_cast<Impl_GPU_Positionlist*>(tids2.get());
                           assert(gpu_tids1!=NULL);
                           assert(gpu_tids2!=NULL);

                           thrust::device_vector<TID>& gpu_tids_device_vector1 = gpu_tids1->getContent();
                           thrust::device_vector<TID>& gpu_tids_device_vector2 = gpu_tids2->getContent();

                           Impl_GPU_Positionlist* result_tids(new Impl_GPU_Positionlist());
                           thrust::device_vector<TID>& result_tids_vector = result_tids->getContent();

                           result_tids_vector.resize(gpu_tids_device_vector1.size()+gpu_tids_device_vector2.size());


                           thrust::device_vector<TID>::iterator result_end = thrust::set_union(gpu_tids_device_vector1.begin(),gpu_tids_device_vector1.end(),
                                                               gpu_tids_device_vector2.begin(),gpu_tids_device_vector2.end(), 
                                                               result_tids_vector.begin());

                           thrust::device_vector<TID>::iterator begin = result_tids_vector.begin();
                           unsigned int result_size=result_end-begin;
                           result_tids_vector.resize(result_size);        

                            return GPU_PositionlistPtr(result_tids);
                        }
                        const GPU_PositionlistPtr GPU_Operators::computePositionListIntersection(GPU_PositionlistPtr tids1, GPU_PositionlistPtr tids2){
                           assert(tids1!=NULL);
                           assert(tids2!=NULL);
                           Impl_GPU_Positionlist* gpu_tids1 = dynamic_cast<Impl_GPU_Positionlist*>(tids1.get());
                           Impl_GPU_Positionlist* gpu_tids2 = dynamic_cast<Impl_GPU_Positionlist*>(tids2.get());
                           assert(gpu_tids1!=NULL);
                           assert(gpu_tids2!=NULL);

                           thrust::device_vector<TID>& gpu_tids_device_vector1 = gpu_tids1->getContent();
                           thrust::device_vector<TID>& gpu_tids_device_vector2 = gpu_tids2->getContent();

                           Impl_GPU_Positionlist* result_tids(new Impl_GPU_Positionlist());
                           thrust::device_vector<TID>& result_tids_vector = result_tids->getContent();

                           //result_tids_vector.resize(std::max(gpu_tids_device_vector1.size(),gpu_tids_device_vector2.size()));
                           result_tids_vector.resize(gpu_tids_device_vector1.size()+gpu_tids_device_vector2.size());

                           thrust::device_vector<TID>::iterator result_end = thrust::set_intersection(gpu_tids_device_vector1.begin(),gpu_tids_device_vector1.end(),
                                                               gpu_tids_device_vector2.begin(),gpu_tids_device_vector2.end(), 
                                                               result_tids_vector.begin());

                           thrust::device_vector<TID>::iterator begin = result_tids_vector.begin();
                           unsigned int result_size=result_end-begin;
                           result_tids_vector.resize(result_size);         

                            return GPU_PositionlistPtr(result_tids);
                        
                        }

                        void GPU_Operators::sortPositionList(GPU_PositionlistPtr tids){
                           assert(tids!=NULL);
                           Impl_GPU_Positionlist* gpu_tids = dynamic_cast<Impl_GPU_Positionlist*>(tids.get());
                           assert(gpu_tids!=NULL);

                           thrust::device_vector<TID>& gpu_tids_device_vector = gpu_tids->getContent();
                           thrust::sort(gpu_tids_device_vector.begin(),gpu_tids_device_vector.end());
                        
                        }


			Impl_GPU_Positionlist::Impl_GPU_Positionlist(){
			
			}

			Impl_GPU_Positionlist::Impl_GPU_Positionlist(const PositionList* pos_list) {

				assert(pos_list!=NULL);

				pos_list_.resize(pos_list->size());
				thrust::copy(pos_list->begin(),pos_list->end(),pos_list_.begin());
	
			}

			Impl_GPU_Positionlist::Impl_GPU_Positionlist(const thrust::device_vector<TID>& pos_list) : pos_list_(pos_list){

			}

			Impl_GPU_Positionlist::~Impl_GPU_Positionlist(){

			}

			void Impl_GPU_Positionlist::clear(){
				pos_list_.clear();
			}
			
			const GPU_PositionlistPtr Impl_GPU_Positionlist::aggregate(GPU_PositionlistPtr dev_tids) const{ //dev_tids indexes the position list this points to, therefore dev_tids is the map and this->pos_list_ are the values of the gather operation, whereas the new created GPU position list is the out put
				shared_pointer_namespace::shared_ptr<Impl_GPU_Positionlist> dev_pos_list = shared_pointer_namespace::static_pointer_cast<Impl_GPU_Positionlist> (dev_tids);

				Impl_GPU_Positionlist* aggregated_pos_list = new Impl_GPU_Positionlist();

				aggregated_pos_list->pos_list_.resize(dev_pos_list->pos_list_.size());
				//std::cout << "Calling Gather..." << std::endl;
				 thrust::gather(dev_pos_list->pos_list_.begin(), dev_pos_list->pos_list_.end(), // begin and end of map
                				 this->pos_list_.begin(), //from this object
                 				 aggregated_pos_list->pos_list_.begin()); //from created column
            	//std::cout << "Result size: " << dev_pos_list->pos_list_.size() << " (" << dev_pos_list->pos_list_.end()-dev_pos_list->pos_list_.begin() << ")" << std::endl;     				 		std::cout << "Result size stored in vector!: " << aggregated_pos_list->pos_list_.size() << std::endl;  
				return GPU_PositionlistPtr(aggregated_pos_list);
			}

			bool Impl_GPU_Positionlist::insert(std::vector<TID>* pos_list){
				assert(pos_list!=NULL);

				pos_list_.resize(pos_list->size());
				thrust::copy(pos_list->begin(),pos_list->end(),pos_list_.begin());
	
				return true;
			}

			size_t Impl_GPU_Positionlist::size() const throw(){
				return pos_list_.size();
			}

			const std::type_info& Impl_GPU_Positionlist::type() const throw(){
				return typeid(Impl_GPU_Positionlist);
			}


			const boost::any Impl_GPU_Positionlist::get(TID tid){
				return boost::any(pos_list_[tid]);
			}


			const PositionListPtr Impl_GPU_Positionlist::toHostPositionlist() const{
				//std::cout << "toHostPositionlist: " << std::endl;
//TODO: POSSIBLE BOTTLENECK
				PositionListPtr host_vec_ptr = PositionListPtr(new PositionList(pos_list_.size()));
				
				//std::cout << "host_vec_ptr size: " << host_vec_ptr << std::endl;
				thrust::copy(pos_list_.begin(),pos_list_.end(),host_vec_ptr->begin());
				return host_vec_ptr;
			}

			void Impl_GPU_Positionlist::print() const throw(){
				thrust::host_vector<TID> v(pos_list_);
				//std::cout << "GPU Position List: " << std::endl;
				for(unsigned int i=0;i<v.size();i++){
					std::cout << v[i] << std::endl;
				}
			}

			thrust::device_vector<TID>& Impl_GPU_Positionlist::getContent(){
				return pos_list_;
			}


	}; //end namespace gpu
}; //end namespace CogaDB

