//thrust includes
#include <thrust/version.h>
#include <thrust/device_vector.h>
#include <thrust/iterator/zip_iterator.h>
#include <thrust/partition.h>
#include <thrust/copy.h>
#include <thrust/functional.h>
#include <thrust/sequence.h>
#include <thrust/generate.h>

#include <cstdlib>
#include <cstdio>
#include <iostream>

#include <gpu/gpu_algorithms.hpp>
#include <gpu/gpu_typed_base_column.hpp>
#include <gpu/gpu_positionlist.hpp>
#include <gpu/gpu_functors.hpp>

using namespace std;

namespace CoGaDB {

    namespace gpu {

        /*! \todo remove the copying of the input data from the function. Instead, use GPU_Typed_Base_Column<T>::copy to create a copy of your data (selection modifies the data during processing. Hence the input column cannot be reused again -> Instead, make copy beforehand and remove the copy in the funtion)*/

        template <typename T>
        GPU_PositionlistPtr my_sort(shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<T> > dev_col, SortOrder order) {



            assert(dev_col != NULL);

            //dev_col->print();
            //std::cout << "Allocating  " << dev_col->size() << " Elements of type " << typeid (T).name() << " on GPU" << std::endl;
            thrust::device_vector<T>& values = dev_col->getContent();



            // initialize a device_vector with the list
            thrust::device_vector<T> D(values.begin(), values.end()); //copy data!

            //thrust::device_vector<T>& D = values;

            //GPU_PositionlistPtr result_tids = GPU_PositionlistPtr(new Impl_GPU_Positionlist());
            boost::shared_ptr<Impl_GPU_Positionlist> gpu_position_list ( new Impl_GPU_Positionlist() );

	    //working on internal data of position list to avoid copying data
            thrust::device_vector<TID>& device_tids = gpu_position_list->getContent(); 
	    //set to same size as input column 
            device_tids.resize(D.size());
            // set the elements of device_tids to 0, 1, 2, 3, ...
            thrust::sequence(device_tids.begin(), device_tids.end());

            assert(D.size() == device_tids.size());

            //std::cout << "Free Device Memory: " << getFreeGPUMemorySizeInByte() << "Byte\tTotal Device Memory: " << getTotalGPUMemorySizeInByte() << "Byte" << std::endl;

            try {
		//std::cout << "GPU SORT: Input Column: " << D.size() << " entries" << std::endl;
                //thrust::copy(D.begin(), D.end(), std::ostream_iterator<T>( std::cout,"\n"));
		//std::cout << "GPU SORT: Generated IDS: " << device_tids.size() << " entries" << std::endl;
                //thrust::copy(device_tids.begin(), device_tids.end(), std::ostream_iterator<T>( std::cout,"\n"));		
                //note, that the keys we want to sort are the values in the column -> the order of the "values" is the desired order of tids that describe the the new order of tuples
                if (order == ASCENDING) {
                    //thrust::stable_sort_by_key(D.begin(),D.end(), device_tids.begin(),thrust::less_equal<T>());
                    //thrust::sort_by_key(D.begin(), D.end(), device_tids.begin(), thrust::less_equal<T>());
		    //thrust::sort_by_key(D.begin(), D.end(), device_tids.begin());
		    thrust::stable_sort_by_key(D.begin(),D.end(), device_tids.begin());
                } else if (order == DESCENDING) {
                    //thrust::stable_sort_by_key(D.begin(),D.end(), device_tids.begin(), thrust::greater_equal<T>());
		    std::cout << "DESCINDING SORTING currently not supported on GPU!" << std::endl;
		    exit(-1);
                    //thrust::sort_by_key(D.begin(), D.end(), device_tids.begin(), thrust::greater_equal<T>());
		
                } else {
                    std::cout << "FATAL ERROR: Unknown Sorting Order!" << std::endl;
                }
            } catch (thrust::system::detail::bad_alloc& e) {
                cout << e.what() << endl;
                printGPUStatus();
                return GPU_PositionlistPtr();
            } catch (thrust::system::system_error& e){
                std::cout << "FATAL ERROR: In GPU_Operators::sort()!" << std::endl;
                cout << e.what() << endl;
                //printGPUStatus();
		exit(-1);		
	    }

            //  std::cout << "Result Sorting:" << std::endl;
            //  thrust::copy(D.begin(), D.end(), std::ostream_iterator<T>( std::cout,"\n"));

            //	std::cout << "Result TIDS:" << std::endl;
            //	thrust::copy(device_tids.begin(), device_tids.end(), std::ostream_iterator<unsigned int>( std::cout,"\n"));
            //	std::cout << "==============================================" << std::endl;
            //

            return gpu_position_list;
        }

        /*
template <typename T>
struct Functor_Sort_Tuple_Less_Equal : public thrust::binary_function<thrust::tuple<T,unsigned int>&,thrust::tuple<T,unsigned int>&,bool>
{

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x, const thrust::tuple<T,unsigned int>& y) const
  {
    return ( thrust::get<0>(x)<=thrust::get<0>(y));
  }
};

template <typename T>
struct Functor_Sort_Tuple_Greater_Equal : public thrust::binary_function<thrust::tuple<T,unsigned int>&,thrust::tuple<T,unsigned int>&,bool>
{

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x, const thrust::tuple<T,unsigned int>& y) const
  {
    return ( thrust::get<0>(x)>=thrust::get<0>(y));
  }
};

template <typename T>
//const GPU_PositionlistPtr selection_stable_partition_approach(thrust::device_vector<T>& values_, T value, ValueComparator comp){
GPU_PositionlistPtr my_sort(shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<T> > dev_col, SortOrder order){



        assert(dev_col!=NULL);
        thrust::device_vector<T>& original_values = dev_col->getContent();
        //make a COPY of input vector, since this funtion will change the content of the processed data
        thrust::device_vector<T> values(original_values);

        typedef typename thrust::tuple<typename thrust::device_vector<T>::iterator,thrust::device_vector<unsigned int>::iterator> Value_TID_Iterator_Tuple;

//thrust::copy(values.begin(), values.end(), std::ostream_iterator<unsigned int>( std::cout,"\n"));
        std::cout << "Sort data with #elements=" << values.size()  << std::endl;

        //thrust::sort(values.begin(), values.end());

        thrust::device_vector<unsigned int> tids(values.size()); //reserve memory for tids according to number of values
        std::cout << "Generating TIDs data!" << std::endl;
        thrust::sequence(tids.begin(),tids.end(), 0); //fill with sequence 0,1,2,3 ...

        //first and last zip iterator, treat both sequences as one tupel sequence
        thrust::zip_iterator<Value_TID_Iterator_Tuple> first = thrust::make_zip_iterator(thrust::make_tuple(values.begin(), tids.begin()));
        thrust::zip_iterator<Value_TID_Iterator_Tuple> last  = thrust::make_zip_iterator(thrust::make_tuple(values.end(), tids.end()));

        //iterator will point to split point in array
        //thrust::zip_iterator<Value_TID_Iterator_Tuple> bound2;

        std::cout << "Call thrust::sort()" << std::endl;
        if(order==ASCENDING){

                //thrust::stable_sort_by_key(D.begin(),D.end(), device_tids.begin(),thrust::less_equal<T>());
                thrust::sort(first, last, Functor_Sort_Tuple_Less_Equal<T>()); // thrust::less_equal<Value_TID_Iterator_Tuple>());
        }else if(order==DESCENDING){
                //thrust::stable_sort_by_key(D.begin(),D.end(), device_tids.begin(), thrust::greater_equal<T>());
                thrust::sort(first, last, Functor_Sort_Tuple_Greater_Equal<T>());
                //thrust::sort(first, last, thrust::greater_equal<Value_TID_Iterator_Tuple>());
        }else{
                std::cout << "FATAL ERROR: Unknown Sorting Order!" << std::endl;
        }

//	std::cout << "Result TIDS:" << std::endl;
//	thrust::copy(result_tids.begin(), result_tids.end(), std::ostream_iterator<unsigned int>( std::cout,"\n"));
//	std::cout << "==============================================" << std::endl;


        return GPU_PositionlistPtr(new Impl_GPU_Positionlist(tids));

}
         */

        const GPU_PositionlistPtr GPU_Operators::sort(GPU_Base_ColumnPtr device_column, SortOrder order) {

            if (!device_column)
                return GPU_PositionlistPtr();

            string name = device_column->getName();
            AttributeType type = device_column->getType();
            if(!quiet && verbose && debug)
                cout << "Processing Column " << name << " with type " << type << endl;
            GPU_PositionlistPtr result_tids;
            if (type == INT) {
                shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<int> > dev_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<int> > (device_column);
                result_tids = my_sort(dev_col, order);
            } else if (type == FLOAT) {
                shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<float> > dev_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<float> > (device_column);
                result_tids = my_sort(dev_col, order);
            } else if (type == VARCHAR) {
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
                        << " Note: VARCHAR is currently not supported, will be added again in the future!" << endl;
            } else if (type == BOOLEAN) {
                //ptr=ColumnPtr(new Column<bool>(name,BOOLEAN));
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
                        << " Note: bool is currently not supported, will be added again in the future!" << endl;
            } else {
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name << endl;
            }

            return result_tids;
        }

    }; //end namespace gpu

}; //end namespace CoGaDB

