

#define ENABLE_SIMD_ACCELERATION


#include <fstream>

#include <unittests/unittests.hpp>
#include <util/tpch_benchmark.hpp>
#include <util/star_schema_benchmark.hpp>
#include <core/runtime_configuration.hpp>
#include <util/reduce_by_keys.hpp>

#include <core/gpu_column_cache.hpp>

#include <query_processing/query_processor.hpp>
#include <parser/commandline_interpreter.hpp>

#include <boost/lexical_cast.hpp>

//#include <boost/math/distributions/beta.hpp>

#include <boost/random.hpp>
#include <boost/generator_iterator.hpp>

#include <utility>
#include <parser/generated/Parser.h>

using namespace std;
using namespace CoGaDB;
using namespace query_processing;




int main(int argc, char* argv[]){

/*efficient support of null values in GPU-accelerated DBMS*/

/* \todo add Null value support!*/
	//RuntimeConfiguration::instance().setPathToDatabase("./ssb_sf1");

//	Unittest_Create_Star_Schema_Benchmark_Database("/media/1A2AA9012AA8DAD1_/benchmarks/star_schema_benchmark/SF1/");

//	return 0;

	cout << "CoGaDB version " << COGADB_VERSION << endl;
        
#ifdef ENABLE_SIMD_ACCELERATION
        cout << "SIMD ACCELERATION ENABLED" << endl;
#else
        cout << "SIMD ACCELERATION DISABLED" << endl;        
#endif
       
#ifdef FORCE_USE_UNSTABLE_SORT_FOR_BENCHMARK
 cout << "Warning: Compiled with FORCE_USE_UNSTABLE_SORT_FOR_BENCHMARK defined, order by for multiple columns will not work!" << endl;        
#endif        
        
        //cout << "Copyright: Database and Information Systems Group, Otto von Guericke University Magdeburg, Germany" << endl;
        //cout << "For questions and bug reports contact the development team by Sebastian Breß: bress@iti.cs.uni-magdeburg.de" << endl;
        cout << "Enter 'help' for instructions" << endl;
        
// cout << "   _____       _____       _____  ____     ___   ___  " << endl;
// cout << "  / ____|     / ____|     |  __ \\|  _ \\   / _ \\ |__ \\ " << endl;
// cout << " | |     ___ | |  __  __ _| |  | | |_) | | | | |   ) |" << endl;
// cout << " | |    / _ \\| | |_ |/ _` | |  | |  _ <  | | | |  / / " << endl;
// cout << " | |___| (_) | |__| | (_| | |__| | |_) | | |_| | / /_ " << endl;
// cout << "  \\_____\\___/ \\_____|\\__,_|_____/|____/   \\___(_)____|" << endl;

	std::string prompt("CoGaDB>");
	std::string input;
        CommandLineInterpreter cmd;

//   typedef bool (*SimpleCommandHandlerPtr)();
//	typedef std::map<string,SimpleCommandHandlerPtr> CommandMap;
//	CommandMap command_map;

//	command_map.insert(std::make_pair("loaddatabase",&loadTables));
//	command_map.insert(std::make_pair("unittests",&CoGaDB::unit_tests::executeUnitTests));
//	command_map.insert(std::make_pair("help",&help));	

//    Parser parser;

//    parser.parse();
        
	//RuntimeConfiguration::instance().setPrintQueryPlan(true);

        //read commands from startup.coga
        {
            ifstream in_stream;
            //cout << "Executing coga script file: " << argv[1] << endl;
            in_stream.open("startup.coga");
            if(in_stream.good()){
                while(!in_stream.eof()){
                    std::getline(in_stream, input);
                    //cout << "Execute Command " << input << endl;
                    if(input=="quit") return 0;
                    if(!cmd.execute(input)){ cout << "Error! Command '" << input << "' failed!" << endl; return -1;}
                }
            }
        }
        
        //if script file was specified, execute script and exit
        if(argc>1){
            ifstream in_stream;
            cout << "Executing coga script file: " << argv[1] << endl;
            in_stream.open(argv[1]);
            
            while(!in_stream.eof()){
                std::getline(in_stream, input);
                //cout << "Execute Command '" << input << "'" << endl;
                if (input.substr(0, 12) == "parallelExec") {
                    cmd.parallelExecution(input, &in_stream);
                    return 0;
                }
                if(input=="quit") return 0;
//                if(!cmd.execute(input)){ cout << "Error! Command '" << input << "' failed!" << endl; return -1;}
                CoGaDB::Timestamp begin = CoGaDB::getTimestamp();
                bool ret = cmd.execute(input);
                CoGaDB::Timestamp end = CoGaDB::getTimestamp();
                if(!ret){ 
                    cout << "Error! Command '" << input << "' failed!" << endl; return -1;
                }else{
                    cout << input << "\t(" << double(end-begin)/(1000*1000) << "ms)" << endl;
                }
            }

            in_stream.close();
            return 0;
        }    
        
        //show interactive shell
	while(true){
		if (cmd.getline(prompt, input)) {
			cout << endl;
			return 0;
		}

		if(input=="quit") return 0;

		if(!cmd.execute(input)) cout << "Error! Command '" << input << "' failed!" << endl;
	

//		CommandMap::iterator it = command_map.find(input); 
//		if(it!=command_map.end()){
//			it->second();
//		}else{
//			cout << "Error! Command '" << input << "' not found!" << endl;
//		}		


	}

 return 0;
}
