#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <fstream>
#include <iostream>

#include <core/table.hpp>
#include <lookup_table/lookup_table.hpp>
#include <lookup_table/join_index.hpp>
#include <parser/commandline_interpreter.hpp>
#include <boost/tokenizer.hpp>


#include <unittests/unittests.hpp>
#include <util/tpch_benchmark.hpp>
#include <util/star_schema_benchmark.hpp>
#include <core/runtime_configuration.hpp>

#include <boost/lexical_cast.hpp>

#include <boost/random.hpp>
#include <boost/generator_iterator.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/thread.hpp>
#include <boost/bind.hpp>

#include <utility>
#include <string>
#include <queue>
#include <algorithm>

#include <parser/generated/Parser.h>
#include <query_processing/query_processor.hpp>
#include <core/gpu_column_cache.hpp>

#include <sql/server/sql_driver.hpp>
#include <hype.h>
#include <core/data_dictionary.hpp>

#ifdef LIBREADLINE_FOUND
#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>
#endif

using namespace std;

namespace CoGaDB {
    using namespace query_processing;
    //hype::DeviceConstraint default_device_constraint(hype::ANY_DEVICE);

    
//    bool optimize_and_execute(const std::string& query_name, LogicalQueryPlan& log_plan){
//        if(RuntimeConfiguration::instance().getPrintQueryPlan()){
//            log_plan.print();
//        }
//        
//        CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();
//        if(RuntimeConfiguration::instance().getPrintQueryPlan()){
//            cout << "Executing Query " << query_name << ":" << endl;
//            plan->print();
//        }
//        //execute the query
//        plan->run();
//        if(RuntimeConfiguration::instance().getPrintQueryPlan()){
//            plan->print();
//        }
//        TablePtr result = plan->getResult();
//        if(!result) return false;
//        result->print();
//        return true;
//    }
    
    
    bool printStatusOfGPUCache() {
        GPU_Column_Cache::instance().printStatus();
        return true;
    }
    
    bool printHypeStatus(){
        hype_printStatus();
        return true;
    }

//    static char** cogadb_completion( const char * text, int start,  int end)
//    {
//        char **matches;
//
//        matches = (char **)NULL;
//
//        if (start == 0)
//            matches = rl_completion_matches ((char*)text, &autocompletion_candidate_generator);
//        else
//            rl_bind_key('\t',rl_abort);
//
//        return (matches);
//
//    }

//    char* autocompletion_candidate_generator(const char* text, int state)
//    {
//        static int list_index, len;
//        char *name;
//
//        if (!state) {
//            list_index = 0;
//            len = strlen (text);
//        }
//
//        while (name = cmd[list_index]) {
//            list_index++;
//
//            if (strncmp (name, text, len) == 0)
//                return (dupstr(name));
//        }
//
//        /* If no names matched, then return NULL. */
//        return ((char *)NULL);
//
//    }
//
//    char * createNewCString (char* s) {
//      char *r;
//      //use C function, because Readline will call free() later
//      r = (char*) malloc ((strlen (s) + 1));
//      assert(r!=NULL);
//      strcpy (r, s);
//      return (r);
//    }
    
    
    
    bool simple_SSB_Query_Selection() {
        //hype::DeviceConstraint default_device_constraint(hype::ANY_DEVICE); //hype::CPU_ONLY); //if you have only a CPU
        hype::DeviceConstraint default_device_constraint = CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint();
                
        boost::shared_ptr<logical_operator::Logical_Scan> scan_lineorder(new logical_operator::Logical_Scan("LINEORDER"));


        //Attribut(INT,"YEARMONTHNUM")); //  numeric (YYYYMM) -- e.g. 199803

        boost::shared_ptr<logical_operator::Logical_Selection> selection_lineorder(new logical_operator::Logical_Selection("LO_DISCOUNT", boost::any(float(0)), GREATER, LOOKUP, default_device_constraint));

        selection_lineorder->setLeft(scan_lineorder);

        LogicalQueryPlan log_plan(selection_lineorder);
        log_plan.print();

        //cout << "Executing Query: " << endl;
        CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();
        plan->print();

        //Timestamp begin=getTimestamp();
        plan->run();
        //Timestamp end=getTimestamp();
        //assert(end>=begin);
        //cout << "Needed " << double(end-begin)/(1000*1000) << "ms ("<< double(end-begin)/(1000*1000*1000) <<"s) to process query '" << "1" <<  "' ..." << endl;
        //cout << "Result has " << plan->getResult()->getNumberofRows() << " rows" << endl;

        return true;
    }

    bool simple_SSB_Query_Join() {
        //hype::DeviceConstraint default_device_constraint(hype::ANY_DEVICE); //hype::CPU_ONLY); //if you have only a CPU
        hype::DeviceConstraint default_device_constraint = CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint();
        
        boost::shared_ptr<logical_operator::Logical_Scan> scan_lineorder(new logical_operator::Logical_Scan("LINEORDER"));
        boost::shared_ptr<logical_operator::Logical_Scan> scan_date(new logical_operator::Logical_Scan("DATE"));

        //Attribut(INT,"YEARMONTHNUM")); //  numeric (YYYYMM) -- e.g. 199803

        boost::shared_ptr<logical_operator::Logical_Selection> selection_date(new logical_operator::Logical_Selection("D_YEARMONTHNUM", boost::any(199803), EQUAL, LOOKUP, default_device_constraint));

        boost::shared_ptr<logical_operator::Logical_Join> join(new logical_operator::Logical_Join("D_DATEKEY", "LO_ORDERDATE", LOOKUP, hype::DeviceConstraint(hype::CPU_ONLY))); //GPU Join not supported


        selection_date->setLeft(scan_date);
        join->setLeft(selection_date);
        join->setRight(scan_lineorder);

        LogicalQueryPlan log_plan(join);
        log_plan.print();

        //cout << "Executing Query: " << endl;
        CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();
        plan->print();
        //Timestamp begin=getTimestamp();
        plan->run();
        //Timestamp end=getTimestamp();
        //assert(end>=begin);
        //cout << "Needed " << double(end-begin)/(1000*1000) << "ms ("<< double(end-begin)/(1000*1000*1000) <<"s) to process query '" << "1" <<  "' ..." << endl;
        //cout << "Result has " << plan->getResult()->getNumberofRows() << " rows" << endl;

        return true;
    }


    bool simple_SSB_Query_Aggregation() {
        hype::DeviceConstraint default_device_constraint = CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint();
        int value = 22;
        boost::shared_ptr<logical_operator::Logical_Scan> scan_lineorder(new logical_operator::Logical_Scan("LINEORDER"));
        boost::shared_ptr<logical_operator::Logical_Scan> scan_date(new logical_operator::Logical_Scan("DATE"));
        //boost::shared_ptr<logical_operator::Logical_ColumnConstantOperator>  column_constant_operation_sale(new logical_operator::Logical_ColumnConstantOperator("Price",value,"Add_Price_Value_LogOp",ADD,hype::DeviceConstraint(hype::CPU_ONLY)));

        boost::shared_ptr<logical_operator::Logical_Selection> selection_date(new logical_operator::Logical_Selection("D_YEAR", boost::any(1993), EQUAL, LOOKUP, default_device_constraint));
        boost::shared_ptr<logical_operator::Logical_Join> join(new logical_operator::Logical_Join("D_DATEKEY", "LO_ORDERDATE", LOOKUP, default_device_constraint)); //GPU Join not supported


        selection_date->setLeft(scan_date);
        join->setLeft(selection_date);
        join->setRight(scan_lineorder);

        std::list<std::string> column_list;
        column_list.push_back("D_YEAR");
        column_list.push_back("LO_EXTENDEDPRICE");
        column_list.push_back("LO_DISCOUNT");
        boost::shared_ptr<logical_operator::Logical_Projection> projection(new logical_operator::Logical_Projection(column_list));

        projection->setLeft(join);

        boost::shared_ptr<logical_operator::Logical_ColumnAlgebraOperator> column_algebra_operation_lineorder(new logical_operator::Logical_ColumnAlgebraOperator("LO_EXTENDEDPRICE", "LO_DISCOUNT", "REVENUE", MUL, default_device_constraint));

        column_algebra_operation_lineorder->setLeft(projection);


        std::list<std::string> sorting_column_names;
        sorting_column_names.push_back("D_YEAR");
        //column_names.push_back("Score");

        boost::shared_ptr<logical_operator::Logical_Sort> sort(new logical_operator::Logical_Sort(sorting_column_names, ASCENDING, LOOKUP, default_device_constraint)); //operation on String column only supported on CPU

        sort->setLeft(column_algebra_operation_lineorder);

        std::list<std::pair<string, AggregationMethod> > aggregation_functions;
        aggregation_functions.push_back(make_pair("REVENUE", SUM));

        boost::shared_ptr<logical_operator::Logical_Groupby> groupby(new logical_operator::Logical_Groupby(sorting_column_names, aggregation_functions, MATERIALIZE, default_device_constraint));

        groupby->setLeft(sort);

        LogicalQueryPlan log_plan(groupby); //column_algebra_operation_lineorder);
        log_plan.print();

        CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();
        plan->print();
        plan->run();

        if (!plan->getResult())
            return false;

        plan->getResult()->print();
        //boost::shared_ptr<logical_operator::Logical_Scan>  scan_sale(new logical_operator::Logical_Scan("Sale"));
        //boost::shared_ptr<logical_operator::Logical_ColumnAlgebraOperator>  column_algebra_operation_sale(new logical_operator::Logical_ColumnAlgebraOperator("Price","Sales","Add_Price_Sales_LogOp",ADD,hype::DeviceConstraint(hype::CPU_ONLY)));
        return true;
    }

    bool simple_SSB_Queries() {
        typedef bool (*SimpleCommandHandlerPtr)();
        typedef std::map<std::string, SimpleCommandHandlerPtr> SimpleCommandMap;
        SimpleCommandMap map;
//        map.insert(std::make_pair("Q1", &simple_SSB_Query_Join));
//        map.insert(std::make_pair("Q2", &simple_SSB_Query_Selection));
//        map.insert(std::make_pair("Q3", &simple_SSB_Query_Aggregation));
        map.insert(std::make_pair("ssb11", &SSB_Q11));
        map.insert(std::make_pair("ssb12", &SSB_Q12));
        map.insert(std::make_pair("ssb13", &SSB_Q13));
        map.insert(std::make_pair("ssb21", &SSB_Q21));
        map.insert(std::make_pair("ssb22", &SSB_Q22));
        map.insert(std::make_pair("ssb23", &SSB_Q23));
        map.insert(std::make_pair("ssb31", &SSB_Q31));
        map.insert(std::make_pair("ssb32", &SSB_Q32));
        map.insert(std::make_pair("ssb33", &SSB_Q33));
        map.insert(std::make_pair("ssb34", &SSB_Q34));
        map.insert(std::make_pair("ssb41", &SSB_Q41));
        map.insert(std::make_pair("ssb42", &SSB_Q42));
        map.insert(std::make_pair("ssb43", &SSB_Q43));        
        string input;
        cout << "Star Schema Query Test" << endl;
        cout << "Type 'back' to return to cmd" << endl;
        for (SimpleCommandMap::iterator it = map.begin(); it != map.end(); ++it) {
            cout << it->first << endl;
        }
        //workaroudn to use the CommandLineInterpreters History in this sub shell
        CommandLineInterpreter cmd;
        std::string prompt="CoGaDB>";
		
        while (true) {
            //cout << "CoGaDB>"; // << endl;
//            std::getline(std::cin, input);
//            if (cin.eof()) {
//                cout << endl;
//                return true;
//            }            
            if (cmd.getline(prompt, input)){ continue;}

            if (input == "back") return true;

            SimpleCommandMap::iterator it = map.find(input);
            if (it != map.end()) {
                cout << "Executing Query: " << it->first << endl;
                Timestamp begin = getTimestamp();
                if (!it->second()) cout << "Error: while processing Query '" << input << "'" << endl;
                Timestamp end = getTimestamp();
                cout << "Needed " << double(end - begin) / (1000 * 1000) << "ms (" << double(end - begin) / (1000 * 1000 * 1000) << "s) to process query '" << it->first << "' ..." << endl;

            } else {
                cout << "Query '" << input << "' not found" << endl;
            }
            //if(!cmd.execute(input)) cout << "Error! Command '" << input << "' failed!" << endl;

        }

        return true;
    }


    bool complex_tpch_queries() {
        hype::DeviceConstraint default_device_constraint = CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint();
        cout << "Running Tests based on the TPC-H benchmark dataset" << endl;

        //scan operations
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_nation(new logical_operator::Logical_Scan("NATION"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_region(new logical_operator::Logical_Scan("REGION"));
        boost::shared_ptr<logical_operator::Logical_Scan> scan_part(new logical_operator::Logical_Scan("PART"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_supplier(new logical_operator::Logical_Scan("SUPPLIER"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_partsupp(new logical_operator::Logical_Scan("PARTSUPP"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_customer(new logical_operator::Logical_Scan("CUSTOMER"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_orders(new logical_operator::Logical_Scan("ORDERS"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_lineitem(new logical_operator::Logical_Scan("LINEITEM"));


        /*
        Ich suche alle Lieferanten X, die ein Produkt von HERSTELLER Y verkaufen, welches einen Wert größer als Z hat. Dabei werden nur Produkte betrachtet, deren Lagerstand < A ist.
         */

        //T1 Suche alle Nationen einer Region -> EUROPE heraus
        //boost::shared_ptr<logical_operator::Logical_Selection> regionselection(new logical_operator::Logical_Selection("R_NAME", boost::any(std::string("EUROPE")), EQUAL, LOOKUP,  hype::DeviceConstraint(hype::CPU_ONLY)));
        //boost::shared_ptr<logical_operator::Logical_Join> regionnationjoin(new logical_operator::Logical_Join("R_REGIONKEY", "N_REGIONKEY", LOOKUP, default_device_constraint)); //GPU Join not supported
        //T2 Suche alle Produkte eines Herstellers heraus die einen gewissen Preis haben
        boost::shared_ptr<logical_operator::Logical_Selection> partselection1(new logical_operator::Logical_Selection("P_SIZE", boost::any(35), LESSER, LOOKUP, default_device_constraint));
        boost::shared_ptr<logical_operator::Logical_Selection> partselection2(new logical_operator::Logical_Selection("P_RETAILPRICE", boost::any(float(1100.99)), GREATER, LOOKUP, default_device_constraint));
        //T3 Suche Zuordnungen heraus, die eine bestimmte Anzahl haben und joine mit T2
        //boost::shared_ptr<logical_operator::Logical_Selection> partsuppselection(new logical_operator::Logical_Selection("PS_AVAILQTY", boost::any(1000), LESSER, LOOKUP, default_device_constraint));
        //boost::shared_ptr<logical_operator::Logical_Join> partpartsuppjoin(new logical_operator::Logical_Join("P_PARTKEY", "PS_PARTKEY", LOOKUP, default_device_constraint)); //GPU Join not supported
        //T4 Suche alle Supplier einer Region heruas (mit Hilfe von T1)
        //boost::shared_ptr<logical_operator::Logical_Join> t1supplierjoin(new logical_operator::Logical_Join("N_NATIONKEY", "S_NATIONKEY", LOOKUP, default_device_constraint)); //GPU Join not supported
        //T5 führe T3 und T4 zusammen
        // boost::shared_ptr<logical_operator::Logitidscal_Join> t3t4join(new logical_operator::Logical_Join("S_SUPPKEY", "PS_SUPPKEY", LOOKUP, default_device_constraint)); //GPU Join not supported
        //T6 sortiere die Ergebnisse nach verfügbarer Menge
        //boost::shared_ptr<logical_operator::Logical_Sort> availqtysort(new logical_operator::Logical_Sort("PS_AVAILQTY", ASCENDING, LOOKUP, default_device_constraint)); //operation on String column only supported on CPU
        //T7 Pjection
        //std::list<std::string> column_list;
        // column_list.push_back("S_NAME");
        // column_list.push_back("P_NAME");
        //column_list.push_back("S_ACCTBAL");
        //column_list.push_back("P_RETAILPRICE");
        //boost::shared_ptr<logical_operator::Logical_Projection> projection(new logical_operator::Logical_Projection(column_list)); //GPU Projection not supported



        //Führe T1 aus
        //regionselection->setLeft(scan_region);

        //regionnationjoin->setLeft(regionselection);
        //regionnationjoin->setRight(scan_nation);
        //Führe T2 aus
        partselection1->setLeft(scan_part);
        partselection2->setLeft(partselection1);
        //Führe T3 aus
        //partsuppselection->setLeft(scan_partsupp);

        //partpartsuppjoin->setLeft(partselection2);
        //partpartsuppjoin->setRight(partsuppselection);
        //Führe T4 aus
        //t1supplierjoin->setLeft(regionnationjoin);
        //t1supplierjoin->setRight(scan_supplier);
        //Führe T5 aus
        //t3t4join->setLeft(t1supplierjoin);
        //t3t4join->setRight(partpartsuppjoin);
        //Führe T6 aus
        //availqtysort->setLeft(partpartsuppjoin);
        //Führe T7 aus
        //projection->setLeft(partselection2);


        //Erstelle logischen Plan
        LogicalQueryPlan log_plan(partselection2);

        log_plan.print();
        //Führe Physischen Plan 30 mal aus
        for (unsigned int i = 0; i < 30; i++) {
            cout << "Executing Query: " << endl;
            CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();

            plan->print();

            plan->run();

            //typename PhysicalQueryPlan::Type
            TablePtr ptr = plan->getResult();
            if (ptr) {
                // ptr->print();
            } else {
                cout << "Error while processing query: pointer to result table is NULL, no results could be retrieved" << endl;

                return false;
            }
        }

        return true;
    }

    bool printDatabaseSchema() {

        std::vector<TablePtr>& tables = getGlobalTableList();
        cout << "Tables in database '" << RuntimeConfiguration::instance().getPathToDatabase() << "':" << endl;

        for (unsigned int i = 0; i < tables.size(); i++) {

            cout << tables[i]->getName() << " (";

            TableSchema schema = tables[i]->getSchema();
            TableSchema::iterator it;
            for (it = schema.begin(); it != schema.end(); ++it) {

                if (it != schema.begin()) cout << ",";
                cout << it->second;

            }

            cout << ") [" << tables[i]->getNumberofRows() << " rows]" << endl;
        }
        return true;
    }

    bool toggleQueryChopping() {
        RuntimeConfiguration::instance().setQueryChoppingEnabled(!RuntimeConfiguration::instance().isQueryChoppingEnabled());
        if (RuntimeConfiguration::instance().isQueryChoppingEnabled()) {
            cout << "Query Chopping is now activated." << endl;
        } else {
            cout << "Query Chopping is now deactivated." << endl;
        }
		return true;
    }

    bool about(){
        
        cout << "************************************************************************************************************" << endl;
        cout << "Copyright (c) 2014, Sebastian Breß, Otto-von-Guericke University of Magdeburg, Germany. All rights reserved." << endl;
        cout << "" << endl;
        cout << "This program and accompanying materials are made available under the terms of the " << endl;
        cout << "GNU GENERAL PUBLIC LICENSE - Version 3, http://www.gnu.org/licenses/gpl-3.0.txt" << endl;
        cout << "************************************************************************************************************" << endl;
        
        cout << endl << "Credits:" << endl << endl;
        cout << "Project members:" << endl;
        cout << "\tSebastian Breß (University of Magdeburg)" << endl;
        cout << "\tRobin Haberkorn (University of Magdeburg)" << endl;
        cout << "\tSteven Ladewig (University of Magdeburg)" << endl;
        cout << "\tTobias Lauer (Jedox AG)" << endl;
        cout << "\tManh Lan Nguyen (University of Magdeburg)" << endl;
        cout << "\tGunter Saake (University of Magdeburg)" << endl;
        cout << "\tNorbert Siegmund (University of Passau)" << endl;
        cout << endl;
        
        cout << "Contributors:" << endl;
        cout << "\tDarius Brückers (contributed Compression Technique: Run Length Encoding)"  << endl;
        cout << "\tSebastian Krieter (contributed Compression Technique: Delta Coding)"  << endl;
        cout << "\tSteffen Schulze (contributed Compression Technique: Bit Vector Encoding)"  << endl;
        cout << endl;

        cout << "Former project members:" << endl;
        cout << "\tRené Hoyer (University of Magdeburg)" << endl;
        cout << "\tPatrick Sulkowski (University of Magdeburg)" << endl;
        cout << endl;

        return true;
    }
    
    bool version(){
        cout << "CoGaDB " << COGADB_VERSION << endl;
        return true;
    }
    
    bool printSystemTableIntegrityConstraints(){
        TablePtr tab = DataDictionary::instance().getTableWithIntegrityConstraints();
        if(!tab) return false;
        std::cout << tab->toString() << std::endl;
        return true;
    }
    
    bool printSystemTableJoinIndexes(){
        TablePtr tab = JoinIndexes::instance().getSystemTable();
        if(!tab) return false;
        std::cout << tab->toString() << std::endl;
        return true;
    }
        
        
    bool help() {

           cout << "supported commands:" << endl
                << "loaddatabase\t\t\tloads complete database in main memory" << endl
                << "unittests\t\t\tperforms a self check of CoGaDB" << endl
                << "printschema\t\t\tprints the schema of the active database" << endl
                << "showgpucache\t\t\tprints status information of the GPU column cache"
                << "simple_ssb_queries\tsimple demonstrator for queries on SSB Benchmark data set" << endl
                << "set <variablename>=<variablevalue>" << endl
                << "print <variable>\t\tprint value of variable" << endl
                << "\t" << "Variables:" << endl
                << "\t" << "path_to_database\tabsolute or relative path to directory where the database is stored" << endl
                << "\t" << "optimizer\t\tname of optimizer used to optimize query plans (default_optimizer, star_join_optimizer)" << endl
                << "\t" << "print_query_plan\tprint the generated query plans for a SQL query (true,false)" << endl
                << "\t" << "enable_profiling\tprint the query execution plan after execution with timings for each operator (true,false)" << endl
                << "exec <SQL>\t\t\tExecute SQL statements" << endl
                << "explain <SQL>\t\t\tDisplay query plan generated from SQL expression" << endl
                << "explain_unoptimized <SQL>\tAs above, but does not apply logical optimizer before showing the plan" << endl
                << "hypestatus \t\t\tPrints all operations and corresponding algorithms registered in HyPE for CoGaDB'S operators" << endl
                << "integrityconstraints \t\tPrints integrity constraints configured for current database" << endl
                << "joinindexes \t\t\tPrints join indexes for current database" << endl
                << "toggleQC\t\t\tToggle the state of Query Chopping activation. Per default QC is off." << endl
                << "ssbXY \t\t\t\tExecute SSB-Query X.Y (X has to be a number between 1 and 4; Y has to be a number between 1 and 3 except when X is 3, in this case 4 is valid for Y as well)" << endl
                << "setdevice <DEVICE> \t\tSets the default device, which is used for execution. Possible values are 'cpu', 'gpu' or 'any' to use either the CPU or the GPU or both." << endl << endl
                << "setparallelizationmode <PARALLELIZATION MODE> \tSets the default parallelization mode for Subplans generated during Two Phase Physical Optimization (TOPPO) in the second phase (currently only for complex selections). Valid values are 'serial' and 'parallel'" << endl
                << "create_tpch_database <path to *.tbl files>\timport tables of TPC-H benchmark in CoGaDB" << endl
                << "create_ssb_database <path to *.tbl files>\timport tables of star schema benchmark in CoGaDB" << endl << endl
                << "about \t shows credits" << endl
                << "version\t shows version of CoGaDB" << endl
                << "quit" << endl;

        return true;
    }

    bool setVariable(const std::string& input) {

        unsigned int tokencounter = 0;
        typedef boost::tokenizer<boost::char_separator<char> >
                tokenizer;
        boost::char_separator<char> sep("=");
        tokenizer tok(input, sep);
        std::string variable_name;
        std::string parameter;
        for (tokenizer::iterator beg = tok.begin(); beg != tok.end(); ++beg) {
            //cout << *beg << "\n";
            if (tokencounter == 0) variable_name = *beg;
            if (tokencounter == 1) parameter = *beg;
            tokencounter++;
        }

        if (tokencounter < 2) {
            cout << "Error! For Command SET: Invalid Expression: \"" << input << "\" <variable>=<value>" << endl;
            return false; 
        }
        if (tokencounter > 2) {
            cout << "Error! For Command SET: To many parameters: <variable>=<value>" << endl;
            return false;
        }



        if (variable_name == "path_to_database") {

            RuntimeConfiguration::instance().setPathToDatabase(parameter);
        }else if (variable_name == "optimizer"){
            if(parameter!="default_optimizer" && parameter!="star_join_optimizer"){
                cout << "Error! For Command SET: Invalid parameter: " << parameter << endl;
                cout << "Valid values are: default_optimizer, star_join_optimizer" << endl;
                return false;
            }
            RuntimeConfiguration::instance().setOptimizer(parameter);
        }else if (variable_name == "print_query_plan"){
            if(parameter=="true" 
            || parameter=="TRUE"
            || parameter=="ON"){
                RuntimeConfiguration::instance().setPrintQueryPlan(true);
            }else if(parameter=="false"
            || parameter=="FALSE"
            || parameter=="OFF"){
                RuntimeConfiguration::instance().setPrintQueryPlan(false);
            }else{
                cout << "Error! For Command SET: Invalid parameter: " << parameter << endl;
                cout << "Valid values are: true, false" << endl;
                return false;
            }
        }else if (variable_name == "enable_profiling"){
            if(parameter=="true" 
            || parameter=="TRUE"
            || parameter=="ON"){
                RuntimeConfiguration::instance().setProfileQueries(true);
            }else if(parameter=="false"
            || parameter=="FALSE"
            || parameter=="OFF"){
                RuntimeConfiguration::instance().setProfileQueries(false);
            }else{
                cout << "Error! For Command SET: Invalid parameter: " << parameter << endl;
                cout << "Valid values are: true, false" << endl;
                return false;
            }
        }else{
                cout << "Error! For Command SET: Unknown variable: '" << variable_name << "'" << endl;
                return false;
        }
        cout << "SET Variable: '" << variable_name << "' to '" << parameter << "'" << endl;
        return true;
    }

    bool setDefaultDevice(const std::string& input) {
        if (input == "cpu") {
            //default_device_constraint = hype::DeviceConstraint(hype::CPU_ONLY);
            CoGaDB::RuntimeConfiguration::instance().setGlobalDeviceConstraint(hype::DeviceConstraint(hype::CPU_ONLY));
            cout << "Set device to CPU." << endl;
        } else if (input == "gpu") {
            //default_device_constraint = hype::DeviceConstraint(hype::GPU_ONLY);
            CoGaDB::RuntimeConfiguration::instance().setGlobalDeviceConstraint(hype::DeviceConstraint(hype::GPU_ONLY));
            cout << "Set device to GPU." << endl;
        } else if (input == "any") {
            //default_device_constraint = hype::DeviceConstraint(hype::ANY_DEVICE);
            CoGaDB::RuntimeConfiguration::instance().setGlobalDeviceConstraint(hype::DeviceConstraint(hype::ANY_DEVICE));            
            cout << "Set device to any device." << endl;
        } else {
            cout << "No device with the given devicecode could be found. Try 'cpu', 'gpu' or 'any' to switch the defaultdevice." << endl;

            return false;
        }
        return true;
    }
    
    
        bool setDefaultParallelizationModeTwoPhasePhysicalOptimization(const std::string& input) {
        if (input == "serial") {
            //default_device_constraint = hype::DeviceConstraint(hype::CPU_ONLY);
            CoGaDB::RuntimeConfiguration::instance().setParallelizationModeForTwoPhasePhysicalOptimizationQueryPlans(CoGaDB::SERIAL);
            cout << "Set TOPPO parallelization mode to serial." << endl;
        } else if (input == "parallel") {
            //default_device_constraint = hype::DeviceConstraint(hype::GPU_ONLY);
            CoGaDB::RuntimeConfiguration::instance().setParallelizationModeForTwoPhasePhysicalOptimizationQueryPlans(CoGaDB::PARALLEL);
            cout << "Set TOPPO parallelization mode to parallel." << endl;
        } else {
            cout << "Unknown TOPPO Parallelization mode! Try 'serial' or 'parallel' to switch the TOPPO (Two Phase Physical Optimization) parallelization mode." << endl;

            return false;
        }
        return true;
    }
    
    
    
    

    bool printVariableValue(const std::string& input){

        if (input == "path_to_database") {
            cout << "'" << RuntimeConfiguration::instance().getPathToDatabase() << "'" << endl;
        }else if (input == "optimizer") {
            cout << "'" << RuntimeConfiguration::instance().getOptimizer() << "'" << endl;
        }else if (input == "print_query_plan"){
            if(RuntimeConfiguration::instance().getPrintQueryPlan()){
            cout << "'true'" << endl;  
            }else{
            cout << "'false'" << endl;
            }
        }else if (input == "enable_profiling"){
            if(RuntimeConfiguration::instance().getProfileQueries()){
            cout << "'true'" << endl;  
            }else{
            cout << "'false'" << endl;
            }
//            cout << "'" << RuntimeConfiguration::instance().getPrintQueryPlan() << "'" << endl;
        }else{
            cout << "Error in print: Invalid variable name: '" << input << "'" << endl;
        }

        return true;
    }

    CommandLineInterpreter::CommandLineInterpreter() : simple_command_map_(), command_map_(), prompt_("CoGaDB>") {

        simple_command_map_.insert(std::make_pair("loaddatabase", &loadTables));
        simple_command_map_.insert(std::make_pair("unittests", &CoGaDB::unit_tests::executeUnitTests));
        simple_command_map_.insert(std::make_pair("printschema", &printDatabaseSchema));
        simple_command_map_.insert(std::make_pair("help", &help));
        simple_command_map_.insert(std::make_pair("about", &about));
        simple_command_map_.insert(std::make_pair("version", &version));        
        simple_command_map_.insert(std::make_pair("simple_ssb_queries", &simple_SSB_Queries));
        simple_command_map_.insert(std::make_pair("showgpucache", &printStatusOfGPUCache));
        simple_command_map_.insert(std::make_pair("toggleQC", &toggleQueryChopping));
        simple_command_map_.insert(std::make_pair("hypestatus", &printHypeStatus));
        simple_command_map_.insert(std::make_pair("integrityconstraints",&printSystemTableIntegrityConstraints));
        simple_command_map_.insert(std::make_pair("joinindexes",&printSystemTableJoinIndexes));
        
        
        //workaround so ssb queries work in batch mode
        simple_command_map_.insert(std::make_pair("ssb11", &SSB_Q11));
        simple_command_map_.insert(std::make_pair("ssb12", &SSB_Q12));
        simple_command_map_.insert(std::make_pair("ssb13", &SSB_Q13));
        simple_command_map_.insert(std::make_pair("ssb21", &SSB_Q21));
        simple_command_map_.insert(std::make_pair("ssb22", &SSB_Q22));
        simple_command_map_.insert(std::make_pair("ssb23", &SSB_Q23));
        simple_command_map_.insert(std::make_pair("ssb31", &SSB_Q31));
        simple_command_map_.insert(std::make_pair("ssb32", &SSB_Q32));
        simple_command_map_.insert(std::make_pair("ssb33", &SSB_Q33));
        simple_command_map_.insert(std::make_pair("ssb34", &SSB_Q34));
        simple_command_map_.insert(std::make_pair("ssb41", &SSB_Q41));
        simple_command_map_.insert(std::make_pair("ssb42", &SSB_Q42));
        simple_command_map_.insert(std::make_pair("ssb43", &SSB_Q43));
        
        //test commands, not an official part of the commandline interface!
        simple_command_map_.insert(std::make_pair("join_performance_test", &CoGaDB::unit_tests::cdk_join_performance_test));
        simple_command_map_.insert(std::make_pair("bitmap_selection_test", &CoGaDB::unit_tests::cdk_selection_bitmap_test));
        simple_command_map_.insert(std::make_pair("selection_performance_test", &CoGaDB::unit_tests::cdk_selection_performance_test));

        simple_command_map_.insert(std::make_pair("invisible_join_test", &CoGaDB::unit_tests::cdk_invisiblejoin_test));

        command_map_.insert(std::make_pair("set", &setVariable));
        command_map_.insert(std::make_pair("create_tpch_database", &Unittest_Create_TPCH_Database));
        command_map_.insert(std::make_pair("create_ssb_database", &Unittest_Create_Star_Schema_Benchmark_Database));
        command_map_.insert(std::make_pair("print", &printVariableValue));
        command_map_.insert(std::make_pair("setdevice", &setDefaultDevice));
        command_map_.insert(std::make_pair("setparallelizationmode", &setDefaultParallelizationModeTwoPhasePhysicalOptimization));

        command_map_.insert(std::make_pair("exec", &SQL::commandlineExec));
        //command_map_.insert(std::make_pair("explain", &SQL::commandlineExplain));
        command_map_.insert(std::make_pair("graphical_explain", &SQL::commandlineExplain));
        command_map_.insert(std::make_pair("explain", &SQL::commandlineExplainStatementsWithOptimization));
        command_map_.insert(std::make_pair("explain_unoptimized", &SQL::commandlineExplainStatementsWithoutOptimization));

        //
        //	create_ssb_database /media/1A2AA9012AA8DAD1_/benchmarks/star_schema_benchmark/SF1/

        //	Unittest_Create_Star_Schema_Benchmark_Database("/media/1A2AA9012AA8DAD1_/benchmarks/star_schema_benchmark/SF1/");

        //Unittest_Create_TPCH_Database
    }

    bool CommandLineInterpreter::parallelExecution(const string& input, ifstream *in_stream) {
        string input_trimmed = boost::algorithm::trim_copy(input.substr(13));
        int threads = boost::lexical_cast<int>(input_trimmed);
        cout << "Used Threads: " << threads << endl;

        string putput;
        while (!(*in_stream).eof()) {
            std::getline(*in_stream, putput);
            if (putput == "quit") {
                break;
            }
            commandQueue.push(putput);
            cout << "Execute Command " << putput << endl;
        }
        boost::thread_group threadgroup;
        for (int i = 0; i < threads; ++i) {
            threadgroup.add_thread(new boost::thread(boost::bind(&CommandLineInterpreter::threadedExecution, this)));

        }
        threadgroup.join_all();
        return 1;
    }

    void CommandLineInterpreter::threadedExecution() {

        string tid = boost::lexical_cast<string>(boost::this_thread::get_id());
        ofstream filestr(("parallel_results_tid_" + tid + ".txt").c_str(), fstream::app);
        std::cout.rdbuf(filestr.rdbuf());

        if (!commandQueue.empty()) {
            string command = commandQueue.front();
            commandQueue.pop();
            cout << "Execute Command " << command << endl;
            execute(command);
            threadedExecution();
        }
        std::cout.flush();
        filestr.close();
    }

    bool CommandLineInterpreter::execute(const string& input) {
        string input_trimmed = boost::algorithm::trim_copy(input);

        if (input_trimmed.empty())
            return true;
        if (input_trimmed[0] == '#')
            return true;

        {
            SimpleCommandMap::iterator it = simple_command_map_.find(input_trimmed);

            if (it != simple_command_map_.end())
                return it->second();
        }

        size_t command_len = input_trimmed.find_first_of(" \t");
        size_t param_begin = input_trimmed.find_first_not_of(" \t", command_len + 1);

        string command = input.substr(0, command_len);
        string parameter = input.substr(param_begin);

        ParameterizedCommandMap::iterator it = command_map_.find(command);
        if (it != command_map_.end())
            return it->second(parameter);

        cout << "Error! Command '" << command << "' not found!" << endl;

        return false;
    }

#ifdef LIBREADLINE_FOUND

    bool CommandLineInterpreter::getline(const string &prompt, string &result) {
        static char *line = NULL;

        if (line)
            free(line);

        line = readline(prompt.c_str());
        if (!line)
            return true; /* EOF */

        if (*line)
            add_history(line);
        result = line;

        return false;
    }
#else /* !LIBREADLINE_FOUND */

    bool CommandLineInterpreter::getline(const string &prompt, string &result) {
        cout << prompt;
        std::getline(cin, result);

        return cin.eof();
    }
#endif /* !LIBREADLINE_FOUND */
} //end namespace CogaDB


