
//#include <persistance/storage_manager.hpp>
#include <persistence/storage_manager.hpp>
#include <core/runtime_configuration.hpp>
#include <util/filesystem.hpp>
#include <util/time_measurement.hpp>

#include <fstream>
#include <iostream>

//serialization
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/binary_object.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/list.hpp>


namespace CoGaDB{

using namespace std;



const TablePtr loadTable(const std::string& table_name){

	string path(RuntimeConfiguration::instance().getPathToDatabase());
	if(!quiet && verbose && debug) cout << "Load table '" << table_name << "' ..." << endl;
	if(!quiet && verbose && debug) cout << "Searching for tables..." << endl;
	std::vector<std::string> v = getFilesinDirectory(path);
	for(unsigned int i=0;i<v.size();i++){
		if(!quiet && verbose && debug) cout << "Found: " << v[i] << endl;
		if(table_name==v[i]){
			path += "/";
			path += table_name;
			break;
		}
		if(i==v.size()-1) {
			cout << "Could not find Table: '" << table_name << "' Aborting..." << endl;
			exit(-1);
		}
	}
	
		Timestamp begin=getTimestamp();
	//load table schema

	   if(!quiet && verbose && debug) cout << "Loading Table Schema '" << table_name << "' from file '" << path << "'..." << endl;
		//string path("data/");
//		path += "/";
//		path += name_;

		TableSchema schema;
		
		//cout << "Opening File '" << path << "'..." << endl;
		ifstream infile (path.c_str(),std::ios_base::binary | std::ios_base::in);
		boost::archive::binary_iarchive ia(infile);
		ia >> schema;
		infile.close();


		TablePtr tab(new Table(table_name,schema)); 

		tab->load();
		
	Timestamp end=getTimestamp();
	assert(end>=begin);
	cout << "Needed " << end-begin << "ns ("<< double(end-begin)/(1000*1000*1000) <<"s) to load Table '" << table_name <<  "' ..." << endl;
	return tab;

}

bool storeTable(const TablePtr tab){

		//store schema
		TableSchema schema = tab->getSchema();
		string table_name = tab->getName();
	
		string path(RuntimeConfiguration::instance().getPathToDatabase());
		path += "/";
		path += table_name;
		cout << "Storing Table schema of Table '" << table_name << "' in File '" << path << "' ..." << endl;
		ofstream outfile (path.c_str(),std::ios_base::binary | std::ios_base::out);
		boost::archive::binary_oarchive oa(outfile);

		oa << schema;

		outfile.flush();
		outfile.close();
		
		//store Table data
		bool error_code = tab->store();
//                if(error_code==true){
//                    addToGlobalTableList();
//                }
                return error_code;
//		if(error_code==false){
//                    return false;
//                }
                
                
		//return true;
}

bool loadTables(){
	vector<TablePtr>& tables=getGlobalTableList();

	string path(RuntimeConfiguration::instance().getPathToDatabase()+"/");

	cout << "Searching for tables..." << endl;
	std::vector<std::string> v = getFilesinDirectory(path);
	
	for(unsigned int i=0;i<v.size();i++){
		if(is_regular_file (path+v[i])){
			cout << "Loading table '" << v[i] << "' ..." << endl;
			const TablePtr tab = loadTable(v[i]);
			tables.push_back(tab);
		}
	}

	cout << "Tables:" << endl;
	for(unsigned int i=0;i<tables.size();i++){
		cout << tables[i]->getName() << endl;
		//tables[i]->print();
	}
	return true;
}

std::vector<TablePtr>& getGlobalTableList(){
	static std::vector<TablePtr> tables;
	return tables;
}

const TablePtr getTablebyName(const std::string& name){

	const vector<TablePtr>& tables=getGlobalTableList();
	vector<TablePtr>::const_iterator it;
	for(it= tables.begin();it!= tables.end();it++){
		if ((*it)->getName()==name){
			return *it;
		}
	}
	cout << "Table '" << name << "' not found..." << endl;
	return TablePtr();

}

bool addToGlobalTableList(TablePtr new_table){
    if(!new_table) return false;
    TablePtr tab = getTablebyName(new_table->getName());
    //if table with the same name exist, do nothing
    if(tab) return false;
    //add to global table list
    getGlobalTableList().push_back(new_table);
    return true;
    
}



}; //end namespace CogaDB


