
#include <query_processing/column_processing/bitmap_operator.hpp>
#include <util/hardware_detector.hpp>

namespace CoGaDB {

    namespace query_processing {

        //Map_Init_Function init_function_Selection_operator=physical_operator::map_init_function_Selection_operator;

        namespace physical_operator {


                CPU_Bitmap_Operator::CPU_Bitmap_Operator(const hype::SchedulingDecision& sched_dec,
                        TypedOperatorPtr left_child,
                        TypedOperatorPtr right_child,
                        BitmapOperation op,
                        MaterializationStatus mat_stat) : BinaryOperator<ColumnPtr, ColumnPtr, ColumnPtr>(sched_dec, left_child, right_child), BitmapOperator(),
                op_(op),
                mat_stat_(mat_stat) {
                }

                bool CPU_Bitmap_Operator::execute() {
                    if (!quiet && debug && verbose) std::cout << "Execute Column Operator CPU" << std::endl;
                    
                    //assert(!this->getInputDataLeftChild()->isMaterialized() && !this->getInputDataLeftChild()->isCompressed() );
//                    //Workaround: We can only work on ColumnPtr and cannot mix PostionListPtr in the plan, so we need a workaround
//                    //create LookupArrays, which store tid lists in the filter outine and get TID lists in the Positionlistroutine
//                    BitmapPtr tids_left_child=getPositonListfromLookupArray(this->getInputDataLeftChild());
//                    BitmapPtr tids_right_child=getPositonListfromLookupArray(this->getInputDataRightChild());
                    //boost::dynamic_pointer_cast<BitmapOperator*>(this->right_child_.get());
                    BitmapOperator* left_bitmap_op = dynamic_cast<BitmapOperator*>(this->left_child_.get());
                    BitmapOperator* right_bitmap_op = dynamic_cast<BitmapOperator*>(this->right_child_.get());
                    assert(left_bitmap_op!=NULL);
                    assert(right_bitmap_op!=NULL);
                    assert(left_bitmap_op->hasResultBitmap() || left_bitmap_op->hasCachedResult_GPU_Bitmap());
                    assert(right_bitmap_op->hasResultBitmap() || right_bitmap_op->hasCachedResult_GPU_Bitmap());
                    
                    BitmapPtr bitmap_left_child;
                    BitmapPtr bitmap_right_child;
                    
                    if(left_bitmap_op->hasResultBitmap()){
                        bitmap_left_child=left_bitmap_op->getResultBitmap();
                    }else if(left_bitmap_op->hasCachedResult_GPU_Bitmap()){
                        bitmap_left_child=gpu::copy_Bitmap_device_to_host(left_bitmap_op->getResult_GPU_Bitmap());
                    }
                    
                    if(right_bitmap_op->hasResultBitmap()){
                        bitmap_right_child=right_bitmap_op->getResultBitmap();
                    }else if(right_bitmap_op->hasCachedResult_GPU_Bitmap()){
                        bitmap_right_child=gpu::copy_Bitmap_device_to_host(right_bitmap_op->getResult_GPU_Bitmap());
                    }
                    
                    assert(bitmap_left_child!=NULL);
                    assert(bitmap_right_child!=NULL);
                    assert(bitmap_left_child->size()==bitmap_right_child->size());
                    
                    if(op_ == BITMAP_AND){
                        this->cpu_bitmap_=CDK::selection::bitwise_and(bitmap_left_child,bitmap_right_child);
                        this->result_size_=CDK::selection::countSetBitsInBitmap(this->cpu_bitmap_->data(),this->cpu_bitmap_->size());
                    }else if(op_ == BITMAP_OR){
                        this->cpu_bitmap_=CDK::selection::bitwise_or(bitmap_left_child,bitmap_right_child);
                        this->result_size_=CDK::selection::countSetBitsInBitmap(this->cpu_bitmap_->data(),this->cpu_bitmap_->size());
                    }else{
                        COGADB_FATAL_ERROR("Invalid BitmapOperation: " << (int) op_,"");
                    }
                    
                    return true;
                    
                }

                CPU_Bitmap_Operator::~CPU_Bitmap_Operator() {
                }

                
                
                GPU_Bitmap_Operator::GPU_Bitmap_Operator(const hype::SchedulingDecision& sched_dec,
                        TypedOperatorPtr left_child,
                        TypedOperatorPtr right_child,
                        BitmapOperation op,
                        MaterializationStatus mat_stat) : BinaryOperator<ColumnPtr, ColumnPtr, ColumnPtr>(sched_dec, left_child, right_child), BitmapOperator(),
                op_(op),
                mat_stat_(mat_stat){
                }

                bool GPU_Bitmap_Operator::execute() {
                    if (!quiet && debug && verbose) std::cout << "Execute Column Operator CPU" << std::endl;
                    
                    //assert(!this->getInputDataLeftChild()->isMaterialized() && !this->getInputDataLeftChild()->isCompressed() );
//                    //Workaround: We can only work on ColumnPtr and cannot mix PostionListPtr in the plan, so we need a workaround
//                    //create LookupArrays, which store tid lists in the filter outine and get TID lists in the Positionlistroutine
//                    BitmapPtr tids_left_child=getPositonListfromLookupArray(this->getInputDataLeftChild());
//                    BitmapPtr tids_right_child=getPositonListfromLookupArray(this->getInputDataRightChild());
                    //boost::dynamic_pointer_cast<BitmapOperator*>(this->right_child_.get());
                    BitmapOperator* left_bitmap_op = dynamic_cast<BitmapOperator*>(this->left_child_.get());
                    BitmapOperator* right_bitmap_op = dynamic_cast<BitmapOperator*>(this->right_child_.get());
                    assert(left_bitmap_op!=NULL);
                    assert(right_bitmap_op!=NULL);
                    assert(left_bitmap_op->hasResultBitmap() || left_bitmap_op->hasCachedResult_GPU_Bitmap());
                    assert(right_bitmap_op->hasResultBitmap() || right_bitmap_op->hasCachedResult_GPU_Bitmap());
                    
                    gpu::GPU_BitmapPtr bitmap_left_child;
                    gpu::GPU_BitmapPtr bitmap_right_child;
                    
                    if(left_bitmap_op->hasResultBitmap() && !left_bitmap_op->hasCachedResult_GPU_Bitmap()){
                        bitmap_left_child=gpu::copy_Bitmap_host_to_device(left_bitmap_op->getResultBitmap());
                    }else if(left_bitmap_op->hasCachedResult_GPU_Bitmap()){
                        bitmap_left_child=left_bitmap_op->getResult_GPU_Bitmap();
                    }
                    
                    if(right_bitmap_op->hasResultBitmap() && !right_bitmap_op->hasCachedResult_GPU_Bitmap()){
                        bitmap_right_child=gpu::copy_Bitmap_host_to_device(right_bitmap_op->getResultBitmap());
                    }else if(right_bitmap_op->hasCachedResult_GPU_Bitmap()){
                        bitmap_right_child=right_bitmap_op->getResult_GPU_Bitmap();
                    }
                    
                    assert(bitmap_left_child!=NULL);
                    assert(bitmap_right_child!=NULL);
                    assert(bitmap_left_child->size()==bitmap_right_child->size());
                    
                    
                    
                    assert(op_ == BITMAP_AND);
                    this->gpu_bitmap_=gpu::GPU_Operators::computeBitmapOperationAnd(bitmap_left_child,bitmap_right_child);
                    this->result_size_= 10; //workaround
                    //this->result_size_=CDK::selection::countSetBitsInBitmap(this->cpu_bitmap_->data(),this->cpu_bitmap_->size());
                    
                    
//#define VALIDATE_GPU_RESULTS_ON_CPU          
#ifdef VALIDATE_GPU_RESULTS_ON_CPU
                    BitmapPtr cpu_bitmap_left_child;
                    BitmapPtr cpu_bitmap_right_child;
                    

                    cpu_bitmap_left_child=gpu::copy_Bitmap_device_to_host(bitmap_left_child);
                    cpu_bitmap_right_child=gpu::copy_Bitmap_device_to_host(bitmap_right_child);
                        
                    if(op_ == BITMAP_AND){
                        this->cpu_bitmap_=CDK::selection::bitwise_and(cpu_bitmap_left_child,cpu_bitmap_right_child);
                        this->result_size_=CDK::selection::countSetBitsInBitmap(this->cpu_bitmap_->data(),this->cpu_bitmap_->size());
                    }else if(op_ == BITMAP_OR){
                        this->cpu_bitmap_=CDK::selection::bitwise_or(cpu_bitmap_left_child,cpu_bitmap_right_child);
                        this->result_size_=CDK::selection::countSetBitsInBitmap(this->cpu_bitmap_->data(),this->cpu_bitmap_->size());
                    }else{
                        COGADB_FATAL_ERROR("Invalid BitmapOperation: " << (int) op_,"");
                    }
                    BitmapPtr result_bitmap_build_on_gpu = gpu::copy_Bitmap_device_to_host(this->gpu_bitmap_);
                    assert(*(this->cpu_bitmap_)==*result_bitmap_build_on_gpu);
#endif
                    
                    return true;
                    
                }

                GPU_Bitmap_Operator::~GPU_Bitmap_Operator(){
                
                }
                
            
            
            column_processing::cpu::TypedOperatorPtr create_CPU_Bitmap_Operator(column_processing::cpu::TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, column_processing::cpu::TypedOperatorPtr left_child, column_processing::cpu::TypedOperatorPtr right_child) {
                logical_operator::Logical_Bitmap_Operator& log_algebra_ref = static_cast<logical_operator::Logical_Bitmap_Operator&> (logical_node);
                if (!quiet && debug && verbose) std::cout << "create CPU_Columnalgebra_Operator!" << std::endl;
                if (!left_child) {
                    std::cout << "[HyPE Library]: Fatal Error! in query_processing::physical_operator::create_CPU_Bitmap_Union_Operator()" << std::endl;
                    std::cout << "Left Child of Node " << logical_node.getOperationName() << " is Invalid!" << std::endl;
                    exit(-1);
                }



                assert(right_child != NULL); //binary operator
                return column_processing::cpu::TypedOperatorPtr(new CPU_Bitmap_Operator(sched_dec,
                        left_child,
                        right_child,
                        log_algebra_ref.getBitmapOperation(),
                        log_algebra_ref.getMaterializationStatus()));
            }

            column_processing::cpu::TypedOperatorPtr create_GPU_Bitmap_Operator(column_processing::cpu::TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, column_processing::cpu::TypedOperatorPtr left_child, column_processing::cpu::TypedOperatorPtr right_child) {
                logical_operator::Logical_Bitmap_Operator& log_algebra_ref = static_cast<logical_operator::Logical_Bitmap_Operator&> (logical_node);
                if (!quiet && debug && verbose) std::cout << "create GPU_Columnalgebra_Operator!" << std::endl;
                if (!left_child) {
                    std::cout << "[HyPE Library]: Fatal Error! in query_processing::physical_operator::create_CPU_Bitmap_Union_Operator()" << std::endl;
                    std::cout << "Left Child of Node " << logical_node.getOperationName() << " is Invalid!" << std::endl;
                    exit(-1);
                }



                assert(right_child != NULL); //binary operator
                return column_processing::cpu::TypedOperatorPtr(new GPU_Bitmap_Operator(sched_dec,
                        left_child,
                        right_child,
                        log_algebra_ref.getBitmapOperation(),
                        log_algebra_ref.getMaterializationStatus()));
            }

            column_processing::cpu::Physical_Operator_Map_Ptr map_init_function_cpu_bitmap_operator() {
                //std::cout << sd.getNameofChoosenAlgorithm() << std::endl;
                column_processing::cpu::Physical_Operator_Map map;
                //std::cout << "calling map init function for CPU_Bitmap_Union_Operator operator!" << std::endl;
                //hype::Scheduler::instance().addAlgorithm("JOIN","CPU_NestedLoopJoin_Algorithm",hype::CPU,"Least Squares 2D","Periodic Recomputation");
                //stemod::Scheduler::instance().addAlgorithm("JOIN","CPU_SortMergeJoin_Algorithm",stemod::CPU,"Least Squares 2D","Periodic Recomputation");
                //hype::Scheduler::instance().addAlgorithm("JOIN","CPU_HashJoin_Algorithm",hype::CPU,"Least Squares 2D","Periodic Recomputation");		


                hype::AlgorithmSpecification col_algebra_alg_spec_cpu("CPU_Bitmap_Operator",
                        "Bitmap_Operator",
                        hype::Multilinear_Fitting_2D,
                        hype::Periodic);

                hype::AlgorithmSpecification col_algebra_alg_spec_gpu("GPU_Bitmap_Operator",
                        "Bitmap_Operator",
                        hype::Multilinear_Fitting_2D,
                        hype::Periodic);

                //addAlgorithmSpecificationToHardware();

                const DeviceSpecifications& dev_specs = HardwareDetector::instance().getDeviceSpecifications();

                for (unsigned int i = 0; i < dev_specs.size(); ++i) {
                    if (dev_specs[i].getDeviceType() == hype::CPU) {
                        hype::Scheduler::instance().addAlgorithm(col_algebra_alg_spec_cpu, dev_specs[i]);				
                    } else if (dev_specs[i].getDeviceType() == hype::GPU) {
#ifdef ENABLE_GPU_ACCELERATION
                        hype::Scheduler::instance().addAlgorithm(col_algebra_alg_spec_gpu, dev_specs[i]);
#endif
                    }
                }

                map["CPU_Bitmap_Operator"] = create_CPU_Bitmap_Operator;
                map["GPU_Bitmap_Operator"] = create_GPU_Bitmap_Operator;
                
                return column_processing::cpu::Physical_Operator_Map_Ptr(new column_processing::cpu::Physical_Operator_Map(map));
            }

        }//end namespace physical_operator

    }//end namespace query_processing

}; //end namespace CogaDB

