
#include <query_processing/column_processing/column_bitmap_fetch_join_operator.hpp>

#include <lookup_table/join_index.hpp>
#include <util/hardware_detector.hpp>

#include "query_processing/operator_extensions.hpp"
#include <hardware_optimizations/primitives.hpp>

namespace CoGaDB
{

	namespace query_processing
	{



		namespace physical_operator
		{

			column_processing::cpu::TypedOperatorPtr create_column_bitmap_fetch_join_operator(column_processing::cpu::TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, column_processing::cpu::TypedOperatorPtr left_child, column_processing::cpu::TypedOperatorPtr)
			{
				logical_operator::Logical_Column_Bitmap_Fetch_Join& log_sort_ref = static_cast<logical_operator::Logical_Column_Bitmap_Fetch_Join&>(logical_node);
				//std::cout << "create SCAN Operator!" << std::endl;
				/*
				if(!left_child) {
					std::cout << "Error!" << std::endl;
					exit(-1);
				}
				assert(right_child==NULL); //unary operator
				*/
//				return column_processing::cpu::TypedOperatorPtr( new column_bitmap_fetch_join_operator(sched_dec, log_sort_ref.getTableName(),log_sort_ref.getColumnName()) );
                                return column_processing::cpu::TypedOperatorPtr( new column_bitmap_fetch_join_operator(sched_dec, left_child, log_sort_ref.getPK_FK_Join_Predicate() ) );
			}



			column_processing::cpu::Physical_Operator_Map_Ptr map_init_function_column_bitmap_fetch_join_operator()
			{
				//std::cout << sd.getNameofChoosenAlgorithm() << std::endl;
				column_processing::cpu::Physical_Operator_Map map;
				if(!quiet) std::cout << "calling map init function! (COLUMN_BITMAP_FETCH_JOIN OPERATION)" << std::endl;
				//hype::Scheduler::instance().addAlgorithm("SCAN","TABLE_SCAN",hype::CPU,"Least Squares 1D","Periodic Recomputation");
                hype::AlgorithmSpecification selection_alg_spec_cpu("CPU_COLUMN_BITMAP_FETCH_JOIN",
                        "COLUMN_BITMAP_FETCH_JOIN",
                        hype::Least_Squares_1D,
                        hype::Periodic);	
	

		//addAlgorithmSpecificationToHardware();
		
		const DeviceSpecifications& dev_specs = HardwareDetector::instance().getDeviceSpecifications();
		
		for(unsigned int i=0;i<dev_specs.size();++i){
			if(dev_specs[i].getDeviceType()==hype::CPU){
				hype::Scheduler::instance().addAlgorithm(selection_alg_spec_cpu,dev_specs[i]);
			}
		}
				
				map["CPU_COLUMN_BITMAP_FETCH_JOIN"]=create_column_bitmap_fetch_join_operator;
				//map["GPU_Algorithm"]=create_GPU_SORT_Operator;
				return column_processing::cpu::Physical_Operator_Map_Ptr(new column_processing::cpu::Physical_Operator_Map(map));
			}


                        
                        column_bitmap_fetch_join_operator::column_bitmap_fetch_join_operator(const hype::SchedulingDecision& sched_dec, TypedOperatorPtr child, PK_FK_Join_Predicate pk_fk_join_pred) : UnaryOperator<ColumnPtr,ColumnPtr>(sched_dec, child), BitmapOperator(),
                                pk_fk_join_pred_(pk_fk_join_pred){
                        }
                        
			bool column_bitmap_fetch_join_operator::execute()
			{	if(!quiet && verbose && debug) std::cout << "Execute Column_Bitmap_Fetch_Join"  << std::endl;

                        
                        //assert(this->getInputData()!=NULL);
                        //this->getInputData()->getName() << std::endl;
                        
                        PositionListOperator* pos_list_op = dynamic_cast<PositionListOperator*>(this->child_.get());
                        assert(pos_list_op!=NULL);
                        assert(pos_list_op->hasResultPositionList() || pos_list_op->hasCachedResult_GPU_PositionList());

                        PositionListPtr input_tids = pos_list_op->getResultPositionList();
                        JoinIndexPtr join_index = JoinIndexes::instance().getJoinIndex(getTablebyName(pk_fk_join_pred_.join_pk_table_name), pk_fk_join_pred_.join_pk_column_name, getTablebyName(pk_fk_join_pred_.join_fk_table_name),pk_fk_join_pred_.join_fk_column_name);
                        this->cpu_bitmap_=createBitmapOfMatchingTIDsFromJoinIndex(join_index,input_tids);
                        assert(getTablebyName(pk_fk_join_pred_.join_fk_table_name)->getNumberofRows()==join_index->second->getPositionList()->size());
                        
                        this->result_size_=CDK::selection::countSetBitsInBitmap(this->cpu_bitmap_->data(),this->cpu_bitmap_->size());
			return true;
			}
			
			column_bitmap_fetch_join_operator::~column_bitmap_fetch_join_operator() {}

		}//end namespace physical_operator


		namespace logical_operator
		{
			Logical_Column_Bitmap_Fetch_Join::Logical_Column_Bitmap_Fetch_Join(const PK_FK_Join_Predicate& pk_fk_join_pred) : TypedNode_Impl<ColumnPtr,physical_operator::map_init_function_column_bitmap_fetch_join_operator>(), pk_fk_join_pred_(pk_fk_join_pred)
			{
			}

			unsigned int Logical_Column_Bitmap_Fetch_Join::getOutputResultSize() const
			{
				return 10;
			}

            double Logical_Column_Bitmap_Fetch_Join::getCalculatedSelectivity() const {
				return 1.0;
			}
			std::string Logical_Column_Bitmap_Fetch_Join::getOperationName() const
			{
				return "COLUMN_BITMAP_FETCH_JOIN";
			}
                        const PK_FK_Join_Predicate Logical_Column_Bitmap_Fetch_Join::getPK_FK_Join_Predicate(){
                                return pk_fk_join_pred_;
                        }

                        std::string Logical_Column_Bitmap_Fetch_Join::toString(bool verbose) const
                        {
                            std::string result="COLUMN_BITMAP_FETCH_JOIN";
                            if(verbose){
                                result+=" ON JOIN_INDEX(" ;
                                result+=this->pk_fk_join_pred_.join_pk_table_name;
                                result+=".";  
                                result+=this->pk_fk_join_pred_.join_pk_column_name;
                                result+=","; 
                                result+=this->pk_fk_join_pred_.join_fk_table_name;
                                result+=".";  
                                result+=this->pk_fk_join_pred_.join_fk_column_name;
                                result+=")";         
                            }
                            return result;
			}
                        
			
		}//end namespace logical_operator

	}//end namespace query_processing

}; //end namespace CogaDB
