
#include <query_processing/invisible_join_operator.hpp>
#include <util/hardware_detector.hpp>
#include <query_processing/query_processor.hpp>

#include <lookup_table/lookup_table.hpp>
#include <lookup_table/join_index.hpp>

namespace CoGaDB
{

namespace query_processing{
//#define INVISIBLE_JOIN_USE_POSITIONLIST_ONLY_PLANS
//Map_Init_Function init_function_InvisibleJoin_operator=physical_operator::map_init_function_InvisibleJoin_operator;

#ifndef INVISIBLE_JOIN_USE_POSITIONLIST_ONLY_PLANS
    const query_processing::column_processing::cpu::LogicalQueryPlanPtr createColumnBasedPlan(TablePtr fact_table, const InvisibleJoinSelectionList& dimensions, hype::DeviceConstraint dev_constr){
        using namespace query_processing::column_processing::cpu;
        using namespace query_processing::column_processing;
        using namespace query_processing::logical_operator;
        using namespace query_processing;
        
        InvisibleJoinSelectionList::const_iterator it;
        const int NUM_DIM_TABLES=dimensions.size();
        
        std::queue<query_processing::column_processing::cpu::TypedLogicalNodePtr> bitmaps_queue;
        for(it=dimensions.begin();it!=dimensions.end();++it){
            query_processing::column_processing::cpu::LogicalQueryPlanPtr complex_selection; 
            //omit selection sub plan for dimension table were no selection is performed
            if(it->knf_sel_expr.disjunctions.empty()) continue;
                
            complex_selection = query_processing::createColumnBasedQueryPlan(getTablebyName(it->table_name),it->knf_sel_expr, dev_constr);
            
            //complex_selections->push_back(complex_selection);
            query_processing::column_processing::cpu::TypedLogicalNodePtr selection_sub_tree = complex_selection->getRoot();
            //avoid that our plan is cleaned up when the plan goes out of scope
            complex_selection->setNewRoot(TypedLogicalNodePtr());
            
            PK_FK_Join_Predicate pk_fk_join_pred(it->table_name, it->join_pred.getColumn1Name(), fact_table->getName(), it->join_pred.getColumn2Name());
            boost::shared_ptr<query_processing::logical_operator::Logical_Column_Fetch_Join> fetch_join (new query_processing::logical_operator::Logical_Column_Fetch_Join(pk_fk_join_pred, dev_constr));
            fetch_join->setLeft(selection_sub_tree);
            
            boost::shared_ptr<Logical_Column_Convert_PositionList_To_Bitmap> tids_to_bitmap(new Logical_Column_Convert_PositionList_To_Bitmap(fact_table->getNumberofRows(), dev_constr)); //hype::CPU_ONLY)); //dev_constr)); 
            tids_to_bitmap->setLeft(fetch_join);
            bitmaps_queue.push(tids_to_bitmap);
        }
        
        
        //create bitmap operations (bushy tree)
                //merge together all tids from this disjunction
                while(bitmaps_queue.size()>1){
                    TypedLogicalNodePtr node1 =bitmaps_queue.front();
                    bitmaps_queue.pop();
                    TypedLogicalNodePtr node2=bitmaps_queue.front();
                    bitmaps_queue.pop();
                    query_processing::column_processing::cpu::TypedLogicalNodePtr bitwise_and(new query_processing::logical_operator::Logical_Bitmap_Operator(BITMAP_AND,LOOKUP, dev_constr)); //hype::CPU_ONLY)); //dev_constr)); 
                    //boost::shared_ptr<logical_operator::Logical_PositionList_Operator>  tid_union(new logical_operator::Logical_PositionList_Operator(POSITIONLIST_UNION));    
                    bitwise_and->setLeft(node1);
                    bitwise_and->setRight(node2);                
                    bitmaps_queue.push(bitwise_and);  
                    if(!quiet && verbose && debug) 
                    std::cout << "ADD BITMAP_AND Operator for Childs " << node1->toString() << " and " << node2->toString() << std::endl;
                }
                     
                boost::shared_ptr<Logical_Column_Convert_Bitmap_To_PositionList> bitmap_to_tids(new Logical_Column_Convert_Bitmap_To_PositionList(dev_constr)); //hype::CPU_ONLY)); //dev_constr)); 
                bitmap_to_tids->setLeft(bitmaps_queue.front());
        
                
                return query_processing::column_processing::cpu::LogicalQueryPlanPtr(new query_processing::column_processing::cpu::LogicalQueryPlan(bitmap_to_tids));
        
    }
    
#else
    const query_processing::column_processing::cpu::LogicalQueryPlanPtr createColumnBasedPlan(TablePtr fact_table, const InvisibleJoinSelectionList& dimensions, hype::DeviceConstraint dev_constr){
        using namespace query_processing::column_processing::cpu;
        using namespace query_processing::column_processing;
        using namespace query_processing::logical_operator;
        using namespace query_processing;
        
        InvisibleJoinSelectionList::const_iterator it;
        const int NUM_DIM_TABLES=dimensions.size();
        
        std::queue<query_processing::column_processing::cpu::TypedLogicalNodePtr> positionlist_queue;
        for(it=dimensions.begin();it!=dimensions.end();++it){
            query_processing::column_processing::cpu::LogicalQueryPlanPtr complex_selection; 
            //omit selection sub plan for dimension table were no selection is performed
            if(it->knf_sel_expr.disjunctions.empty()) continue;
                
            complex_selection = query_processing::createColumnBasedQueryPlan(getTablebyName(it->table_name),it->knf_sel_expr, dev_constr);
            
            //complex_selections->push_back(complex_selection);
            query_processing::column_processing::cpu::TypedLogicalNodePtr selection_sub_tree = complex_selection->getRoot();
            //avoid that our plan is cleaned up when the plan goes out of scope
            complex_selection->setNewRoot(TypedLogicalNodePtr());
            
            PK_FK_Join_Predicate pk_fk_join_pred(it->table_name, it->join_pred.getColumn1Name(), fact_table->getName(), it->join_pred.getColumn2Name());
            boost::shared_ptr<query_processing::logical_operator::Logical_Column_Fetch_Join> fetch_join (new query_processing::logical_operator::Logical_Column_Fetch_Join(pk_fk_join_pred, dev_constr));
            fetch_join->setLeft(selection_sub_tree);
            
            positionlist_queue.push(fetch_join);
        }
        
        
        //create bitmap operations (bushy tree)
                //merge together all tids from this disjunction
                while(positionlist_queue.size()>1){
                    TypedLogicalNodePtr node1 =positionlist_queue.front();
                    positionlist_queue.pop();
                    TypedLogicalNodePtr node2=positionlist_queue.front();
                    positionlist_queue.pop();
                    //query_processing::column_processing::cpu::TypedLogicalNodePtr bitwise_and(new query_processing::logical_operator::Logical_Bitmap_Operator(BITMAP_AND,LOOKUP, hype::CPU_ONLY)); //dev_constr)); 
                    boost::shared_ptr<logical_operator::Logical_PositionList_Operator>  tid_intersection(new logical_operator::Logical_PositionList_Operator(POSITIONLIST_INTERSECTION,dev_constr));     
                    tid_intersection->setLeft(node1);
                    tid_intersection->setRight(node2);                
                    positionlist_queue.push(tid_intersection);  
                    if(!quiet && verbose && debug) 
                    std::cout << "ADD PositionList Intersection Operator for Childs " << node1->toString() << " and " << node2->toString() << std::endl;
                }
                     
        
                
                return query_processing::column_processing::cpu::LogicalQueryPlanPtr(new query_processing::column_processing::cpu::LogicalQueryPlan(positionlist_queue.front()));
        
    }
#endif
    
//    void pk_fk_semi_join_thread(TablePtr  filtered_fact_tab, InvisibleJoinSelection inv_join_sel, LookupTablePtr* result, unsigned int thread_id){
//    
//                TablePtr dim_table = getTablebyName(inv_join_sel.table_name);
//                TablePtr join_result = BaseTable::pk_fk_join(dim_table, inv_join_sel.join_pred.getColumn1Name(), filtered_fact_tab, inv_join_sel.join_pred.getColumn2Name(), HASH_JOIN, LOOKUP);
//                
//                std::cout << "Dim " << inv_join_sel.table_name << ": #Rows: " <<  join_result->getNumberofRows() << std::endl;
//                std::list<std::string> columns_dim_table;
//                TableSchema schema = dim_table->getSchema();
//                TableSchema::iterator sit;
//                for(sit=schema.begin();sit!=schema.end();++sit){
//                    columns_dim_table.push_back(sit->second);
//                }
//                
//                TablePtr dimension_table_semi_join =  BaseTable::projection(join_result, columns_dim_table, MATERIALIZE, CPU);
//                LookupTablePtr lookup_table = boost::dynamic_pointer_cast<LookupTable>(dimension_table_semi_join);
//                //dimension_semi_joins.push_back(lookup_table);
//                *result=lookup_table;
//     }
    
    void semi_join_from_join_index_thread(TablePtr fact_table, PositionListPtr fact_table_tids, InvisibleJoinSelection inv_join_sel, LookupTablePtr* result, unsigned int thread_id){

                TablePtr dim_table = getTablebyName(inv_join_sel.table_name);
                JoinIndexPtr join_index = JoinIndexes::instance().getJoinIndex(fact_table, inv_join_sel.join_pred.getColumn2Name(),  dim_table, inv_join_sel.join_pred.getColumn1Name());
                PositionListPtr matching_dim_table_tids = CoGaDB::fetchMatchingTIDsFromJoinIndex(join_index,fact_table_tids);
                std::cout << "#Matching Rows in DimensionTable " << inv_join_sel.table_name << ": " << matching_dim_table_tids->size() << std::endl;
		LookupTablePtr lookup_table = createLookupTableforUnaryOperation("",dim_table, matching_dim_table_tids);
                *result=lookup_table;
                
    }
    
    TablePtr two_phase_physical_optimization_invisible_join(TablePtr fact_table, InvisibleJoinSelectionList dimensions, hype::DeviceConstraint dev_constr){


        InvisibleJoinSelectionList::iterator it;
        const int NUM_DIM_TABLES=dimensions.size();

        
        Timestamp begin_compute_matching_fact_table_rows = getTimestamp();
        //const query_processing::column_processing::cpu::LogicalQueryPlanPtr log_plan = createColumnBasedPlan(fact_table, dimensions, dev_constr);
        const query_processing::column_processing::cpu::LogicalQueryPlanPtr log_plan = createColumnBasedPlan(fact_table, dimensions, dev_constr);
        log_plan->print();
        query_processing::column_processing::cpu::PhysicalQueryPlanPtr phy_plan;
        //phy_plan = log_plan->convertToPhysicalQueryPlan();
        hype::query_optimization::QEPPtr hype_qep = log_plan->convertToQEP();
        hype::query_optimization::optimizeQueryPlan(*hype_qep,hype::GREEDY_HEURISTIC);//hype::BACKTRACKING);
        phy_plan = log_plan->convertQEPToPhysicalQueryPlan(hype_qep);
        
        phy_plan->print();
        //Timestamp end_optimize_query_plan = getTimestamp();
        
        //std::cout << "Time for query plan generation and optimization: " << double(end_optimize_query_plan-begin_compute_matching_fact_table_rows)/(1000*1000) << "ms" << std::endl;

        Timestamp begin = getTimestamp();    
        phy_plan->run();
        Timestamp end = getTimestamp();    
        std::cout << "QEP Runtime: " << double(end-begin)/(1000*1000) << "ms" << std::endl;
                   
        phy_plan->printResults(true,true,false);
        assert(phy_plan->getRoot().get()!=NULL);
        PositionListOperator* pos_list_op = dynamic_cast<PositionListOperator*>(phy_plan->getRoot().get());
        assert(pos_list_op!=NULL);
        assert(pos_list_op->hasResultPositionList() || pos_list_op->hasCachedResult_GPU_PositionList());

        assert(pos_list_op != NULL);
        assert(pos_list_op->hasResultPositionList() || pos_list_op->hasCachedResult_GPU_PositionList());
        PositionListPtr fact_table_tids;
        if (!pos_list_op->hasResultPositionList() && pos_list_op->hasCachedResult_GPU_PositionList()) {
            fact_table_tids = gpu::copy_PositionList_device_to_host(pos_list_op->getResult_GPU_PositionList());
        } else {
            fact_table_tids = pos_list_op->getResultPositionList();
        }
        assert(fact_table_tids != NULL);
        
//        PositionListPtr fact_table_tids = input_tids->getResultPositionList();
        
        //assert sorted TID list
//        for(unsigned int i=0;i<fact_table_tids->size()-1;++i){
//            assert((*fact_table_tids)[i]<(*fact_table_tids)[i+1]);
//        }
        
        
        LookupTablePtr filtered_fact_tab = createLookupTableforUnaryOperation("InvisibleJoin(Fact_Table)", fact_table, fact_table_tids);
        Timestamp end_compute_matching_fact_table_rows = getTimestamp();

//        filtered_fact_tab->concatenate()
        

        std::cout << "Size of Filtered Fact Table (invisible Join): " << filtered_fact_tab->getNumberofRows() << "rows" << std::endl;

        Timestamp begin_construct_result = getTimestamp();
        TablePtr result=filtered_fact_tab;
        
        for(it=dimensions.begin();it!=dimensions.end();++it){
        //	PositionList dim_tab_tids = fetchValuesFromDimTable(it->first, fact_table, fact_table_tids);
        //	LookupColumn dim_table_lc = createLookupColumn(it->first,dim_tab_tids);
        //	result.concat(dim_table_lc);
                Timestamp begin = getTimestamp();
                result = Table::join(result, it->join_pred.getColumn2Name(), getTablebyName(it->table_name), it->join_pred.getColumn1Name(), HASH_JOIN, LOOKUP);

                Timestamp end = getTimestamp();    
                std::cout << "TOPPO Tuple Reconstruct JOIN with "<< it->table_name <<": " << double(end-begin)/(1000*1000) << "ms" << std::endl;
                //result = Table::join(result, it->join_pred.getColumn2Name(), getTablebyName(it->table_name), it->join_pred.getColumn1Name(), HASH_JOIN, LOOKUP); 
               // result = Table::pk_fk_join(getTablebyName(it->table_name), it->join_pred.getColumn1Name(), result, it->join_pred.getColumn2Name(), HASH_JOIN, LOOKUP); 

                std::cout << "Dim " << it->table_name << ": #Rows: " <<  result->getNumberofRows() << std::endl;
        }
        
        
//        //result->print();
//        std::vector<LookupTablePtr> dimension_semi_joins;
//        for(it=dimensions.begin();it!=dimensions.end();++it){
//                TablePtr dim_table = getTablebyName(it->table_name);
//                TablePtr dim_tab_sel_result = Table::selection(dim_table, it->knf_sel_expr, LOOKUP, SERIAL);
//                //TablePtr join_result = BaseTable::join(dim_tab_sel_result, it->join_pred.getColumn1Name(), filtered_fact_tab, it->join_pred.getColumn2Name(), HASH_JOIN, LOOKUP);
//                TablePtr join_result = BaseTable::join(filtered_fact_tab, it->join_pred.getColumn2Name(),  dim_table, it->join_pred.getColumn1Name(), HASH_JOIN, LOOKUP);
//                
////                TablePtr dim_table = getTablebyName(it->table_name);
////                //TablePtr dim_tab_sel_result = Table::selection(dim_table, it->knf_sel_expr, LOOKUP, SERIAL);
////                TablePtr join_result = BaseTable::pk_fk_join(dim_table, it->join_pred.getColumn1Name(), filtered_fact_tab, it->join_pred.getColumn2Name(), HASH_JOIN, LOOKUP);
//                
//                std::cout << "Dim " << it->table_name << ": #Rows: " <<  join_result->getNumberofRows() << std::endl;
//                std::list<std::string> columns_dim_table;
//                TableSchema schema = dim_table->getSchema();
//                TableSchema::iterator sit;
//                for(sit=schema.begin();sit!=schema.end();++sit){
//                    columns_dim_table.push_back(sit->second);
//                }
//                
//                TablePtr dimension_table_semi_join =  BaseTable::projection(join_result, columns_dim_table, MATERIALIZE, CPU);
//                LookupTablePtr lookup_table = boost::dynamic_pointer_cast<LookupTable>(dimension_table_semi_join);
//                dimension_semi_joins.push_back(lookup_table);
//                lookup_table->print();
//        }
  

//        std::vector<LookupTablePtr> dimension_semi_joins(dimensions.size());
//        unsigned int thread_counter=0; 
//        for(it=dimensions.begin();it!=dimensions.end();++it){
//            CDK::join::pk_fk_semi_join_thread(filtered_fact_tab, *it, &dimension_semi_joins[thread_counter], thread_counter);
//            thread_counter++;
//        }
        
//        boost::thread_group tuple_construction_threads;
//        std::vector<LookupTablePtr> dimension_semi_joins(dimensions.size());
//        unsigned int thread_counter=0; 
//        for(it=dimensions.begin();it!=dimensions.end();++it){
//            tuple_construction_threads.add_thread(new boost::thread(boost::bind(&CDK::join::pk_fk_semi_join_thread,filtered_fact_tab, *it, &dimension_semi_joins[thread_counter], thread_counter)));
//            thread_counter++;
//        }
//        tuple_construction_threads.join_all();
        
//        boost::thread_group threads;
//        std::vector<LookupTablePtr> dimension_semi_joins(dimensions.size());
//        unsigned int thread_counter=0; 
//        for(it=dimensions.begin();it!=dimensions.end();++it){
//            threads.add_thread(new boost::thread(boost::bind(&CDK::join::pk_fk_semi_join_thread,filtered_fact_tab, *it, &dimension_semi_joins[thread_counter], thread_counter)));
//            thread_counter++;
//        }
//        threads.join_all();
        
//        std::vector<LookupTablePtr> dimension_semi_joins;
//        for(it=dimensions.begin();it!=dimensions.end();++it){
//                TablePtr dim_table = getTablebyName(it->table_name);
//                JoinIndexPtr join_index = JoinIndexes::instance().getJoinIndex(fact_table, it->join_pred.getColumn2Name(),  dim_table, it->join_pred.getColumn1Name());
//                PositionListPtr matching_dim_table_tids = CoGaDB::fetchMatchingTIDsFromJoinIndex(join_index,fact_table_tids);
//                std::cout << "#Matching Rows in DimensionTable " << it->table_name << ": " << matching_dim_table_tids->size() << std::endl;
//		LookupTablePtr lookup_table = createLookupTableforUnaryOperation("",dim_table, matching_dim_table_tids);
//                dimension_semi_joins.push_back(lookup_table);
//        }
        
/*
        boost::thread_group tuple_construction_threads;
        std::vector<LookupTablePtr> dimension_semi_joins(dimensions.size());
        unsigned int thread_counter=0; 
        for(it=dimensions.begin();it!=dimensions.end();++it){
            //semi_join_from_join_index_thread(fact_table, fact_table_tids, *it, &dimension_semi_joins[thread_counter], thread_counter);
            tuple_construction_threads.add_thread(new boost::thread(boost::bind(&semi_join_from_join_index_thread,fact_table, fact_table_tids, *it, &dimension_semi_joins[thread_counter], thread_counter)));
            thread_counter++;
       }
       tuple_construction_threads.join_all();
        
        LookupTablePtr lookup_table_result=boost::dynamic_pointer_cast<LookupTable>(filtered_fact_tab);
        for(unsigned int i=0;i<dimension_semi_joins.size();++i){
            lookup_table_result=LookupTable::concatenate("",*lookup_table_result,*dimension_semi_joins[i]);
        };
        assert(lookup_table_result->getNumberofRows()==result->getNumberofRows());
        result=lookup_table_result;
        */
        Timestamp end_construct_result = getTimestamp();
        //result.concat(fact_tab);

        
                std::cout << "TOPPO: Time for Creating Matching Rows TID List: " << double(end_compute_matching_fact_table_rows-begin_compute_matching_fact_table_rows)/(1000*1000) << "ms" << std::endl;
//                cout << "Time for Combining Bitmaps: " << double(end_combine_bitmaps-begin_combine_bitmaps)/(1000*1000) << "ms" << std::endl;
                std::cout << "TOPPO: Time for Result Construction: " << double(end_construct_result-begin_construct_result)/(1000*1000) << "ms" << std::endl;

                
                
        //    char* dim_table_bitmap = bitmap_scan(it->first,it->second);
        //    dim_join_column = fetch_values(it->first, ValueValuePredicate->getColumnName(), dim_table_bitmap);

        return result;

    }
    
	namespace physical_operator{

                bool CPU_InvisibleJoin_Operator::execute()
                {                    
#ifdef ENABLE_TWO_PHASE_PHYSICAL_OPTIMIZATION
//                    Timestamp begin_stable_invisible_join = getTimestamp();
//                    TablePtr tab = CDK::join::invisibleJoin(this->getInputData(), dimensions_);
//                    Timestamp end_stable_invisible_join = getTimestamp();
                    

                    Timestamp begin_toppo_invisible_join = getTimestamp();
                    this->result_ = two_phase_physical_optimization_invisible_join(this->getInputData(), dimensions_, this->dev_constr_); //hype::GPU_ONLY);//this->dev_constr_);
                    //this->result_ = CDK::join::invisibleJoin(this->getInputData(), dimensions_);
                    Timestamp end_toppo_invisible_join = getTimestamp();
//                    std::cout << "Stable Invisible Join: " << double(end_stable_invisible_join-begin_stable_invisible_join)/(1000*1000) << "ms" << std::endl;
//                    std::cout << "TOPPO Invisible Join: " << double(end_toppo_invisible_join-begin_toppo_invisible_join)/(1000*1000) << "ms" << std::endl;
//                    assert(this->result_->getNumberofRows()==tab->getNumberofRows());
#else
                    this->result_ = CDK::join::invisibleJoin(this->getInputData(), dimensions_);
#endif
                if (this->result_) {
                    setResultSize((this->result_)->getNumberofRows());
                    return true;
                } else
                    return false;
                }    
            
	TypedOperatorPtr create_CPU_InvisibleJoin_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child)
	{
		logical_operator::Logical_InvisibleJoin& log_selection_ref = static_cast<logical_operator::Logical_InvisibleJoin&>(logical_node);
		//std::cout << "create CPU_InvisibleJoin_Operator!" << std::endl;
		if(!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
			exit(-1);
		}
                
		assert(right_child==NULL); //unary operator
		return TypedOperatorPtr(new CPU_InvisibleJoin_Operator(sched_dec,
		                        left_child,
		                        log_selection_ref.getInvisibleJoinSelectionList(),
                                        log_selection_ref.getDeviceConstraint(),
		                        log_selection_ref.getMaterializationStatus()) );
	}

	Physical_Operator_Map_Ptr map_init_function_invisible_join_operator()
	{
		//std::cout << sd.getNameofChoosenAlgorithm() << std::endl;
		Physical_Operator_Map map;
		if(!quiet) std::cout << "calling map init function for INVISIBLE JOIN operator!" << std::endl;
		//hype::Scheduler::instance().addAlgorithm("SELECTION","CPU_InvisibleJoin_Algorithm",hype::CPU,"Least Squares 1D","Periodic Recomputation");
		//hype::Scheduler::instance().addAlgorithm("SELECTION","GPU_InvisibleJoin_Algorithm",hype::GPU,"Least Squares 1D","Periodic Recomputation");

		hype::AlgorithmSpecification selection_alg_spec_cpu("CPU_InvisibleJoin_Algorithm",
                                                                    "INVISIBLE_JOIN",
                                                                    hype::Least_Squares_1D,hype::Periodic);
                
		const DeviceSpecifications& dev_specs = HardwareDetector::instance().getDeviceSpecifications();
		
		for(unsigned int i=0;i<dev_specs.size();++i){
			if(dev_specs[i].getDeviceType()==hype::CPU){
				hype::Scheduler::instance().addAlgorithm(selection_alg_spec_cpu,dev_specs[i]);
                                //hype::Scheduler::instance().addAlgorithm(selection_alg_spec_cpu_parallel,dev_specs[i]);
			}else if(dev_specs[i].getDeviceType()==hype::GPU){
				//hype::Scheduler::instance().addAlgorithm(selection_alg_spec_gpu,dev_specs[i]);			
			}
		}


		map["CPU_InvisibleJoin_Algorithm"]=create_CPU_InvisibleJoin_Operator;
                //map["CPU_ParallelInvisibleJoin_Algorithm"]=create_CPU_ParallelInvisibleJoin_Operator;
		//map["GPU_InvisibleJoin_Algorithm"]=create_GPU_InvisibleJoin_Operator;
		return Physical_Operator_Map_Ptr(new Physical_Operator_Map(map));
	}

	}//end namespace physical_operator

}//end namespace query_processing

}; //end namespace CogaDB

