
#include <query_processing/join_operator.hpp>
#include <util/hardware_detector.hpp>

namespace CoGaDB {

    namespace query_processing {

        //Map_Init_Function init_function_Selection_operator=physical_operator::map_init_function_Selection_operator;

        namespace physical_operator {

            TypedOperatorPtr create_CPU_NestedLoopJoin_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child) {
                logical_operator::Logical_Join& log_join_ref = static_cast<logical_operator::Logical_Join&> (logical_node);
                if (!quiet && verbose && debug) std::cout << "create CPU_NestedLoopJoin_Operator!" << std::endl;
                if (!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                    exit(-1);
                }



                assert(right_child != NULL); //binary operator
                return TypedOperatorPtr(new CPU_NestedLoopJoin_Operator(sched_dec,
                        left_child,
                        right_child,
                        log_join_ref.getLeftJoinColumnName(),
                        log_join_ref.getRightJoinColumnName(),
                        log_join_ref.getMaterializationStatus()));
            }

            TypedOperatorPtr create_CPU_SortMergeJoin_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child) {
                logical_operator::Logical_Join& log_join_ref = static_cast<logical_operator::Logical_Join&> (logical_node);
                if (!quiet && verbose && debug) std::cout << "create CPU_SortMergeJoin_Operator!" << std::endl;
                if (!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                    exit(-1);
                }



                assert(right_child != NULL); //binary operator
                return TypedOperatorPtr(new CPU_SortMergeJoin_Operator(sched_dec,
                        left_child,
                        right_child,
                        log_join_ref.getLeftJoinColumnName(),
                        log_join_ref.getRightJoinColumnName(),
                        log_join_ref.getMaterializationStatus()));
            }

            TypedOperatorPtr create_CPU_HashJoin_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child) {
                logical_operator::Logical_Join& log_join_ref = static_cast<logical_operator::Logical_Join&> (logical_node);
                if (!quiet && verbose && debug) std::cout << "create CPU_HashJoin_Operator!" << std::endl;
                if (!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                    exit(-1);
                }



                assert(right_child != NULL); //binary operator
                return TypedOperatorPtr(new CPU_HashJoin_Operator(sched_dec,
                        left_child,
                        right_child,
                        log_join_ref.getLeftJoinColumnName(),
                        log_join_ref.getRightJoinColumnName(),
                        log_join_ref.getMaterializationStatus()));
            }
            
            TypedOperatorPtr create_CPU_Parallel_HashJoin_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child) {
                logical_operator::Logical_Join& log_join_ref = static_cast<logical_operator::Logical_Join&> (logical_node);
                if (!quiet && verbose && debug) std::cout << "create CPU_Parallel_HashJoin_Operator!" << std::endl;
                if (!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                    exit(-1);
                }
                assert(right_child != NULL); //binary operator
                return TypedOperatorPtr(new CPU_Parallel_HashJoin_Operator(sched_dec,
                        left_child,
                        right_child,
                        log_join_ref.getLeftJoinColumnName(),
                        log_join_ref.getRightJoinColumnName(),
                        log_join_ref.getMaterializationStatus()));
            }

            TypedOperatorPtr create_CPU_RadixJoin_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child) {
                logical_operator::Logical_Join& log_join_ref = static_cast<logical_operator::Logical_Join&> (logical_node);
                if (!quiet && verbose && debug) std::cout << "create CPU_RadixJoin_Operator!" << std::endl;
                if (!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                    exit(-1);
                }



                assert(right_child != NULL); //binary operator
                return TypedOperatorPtr(new CPU_RadixJoin_Operator(sched_dec,
                        left_child,
                        right_child,
                        log_join_ref.getLeftJoinColumnName(),
                        log_join_ref.getRightJoinColumnName(),
                        log_join_ref.getMaterializationStatus()));
            }
            
            TypedOperatorPtr create_GPU_Join_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child) {
                logical_operator::Logical_Join& log_join_ref = static_cast<logical_operator::Logical_Join&> (logical_node);
                if (!quiet && verbose && debug) std::cout << "create GPU_Join_Operator!" << std::endl;
                if (!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                    exit(-1);
                }

                assert(right_child != NULL); //binary operator
                return TypedOperatorPtr(new GPU_Join_Operator(sched_dec,
                        left_child,
                        right_child,
                        log_join_ref.getLeftJoinColumnName(),
                        log_join_ref.getRightJoinColumnName(),
                        log_join_ref.getMaterializationStatus()));
            }

            Physical_Operator_Map_Ptr map_init_function_join_operator() {
                //std::cout << sd.getNameofChoosenAlgorithm() << std::endl;
                Physical_Operator_Map map;
                if (!quiet) std::cout << "calling map init function for JOIN operator!" << std::endl;
                //hype::Scheduler::instance().addAlgorithm("JOIN","CPU_NestedLoopJoin_Algorithm",hype::CPU,"Least Squares 2D","Periodic Recomputation");
                //stemod::Scheduler::instance().addAlgorithm("JOIN","CPU_SortMergeJoin_Algorithm",stemod::CPU,"Least Squares 2D","Periodic Recomputation");
                //hype::Scheduler::instance().addAlgorithm("JOIN","CPU_HashJoin_Algorithm",hype::CPU,"Least Squares 2D","Periodic Recomputation");


                hype::AlgorithmSpecification join_alg_spec_cpu_nlj("CPU_NestedLoopJoin_Algorithm",
                        "JOIN",
                        hype::Multilinear_Fitting_2D,
                        hype::Periodic);

                hype::AlgorithmSpecification join_alg_spec_cpu_hashjoin("CPU_HashJoin_Algorithm",
                        "JOIN",
                        hype::Multilinear_Fitting_2D,
                        hype::Periodic);
                
               hype::AlgorithmSpecification join_alg_spec_cpu_parallel_hashjoin("CPU_Parallel_HashJoin_Algorithm",
                        "JOIN",
                        hype::Multilinear_Fitting_2D,
                        hype::Periodic);
               
               hype::AlgorithmSpecification join_alg_spec_cpu_radixjoin("CPU_RadixJoin_Algorithm",
                        "JOIN",
                        hype::Multilinear_Fitting_2D,
                        hype::Periodic);
               
                hype::AlgorithmSpecification join_alg_spec_gpu_sort_merge_join("GPU_SortMergeJoin_Algorithm",
                        "JOIN",
                        hype::Multilinear_Fitting_2D,
                        hype::Periodic);

                //addAlgorithmSpecificationToHardware();

                const DeviceSpecifications& dev_specs = HardwareDetector::instance().getDeviceSpecifications();

                for (unsigned int i = 0; i < dev_specs.size(); ++i) {
                    if (dev_specs[i].getDeviceType() == hype::CPU) {
#ifdef ENABLE_CPU_NESTED_LOOP_JOIN
                        hype::Scheduler::instance().addAlgorithm(join_alg_spec_cpu_nlj,dev_specs[i]);
#endif
                        hype::Scheduler::instance().addAlgorithm(join_alg_spec_cpu_hashjoin, dev_specs[i]);
                        //Radix Join has still some stability issues
                        //hype::Scheduler::instance().addAlgorithm(join_alg_spec_cpu_radixjoin, dev_specs[i]);
                        //hype::Scheduler::instance().addAlgorithm(join_alg_spec_cpu_parallel_hashjoin, dev_specs[i]);
                    } else if (dev_specs[i].getDeviceType() == hype::GPU) {
#ifdef ENABLE_GPU_ACCELERATION
     #ifdef ENABLE_GPU_JOIN                     
                        hype::Scheduler::instance().addAlgorithm(join_alg_spec_gpu_sort_merge_join,dev_specs[i]);
     #endif
#endif
                    }
                }



                //stemod::Scheduler::instance().addAlgorithm("SELECTION","GPU_Selection_Algorithm",stemod::GPU,"Least Squares 1D","Periodic Recomputation");
                map["CPU_NestedLoopJoin_Algorithm"]=create_CPU_NestedLoopJoin_Operator;
                //map["CPU_SortMergeJoin_Algorithm"]=create_CPU_SortMergeJoin_Operator;
		map["GPU_SortMergeJoin_Algorithm"]=create_GPU_Join_Operator;
                map["CPU_HashJoin_Algorithm"] = create_CPU_HashJoin_Operator;
                map["CPU_RadixJoin_Algorithm"] = create_CPU_RadixJoin_Operator;
                map["CPU_Parallel_HashJoin_Algorithm"] = create_CPU_Parallel_HashJoin_Operator;
                //map["GPU_Selection_Algorithm"]=create_GPU_Selection_Operator;
                return Physical_Operator_Map_Ptr(new Physical_Operator_Map(map));
            }

        }//end namespace physical_operator

    }//end namespace query_processing

}; //end namespace CogaDB
