
#include <query_processing/sort_operator.hpp>
#include <util/hardware_detector.hpp>

namespace CoGaDB
{

namespace query_processing{

//Map_Init_Function init_function_sort_operator=physical_operator::map_init_function_sort_operator;

	namespace physical_operator{

	TypedOperatorPtr create_CPU_SORT_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child)
	{
		logical_operator::Logical_Sort& log_sort_ref = static_cast<logical_operator::Logical_Sort&>(logical_node);
		//std::cout << "create CPU_SORT_Operator!" << std::endl;
		if(!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
			exit(-1);
		}
		assert(right_child==NULL); //unary operator
		return TypedOperatorPtr(new CPU_Sort_Operator(sched_dec,
		                        left_child,
		                        log_sort_ref.getColumnNames(),
		                        log_sort_ref.getSortOrder(),
		                        log_sort_ref.getMaterializationStatus()) );
	}

	TypedOperatorPtr create_GPU_SORT_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child)
	{
		logical_operator::Logical_Sort& log_sort_ref = static_cast<logical_operator::Logical_Sort&>(logical_node);
		//std::cout << "create GPU_SORT_Operator!" << std::endl;
		if(!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
			exit(-1);
		}
		assert(right_child==NULL); //unary operator
		return TypedOperatorPtr(new GPU_Sort_Operator(sched_dec,
		                        left_child,
		                        log_sort_ref.getColumnNames(),
		                        log_sort_ref.getSortOrder(),
		                        log_sort_ref.getMaterializationStatus()) );
	}



	Physical_Operator_Map_Ptr map_init_function_sort_operator()
	{
		//std::cout << sd.getNameofChoosenAlgorithm() << std::endl;
		Physical_Operator_Map map;
		if(!quiet) std::cout << "calling map init function for Sort Operator!" << std::endl;
		//hype::Scheduler::instance().addAlgorithm("SORT","CPU_Sort_Algorithm",hype::CPU,"Least Squares 1D","Periodic Recomputation");
		//hype::Scheduler::instance().addAlgorithm("SORT","GPU_Sort_Algorithm",hype::GPU,"Least Squares 1D","Periodic Recomputation");
		
		hype::AlgorithmSpecification sort_alg_spec_cpu("CPU_Sort_Algorithm",
																	"SORT",
																	hype::Least_Squares_1D,
																	hype::Periodic);	
	
		hype::AlgorithmSpecification sort_alg_spec_gpu("GPU_Sort_Algorithm",
																	"SORT",
																	hype::Least_Squares_1D,
																	hype::Periodic);																		
																	

		//addAlgorithmSpecificationToHardware();
		
		const DeviceSpecifications& dev_specs = HardwareDetector::instance().getDeviceSpecifications();
		
		for(unsigned int i=0;i<dev_specs.size();++i){
			if(dev_specs[i].getDeviceType()==hype::CPU){
				hype::Scheduler::instance().addAlgorithm(sort_alg_spec_cpu,dev_specs[i]);
			}else if(dev_specs[i].getDeviceType()==hype::GPU){
#ifdef ENABLE_GPU_ACCELERATION
				hype::Scheduler::instance().addAlgorithm(sort_alg_spec_gpu,dev_specs[i]);	
#endif
			}
		}
		
		
		map["CPU_Sort_Algorithm"]=create_CPU_SORT_Operator;
		map["GPU_Sort_Algorithm"]=create_GPU_SORT_Operator;
		return Physical_Operator_Map_Ptr(new Physical_Operator_Map(map));
	}

	}//end namespace physical_operator

}//end namespace query_processing

}; //end namespace CogaDB
