
#pragma once

#include <core/specification.hpp>
#include <core/scheduling_decision.hpp>
#include <core/time_measurement.hpp>

#include <boost/shared_ptr.hpp>

namespace hype {
    namespace queryprocessing {

        class Operator {
        public:
            bool operator() ();
            const hype::core::SchedulingDecision& getSchedulingDecision() const throw ();
            virtual ~Operator();
        protected:
            Operator(const hype::core::SchedulingDecision& sched_dec); // : sched_dec_(sched_dec) {}
            double timeNeeded;
            double timeEstimated;
        public:
            const core::EstimatedTime getEstimatedExecutionTime() const throw ();

            virtual double getRecursiveEstimationTimeInNanoseconds() = 0;
            virtual double getRecursiveExecutionTimeInNanoseconds() = 0;

            void setTimeNeeded(double timeNeeded);
            void setTimeEstimated(double timeEstimated);
            double getTimeNeeded();
            double getTimeEstimated();
            const std::string getAlgorithmName() const throw ();

            const core::Tuple getFeatureValues() const throw ();

            const core::DeviceSpecification getDeviceSpecification() const throw ();
        private:
            virtual bool execute() = 0;
        protected:
            hype::core::SchedulingDecision sched_dec_;
        };


        typedef boost::shared_ptr<Operator> OperatorPtr;




        /*
        template <typename Type>
        class TypedOperator : public Operator{

                typedef Type OperatorOutputType;


                TypedOperator(const stemod::SchedulingDecision& sched_dec, OperatorOutputType result) : Operator(sched_dec),result_(result){}

                const OperatorOutputType getResult(){
                        this(); //execute Functor
                        return result_;
                }

                protected:
                OperatorOutputType result_;
        };


        template <typename OperatorInputType,typename OperatorOutputType>
        class UnaryOperator : public TypedOperator<OperatorOutputType>{ //uses 1D Learning Method

                UnaryOperator(OperatorInputType);

                OperatorInputType getInputData();

        };

        template <typename OperatorInputType,typename OperatorOutputType>
        class BinaryOperator : public TypedOperator<OperatorOutputType>{ //uses 2D Learning Method

                BinaryOperator(OperatorInputType);

                OperatorInputType getInputDataLeftChild();

                OperatorInputType getInputDataRightChild();

        };


         */



        //selectivity???

    }; //end namespace queryprocessing
}; //end namespace hype
