
#include <core/specification.hpp>
#include <core/scheduler.hpp>
#include <util/print.hpp>
#include <util/get_name.hpp>
#include <util/utility_functions.hpp>
#include <util/algorithm_name_conversion.hpp>
#include <list>
#include <queue>
#include <query_optimization/qep.hpp>
#include <sstream>

namespace hype{
        using namespace hype::core;
	namespace query_optimization{

                        QEP_Node::QEP_Node(const OperatorSpecification& op_spec_arg, const DeviceConstraint& dev_constr_arg)
                        : op_spec(op_spec_arg), dev_constr(dev_constr_arg), parent(NULL), childs(), number_childs(0),
                        alg_ptr(NULL), cost(0), sched_dec_ptr(NULL), user_payload(), c_user_payload(NULL){
                            
                        }
                        
                        
                        QEP_Node::QEP_Node(const QEP_Node& other) : op_spec (other.op_spec), 
                                                           dev_constr(other.dev_constr),
                                                           parent(NULL),
                                                           childs(),
                                                           number_childs(other.number_childs),
                                                           alg_ptr(other.alg_ptr), //we NEVER copy algorithm objects, just pointers
                                                           cost(other.cost),
                                                           sched_dec_ptr(NULL),
                                                           user_payload(other.user_payload),
                                                           c_user_payload(other.c_user_payload)
                        {
                            if(other.sched_dec_ptr){
                                this->sched_dec_ptr=new core::SchedulingDecision(*other.sched_dec_ptr);
                            }
                            
                            for(unsigned int i=0;i<other.number_childs;++i){
                                //call copy constructor
                                if(other.childs[i]){
                                    childs[i]=new QEP_Node(*other.childs[i]);
                                    childs[i]->parent=this;
                                }
                            }

                            
                        }
                        
                        QEP_Node& QEP_Node::operator=(const QEP_Node& other){
                            //check for self assignment
                            if(this!=&other){
                                //cleanup old data
                                cleanupChilds();
                                if(this->sched_dec_ptr) delete this->sched_dec_ptr;
                                //assign new data
                                this->alg_ptr=other.alg_ptr;
                                this->cost=other.cost;
                                this->dev_constr=other.dev_constr;
                                this->op_spec=other.op_spec;
                                this->number_childs=other.number_childs;
                                this->parent=other.parent;
                                this->sched_dec_ptr=other.sched_dec_ptr;
                                this->user_payload=other.user_payload;
                                this->c_user_payload=other.c_user_payload;
                                        
                                for(unsigned int i=0;i<other.number_childs;++i){
                                    //call copy constructor
                                    childs[i]=new QEP_Node(*other.childs[i]);
                                    childs[i]->parent=this;
                                }
                                
                            }
                        }
                        
                        void QEP_Node::assignAlgorithm(core::Algorithm* alg_ptr){
                            this->alg_ptr=alg_ptr;
                            //std::cout << "AssignAlg: " << alg_ptr->toString() << std::endl;
                            this->cost=alg_ptr->getEstimatedExecutionTime(op_spec.getFeatureVector()).getTimeinNanoseconds();
                        }
                        
                        void QEP_Node::setChilds(QEP_Node** childs, size_t number_childs){
                            assert(number_childs<=MAX_CHILDS);
                            cleanupChilds();
                            for(unsigned int i=0;i<number_childs;++i){
                                this->childs[i]=childs[i];
                                this->childs[i]->parent=this;
                            }
                            this->number_childs=number_childs;
                        }
                        
                        void QEP_Node::setParent(QEP_Node* parent){
                            this->parent=parent;
                        }
                        
                        bool QEP_Node::isLeave(){
                            if(number_childs==0){
                                return true;
                            }else{
                                return false;
                            }
                        }
                        
                        bool QEP_Node::isRoot(){
                            if(this->parent==NULL){
                                return true;
                            }else{
                                return false;
                            }
                        }
                        
                        size_t QEP_Node::getTotalNumberofChilds(){
                            if(number_childs==0){
                                return 0;
                            }else{
                                size_t num_childs=number_childs;
                                for(unsigned int i=0;i<number_childs;++i){
                                    num_childs+=childs[i]->getTotalNumberofChilds();
                                }
                                return num_childs;
                            }
                        }
                        
                        double QEP_Node::computeCost(){
                            if(number_childs==0){
                                return this->cost;
                            }else{
                                double total_cost=this->cost;
                                double max_cost_of_childs=std::numeric_limits<double>::min(); 
                                for(unsigned int i=0;i<number_childs;++i){
                                    max_cost_of_childs=std::max(max_cost_of_childs, childs[i]->computeCost());
                                }
                                return  max_cost_of_childs+total_cost;
                            }
                        }
                        
                        std::list<QEP_Node*> QEP_Node::preorder_traversal(){
                            
                            std::list<QEP_Node*> ret;
                            for(unsigned int i=0;i<this->number_childs;++i){
                                if(this->childs[i]){
                                    std::list<QEP_Node*> child_order = this->childs[i]->preorder_traversal();
                                    ret.insert(ret.end(),child_order.begin(),child_order.end());
                                }
                            }
                            return ret;
                        }
                        
                        std::list<QEP_Node*> QEP_Node::levelorder_traversal(){
                        
                            std::list<QEP_Node*> result;
                            
                            std::queue<QEP_Node*> queue;
                            
                            queue.push(this);
                            //result.push_back(this);
                            while(!queue.empty()){
                                QEP_Node* node = queue.front();
                                result.push_back(node);
                                for(unsigned int i=0;i<node->number_childs;++i){
                                    if(node->childs[i])
                                        queue.push(node->childs[i]);
                                }
                                //remove processed node from queue
                                queue.pop();
                            }
                            return result;
                        }
                        
                        std::list<QEP_Node*> QEP_Node::reverselevelorder_traversal(){
                            std::list<QEP_Node*> result=levelorder_traversal();
                            result.reverse();
                            return result;
                        }
                        
                        std::list<QEP_Node*> QEP_Node::postorder_traversal(){
                            return std::list<QEP_Node*>();
                        }
                        
                        
                        QEP_Node::~QEP_Node(){
                            //node that we do not delete algorithm pointers on purpose!
                            //algorithms are explicitely managed by their operations, which are in turn managed by the central scheduler instance!

                            if(sched_dec_ptr) delete sched_dec_ptr;
                            //However, we clean up our childs recursively:
                            for(unsigned int i=0;i<number_childs;++i){
                                if(childs[i]) delete childs[i];
                            }
                        }
                        
                        std::string QEP_Node::toString(){
                            std::stringstream ss;
                            if(this->sched_dec_ptr){
                                ss << alg_ptr->getName() << "(" <<  op_spec.getFeatureVector() << "), Cost: " << this->cost/(1000*1000) << "ms"; 
                                ss << " at device: " << (int)sched_dec_ptr->getDeviceSpecification().getProcessingDeviceID()
                                   << " (" << util::getName(sched_dec_ptr->getDeviceSpecification().getDeviceType()) << ", "
                                   << "MemoryID: " << (int)sched_dec_ptr->getDeviceSpecification().getMemoryID() << ")";
                                if(this->c_user_payload){
                                    ss << " User Payload (C Interface): " << this->c_user_payload; // << std::endl;
                                }
                            }else if(alg_ptr){
                                //ss << alg_ptr->getName() << ": " << this->cost;
                                ss << alg_ptr->getName() << "(" <<  op_spec.getFeatureVector() << "): " << this->cost; 
                            }else{
                                //ss << op_spec.getOperatorName();
                                ss << op_spec.getOperatorName() << "(" <<  op_spec.getFeatureVector() << ")";
                            }
                            return ss.str();
                        }
                        
                        void QEP_Node::cleanupChilds(){
                            for(unsigned int i=0;i<number_childs;++i){
                                if( childs[i]) delete childs[i];
                            }
                        }
                        
   /* ============================================================================ */
                        
                    QEP::QEP() : root_(NULL){
                        
                    }   
                    
                    QEP::QEP(QEP_Node* root) : root_(root){
                        
                    }
                    
                    QEP::QEP(const QEP& other) : root_(new QEP_Node(*other.root_)){
                        
                        
                    }
                    
                    QEP& QEP::operator=(const QEP& other){
                        //check for self assignment
                        if(this!=&other){
                            //cleanup root node, which will recursively cleanup childs
                            if(root_){
                                delete root_;
                            }
                            this->root_=new QEP_Node(*other.root_);
                        }
                    }
                    
                    
                    double QEP::computeCost(){
                        if(root_)
                            return root_->computeCost();
                        else
                            return false;
                    }
                    
                    struct Print_QEP_Functor{
                    
                        Print_QEP_Functor(std::string& str) : result_string(str){
                        }
                        
                        Print_QEP_Functor(Print_QEP_Functor& other) : result_string(other.result_string){
                            
                        }
                        Print_QEP_Functor& operator=(Print_QEP_Functor& other){
                            if(this!=&other){
                            result_string=other.result_string;
                            }
                        }
                        
                        
                        void operator()(QEP_Node* node, unsigned int level){
                            std::stringstream ss;
                            if(!result_string.empty()) ss << std::endl;
                            for(unsigned int i=0;i<level;++i){
                                ss << "\t";
                            }
                            ss << node->toString(); //<< std::endl;
                            //std::cout << ss.str();
                            result_string+=ss.str();
                            //std::cout << "result_string: " <<  result_string << std::endl;
                            //std::cout << "Current Line: " <<  ss.str() << std::endl;
                        }

                        std::string& result_string;
                    };
                    
                    
                    std::string QEP::toString(unsigned int indent){
                        if(root_==NULL){
                            return std::string();
                        }
                        std::string result;
                        Print_QEP_Functor functor(result);
                        root_->traverse_preorder(functor, 0);
                        
                       return result; //functor.result_string;
                        
                        
                        std::list<QEP_Node*> nodes = root_->preorder_traversal();
                        
                        
//                        std::stringstream ss;
//                        
//                        
//                        std::list<QEP_Node*> nodes = root_->preorder_traversal();
//                        
//                        
//                        
//                        for(unsigned int i=0;i<indent;++i){
//                            ss << "\t";
//                        }
//                        ss << root_->toString() << std::endl;
//
//                        QEP_Node** childs=root_->childs;
//                        size_t number_childs=root_->number_childs;
//
//                        for(unsigned int i=0;i<number_childs;++i){
////                              if(childs[i]){
////                                  QEP* tmp = new QEP(childs[i]);
////                                  tmp->toString(indent+1);
////                                  delete tmp;
////                              }
//                            if(childs[i]){
//                               ss << QEP(new QEP_Node(*childs[i])).toString(indent+1);
//                            }   
//                        }
//
//                        
//                        return ss.str();
                    }
                    
                    std::list<QEP_Node*> QEP::levelorder_traversal(){
                        if(root_==NULL){
                            return std::list<QEP_Node*>();
                        }
                        return root_->levelorder_traversal();
                    }

                    std::list<QEP_Node*> QEP::reverselevelorder_traversal(){
                        if(root_==NULL){
                            return std::list<QEP_Node*>();
                        }
                        return root_->reverselevelorder_traversal();
                    }
                    
                    size_t QEP::getNumberOfOperators(){
                        if(root_==NULL){
                            return 0;
                        }else{
                            root_->getTotalNumberofChilds();
                        }
                    }
                    
                    QEP_Node* QEP::getRoot(){
                        return this->root_;
                    }
                    
                    QEP_Node* QEP::removeRoot(){
                        QEP_Node* ret_val=this->root_;
                        this->root_=NULL;
                        return ret_val;
                    }
                    
                    void QEP::setRoot(QEP_Node* new_root){
                        this->root_=new_root;
                    }
                    
                    QEP::~QEP(){
                        if(root_)
                                delete root_;
                    }
 
   /* ============================================================================ */
                    
                    
                    
                void advanceAlgorithmIDs(std::vector<int>& current_algorithm_ids, const std::vector<int>& number_of_algorithms, int begin_index = 0){
                    assert(current_algorithm_ids.size()==number_of_algorithms.size());
                    if(begin_index==current_algorithm_ids.size())
                        return;
                    if(begin_index<current_algorithm_ids.size()){
                        if(current_algorithm_ids[begin_index] < number_of_algorithms[begin_index]-1){
                           current_algorithm_ids[begin_index]++; 
                        }else{
                            current_algorithm_ids[begin_index]=0;
                            advanceAlgorithmIDs(current_algorithm_ids, number_of_algorithms, begin_index+1);        
                        }
                    }   
                }
                
                
                void optimizeQueryPlanGreedy(QEP& plan){
                            QEP current_plan(plan);
                            if(!hype::quiet && hype::verbose && hype::debug)
                                std::cout << "Create Plan: " << std::endl;
                            //create plan
                            //for(unsigned int j=0;j<op_seq.size();++j){

                                std::list<hype::query_optimization::QEP_Node*> order = current_plan.levelorder_traversal(); //current_plan.reverselevelorder_traversal(); //plan.levelorder_traversal();
                                std::list<hype::query_optimization::QEP_Node*>::iterator it;
                                unsigned int operator_index=0;
                                for(it=order.begin();it!=order.end();++it,++operator_index){
                                    if(!hype::quiet && hype::verbose && hype::debug)
                                        std::cout << (*it)->op_spec.getOperatorName() <<  " (" << (*it)->op_spec.getFeatureVector() << ")" << std::endl;
                                    //fetch algorithm
                                    
                                    SchedulingDecision sched_dec = core::Scheduler::instance().getOptimalAlgorithm((*it)->op_spec, (*it)->dev_constr);
                                    (*it)->sched_dec_ptr=new core::SchedulingDecision(sched_dec);
                                    core::AlgorithmPtr alg_ptr = core::Scheduler::instance().getAlgorithm(hype::util::toInternalAlgName(sched_dec.getNameofChoosenAlgorithm(), sched_dec.getDeviceSpecification()));
                                    assert(alg_ptr!=NULL);
                                    //decision is load aware, if WTAR is used. 
                                    //However, the single cost of this algorithm 
                                    //does not display the load, only this algorithm's 
                                    //execution time
                                    (*it)->assignAlgorithm(alg_ptr.get());
                                    //hype::core::AlgorithmPtr alg_ptr = available_algorithms_per_operator[operator_index][current_algorithm_ids[operator_index]];
                                    //(*it)->assignAlgorithm(alg_ptr.get());
                                    //at leave level
                                    if((*it)->isLeave()){
                                        if(!(*it)->isRoot()){
                                            insertCopyOperatorInnerNode(*it);
                                        }
                                        insertCopyOperatorLeaveNode(*it);
                                    //at root    
                                    }else if((*it)->isRoot()){
                                        insertCopyOperatorRootNode(current_plan,*it);
                                    //inside plan    
                                    }else{
                                        insertCopyOperatorInnerNode(*it);
                                        if((*it)->parent->op_spec.getMemoryLocation()!=(*it)->op_spec.getMemoryLocation()){

                                        }
                                    }
                                 }
                                
                        //construct scheduling decisions for copy operations
                       order = current_plan.levelorder_traversal(); //current_plan.reverselevelorder_traversal(); //plan.levelorder_traversal();
                        for(it=order.begin();it!=order.end();++it){
                            if((*it)->sched_dec_ptr==NULL){
                                if((*it)->alg_ptr){
                                    (*it)->sched_dec_ptr=new core::SchedulingDecision(*((*it)->alg_ptr), core::EstimatedTime((*it)->cost), (*it)->op_spec.getFeatureVector()); 
                                }else{
                                    HYPE_FATAL_ERROR("In Greedy Strategy: Found Node without assigned algorithm!",std::cout);
                                }
                            }
                            
                        }
                                
                                
                        //assign result plan to input reference
                        plan=current_plan;
                }
                
                
                void optimizeQueryPlanBacktracking(QEP& plan){
                    
                    //size_t num_op = plan.getNumberOfOperators();
                    
                    
                    QEP optimal_plan;
                    double total_cost_optimal_plan=std::numeric_limits<double>::max();
                    
                    //QEP current_plan(plan);
                    
                    hype::core::OperatorSequence op_seq;
                    {
                        std::list<hype::query_optimization::QEP_Node*> order = plan.levelorder_traversal(); //plan.levelorder_traversal();
                        std::list<hype::query_optimization::QEP_Node*>::iterator it;
                        for(it=order.begin();it!=order.end();++it){
                            std::cout << (*it)->op_spec.getOperatorName() << std::endl;
                            op_seq.push_back(std::make_pair((*it)->op_spec,(*it)->dev_constr));
                        }
                    }
                    if(!hype::quiet && hype::verbose && hype::debug){
                        std::cout << "Serialized Plan: " << std::endl;
                        util::print(op_seq,std::cout);
                    }
                    
 //stores for each operators the number of possible algorithms to choose from
                        std::vector<int> number_of_algorithms(op_seq.size());
                        std::vector<std::vector<hype::core::AlgorithmPtr> > available_algorithms_per_operator(op_seq.size());
                        for(unsigned int i=0;i<op_seq.size();++i){
                            //core::Scheduler::instance().get
                            const core::Scheduler::MapNameToOperation& map_operationname_to_operation_=hype::core::Scheduler::instance().getOperatorMap();
                            core::Scheduler::MapNameToOperation::const_iterator it = map_operationname_to_operation_.find(op_seq[i].first.getOperatorName());
                            if(it==map_operationname_to_operation_.end()){
                                //error, return NUll Pointer
                                HYPE_FATAL_ERROR("INVALID OPERATIONNAME!",std::cout);
                                return;
                            }
                            //fill the number of algorithms array
                            const std::vector<hype::core::AlgorithmPtr> alg_ptrs = it->second->getAlgorithms();
                            for(unsigned int j=0;j<alg_ptrs.size();++j){
                                if(op_seq[i].second.getDeviceTypeConstraint()==hype::ANY_DEVICE 
                                  || (op_seq[i].second.getDeviceTypeConstraint()==hype::CPU_ONLY && alg_ptrs[j]->getDeviceSpecification().getDeviceType()==hype::CPU) 
                                  || (op_seq[i].second.getDeviceTypeConstraint()==hype::GPU_ONLY && alg_ptrs[j]->getDeviceSpecification().getDeviceType()==hype::GPU)){
                                        number_of_algorithms[i]++;
                                        available_algorithms_per_operator[i].push_back(alg_ptrs[j]);
                                }
                            }
                        }
                        
                        //compute number of possible plans
                        unsigned int number_of_plans=1;
                        for(unsigned int i=0;i<op_seq.size();++i){
                            number_of_plans*= number_of_algorithms[i];
                        }
                        
                        std::vector<int> current_algorithm_ids (number_of_algorithms.size());


                        //double total_cost_optimal_plan=std::numeric_limits<double>::max();
                        
                        std::cout << "Backtracking exploring " << number_of_plans << " plans" << std::endl;
                        for(unsigned int i=0;i<number_of_plans;++i){
                            
                            QEP current_plan(plan);
                            if(!hype::quiet && hype::verbose && hype::debug)
                                std::cout << "Create Plan: " << std::endl;
                            //create plan
                            //for(unsigned int j=0;j<op_seq.size();++j){

                                std::list<hype::query_optimization::QEP_Node*> order = current_plan.levelorder_traversal(); //current_plan.reverselevelorder_traversal(); //plan.levelorder_traversal();
                                std::list<hype::query_optimization::QEP_Node*>::iterator it;
                                unsigned int operator_index=0;
                                for(it=order.begin();it!=order.end();++it,++operator_index){
                                    if(!hype::quiet && hype::verbose && hype::debug)
                                        std::cout << (*it)->op_spec.getOperatorName() <<  " (" << (*it)->op_spec.getFeatureVector() << ")" << std::endl;
                                    //fetch algorithm
                                    hype::core::AlgorithmPtr alg_ptr = available_algorithms_per_operator[operator_index][current_algorithm_ids[operator_index]];
                                    (*it)->assignAlgorithm(alg_ptr.get());
                                    //at leave level
                                    if((*it)->isLeave()){
                                        if(!(*it)->isRoot()){
                                            insertCopyOperatorInnerNode(*it);
                                        }
                                        insertCopyOperatorLeaveNode(*it);
                                    //at root    
                                    }else if((*it)->isRoot()){
                                        insertCopyOperatorRootNode(current_plan,*it);
                                    //inside plan    
                                    }else{
                                        insertCopyOperatorInnerNode(*it);
                                        if((*it)->parent->op_spec.getMemoryLocation()!=(*it)->op_spec.getMemoryLocation()){

                                        }

                                    }

                                 }
                             //}
                        
                            double cost_current_plan = current_plan.computeCost();
                            if(!hype::quiet && hype::verbose && hype::debug){
                                std::cout << current_plan.toString() << std::endl;
                                std::cout << "Total Cost: " << cost_current_plan/(1100*1000) << "ms" << std::endl;
                            }
                            
                            if(cost_current_plan<total_cost_optimal_plan){
                                total_cost_optimal_plan=cost_current_plan;
                                optimal_plan=current_plan;
                                if(!hype::quiet && hype::verbose && hype::debug)
                                    std::cout << "================ New Optimal Plan ===============" << std::endl;
                            }
                            
                            //advance
                            advanceAlgorithmIDs(current_algorithm_ids, number_of_algorithms);
                             
                    }
                    
//                        std::cout << "Optimal Plan: " << std::endl;
//                        std::cout << optimal_plan.toString() << std::endl;
                    //current_plan->getRoot()

                        //construct scheduling decisions
                        std::list<hype::query_optimization::QEP_Node*> order = optimal_plan.levelorder_traversal(); //current_plan.reverselevelorder_traversal(); //plan.levelorder_traversal();
                        std::list<hype::query_optimization::QEP_Node*>::iterator it;
                        unsigned int operator_index=0;
                        for(it=order.begin();it!=order.end();++it,++operator_index){
                            (*it)->sched_dec_ptr= new core::SchedulingDecision(*((*it)->alg_ptr), core::EstimatedTime((*it)->cost), (*it)->op_spec.getFeatureVector());
                        }
                        //assign result plan to input reference
                        plan=optimal_plan;
                }
                
                bool verifyPlan(QEP& plan){
                    //check the plan whether all QEP_Nodes have a valid Scheduling Decision and that no Device Constraint is violated
                    std::list<hype::query_optimization::QEP_Node*> order = plan.levelorder_traversal(); //current_plan.reverselevelorder_traversal(); //plan.levelorder_traversal();
                    std::list<hype::query_optimization::QEP_Node*>::iterator it;
                    unsigned int operator_index=0;
                    for(it=order.begin();it!=order.end();++it,++operator_index){
                        //valid scheduling decisions?
                        if((*it)->sched_dec_ptr==NULL){
                            std::cerr << "Error! Node '" << (*it)->toString() << "' has no valid Scheduling Decision!" << std::endl;
                            return false;
                        }
                        //device constraints considered?
                        if(!util::satisfiesDeviceConstraint((*it)->sched_dec_ptr->getDeviceSpecification(), (*it)->dev_constr)){
                            std::cerr << "Error! Node '" << (*it)->toString() << "' does not satisfy its DeviceConstraint!" << std::endl;
                            return false; 
                        }

                    }
                    return true;
                }
                
                void optimizeQueryPlan(QEP& plan, const QueryOptimizationHeuristic& heuristic){
                    uint64_t begin = getTimestamp();
                    if(heuristic==BACKTRACKING){
                        optimizeQueryPlanBacktracking(plan);
                    }else if(heuristic==GREEDY_HEURISTIC){
                        optimizeQueryPlanGreedy(plan);
                    }
                    //check whether for each operator a Scheduling Decision was created
                    if(!verifyPlan(plan)){
                        HYPE_FATAL_ERROR("QEP contains errors!",std::cout);
                    }
                    uint64_t end = getTimestamp();
                    std::cout << "Result Plan (" << util::getName(heuristic) << "): " << std::endl << plan.toString() << std::endl;
                    std::cout << "Time for optimization using heuristic '" << util::getName(heuristic) << "': " <<  double(end-begin)/(1000*1000) << "ms" << std::endl;
                }
                
//               DeviceSpecification& Scheduler::getDeviceSpecificationforCopyType(const std::string& copy_type){
//                   if(copy_type=="COPY_CPU_CP"){
//                       return this->dma_cpu_cp_;
//                   }else if(copy_type=="COPY_CP_CPU"){
//                       return this->dma_cp_cpu_;
//                   }else if(copy_type=="COPY_CP_CP"){
//                       return this->dma_cp_cpu_;
//                   }else{
//                        HYPE_FATAL_ERROR(std::string("INVALID COPY OPERATION TYPE: ")+copy_type,std::cout);
//                        return this->dma_cp_cpu_;
//                   }
//               }
               

                
                        void insertCopyOperatorLeaveNode(hype::query_optimization::QEP_Node* leave){
                            assert(leave!=NULL);
                            assert(leave->alg_ptr!=NULL);
                            assert(leave->isLeave());
                            if(leave->alg_ptr->getDeviceSpecification().getMemoryID()!=hype::PD_Memory_0){
                                std::string copy_type = util::getCopyOperationType(hype::PD0,leave->alg_ptr->getDeviceSpecification().getProcessingDeviceID());
                                assert(!copy_type.empty());
                                DeviceSpecification dev_spec = hype::core::Scheduler::instance().getDeviceSpecificationforCopyType(copy_type);
                                core::AlgorithmPtr copy_alg_ptr = hype::core::Scheduler::instance().getAlgorithm(util::toInternalAlgName(copy_type,dev_spec ));
                                assert(copy_alg_ptr!=NULL);
                                //cout << "COPY Operator " << copy_alg_ptr.get() << endl;
                                Tuple t;
                                t.push_back(leave->op_spec.getFeatureVector().front());
                                //DeviceSpecification dev_spec = getDeviceSpecificationforCopyType(copy_type);
                                //coyp data from CPU to co-processor:
                                OperatorSpecification copy_op_spec(copy_type,
                                                                    t,
                                                                    hype::PD_Memory_0, //memory id of input data
                                                                    leave->alg_ptr->getDeviceSpecification().getMemoryID()); //memory id of result data
                                
                                hype::query_optimization::QEP_Node* copy_node = new hype::query_optimization::QEP_Node(copy_op_spec, hype::core::DeviceConstraint());
                                copy_node->assignAlgorithm(copy_alg_ptr.get());
                                //SchedulingDecision copy_operator(*copy_alg_ptr,copy_alg_ptr->getEstimatedExecutionTime(t),t);
                                //result_plan->push_back(copy_operator); 
//                                hype::query_optimization::QEP_Node* old_parent = leave->parent;
//                                hype::query_optimization::QEP_Node* old_child = leave;
                                leave->setChilds(&copy_node,1);

                            }
                        }
                        void insertCopyOperatorRootNode(hype::query_optimization::QEP& qep, hype::query_optimization::QEP_Node* root){
      
                            if(root->alg_ptr->getDeviceSpecification().getMemoryID()!=hype::PD_Memory_0){
                                //cout << "Insert COPY Operation to transfer result to CPU" << endl;
                                std::string copy_type = util::getCopyOperationType(root->alg_ptr->getDeviceSpecification().getProcessingDeviceID(),hype::PD0); //assume that PD0 is CPU!
                                assert(!copy_type.empty());
                                DeviceSpecification dev_spec = hype::core::Scheduler::instance().getDeviceSpecificationforCopyType(copy_type);
                                core::AlgorithmPtr copy_alg_ptr = hype::core::Scheduler::instance().getAlgorithm(util::toInternalAlgName(copy_type,dev_spec ));
                                assert(copy_alg_ptr!=NULL);
                                //cout << "COPY Operator " << copy_alg_ptr.get() << endl;
                                Tuple t;
                                t.push_back(root->op_spec.getFeatureVector().front());
                                
                                //copy data from CPU to co-processor:
                                OperatorSpecification copy_op_spec(copy_type,
                                                                    t,
                                                                    root->alg_ptr->getDeviceSpecification().getMemoryID(), //memory id of input data
                                                                    hype::PD_Memory_0); //memory id of result data
                                
                                hype::query_optimization::QEP_Node* copy_node = new hype::query_optimization::QEP_Node(copy_op_spec, hype::core::DeviceConstraint());
                                copy_node->assignAlgorithm(copy_alg_ptr.get());
                                
                                copy_node->setChilds(&root,1);
                                qep.setRoot(copy_node);
                                //InternalPhysicalOperator copy_operator(copy_alg_ptr, t, copy_alg_ptr->getEstimatedExecutionTime(t).getTimeinNanoseconds());
                                //SchedulingDecision copy_operator(*copy_alg_ptr,copy_alg_ptr->getEstimatedExecutionTime(t),t);
                            }
                            
                            
//                            if(last_operator.alg_ptr->getDeviceSpecification().getMemoryID()!=hype::PD_Memory_0){
//                                //cout << "Insert COPY Operation to transfer result to CPU" << endl;
//                                std::string copy_type = util::getCopyOperationType(last_operator.alg_ptr->getDeviceSpecification().getProcessingDeviceID(),hype::PD0); //assume that PD0 is CPU!
//                                assert(!copy_type.empty());
//                                DeviceSpecification dev_spec = getDeviceSpecificationforCopyType(copy_type);
//                                AlgorithmPtr copy_alg_ptr = this->getAlgorithm(util::toInternalAlgName(copy_type,dev_spec ));
//                                assert(copy_alg_ptr!=NULL);
//                                //cout << "COPY Operator " << copy_alg_ptr.get() << endl;
//                                Tuple t;
//                                t.push_back(last_operator.feature_vector.front());
//                                InternalPhysicalOperator copy_operator(copy_alg_ptr, t, copy_alg_ptr->getEstimatedExecutionTime(t).getTimeinNanoseconds());
//                                //SchedulingDecision copy_operator(*copy_alg_ptr,copy_alg_ptr->getEstimatedExecutionTime(t),t);
//                                result_plan.push_back(copy_operator); 
//                            }
                        }
                        
                        void insertCopyOperatorInnerNode(hype::query_optimization::QEP_Node* inner_node){
                                assert(inner_node!=NULL);
                                assert(!inner_node->isRoot());
                                //assert(!inner_node->isLeave());
                                if(!util::isCopyOperation(inner_node->alg_ptr->getName())){
                                    ProcessingDeviceMemoryID mem_id_current_node=inner_node->alg_ptr->getDeviceSpecification().getMemoryID();
                                    ProcessingDeviceMemoryID mem_id_parent_node=inner_node->parent->alg_ptr->getDeviceSpecification().getMemoryID();
                                    if(mem_id_current_node!=mem_id_parent_node){

                                        std::string copy_type = util::getCopyOperationType(inner_node->alg_ptr->getDeviceSpecification().getProcessingDeviceID(),inner_node->parent->alg_ptr->getDeviceSpecification().getProcessingDeviceID());
                                        assert(!copy_type.empty());
                                        DeviceSpecification dev_spec = hype::core::Scheduler::instance().getDeviceSpecificationforCopyType(copy_type);
                                        core::AlgorithmPtr copy_alg_ptr = hype::core::Scheduler::instance().getAlgorithm(util::toInternalAlgName(copy_type,dev_spec ));
                                        assert(copy_alg_ptr!=NULL);
                                        Tuple t;

                                        t.push_back(inner_node->op_spec.getFeatureVector().front());
                                         
                                        //copy data from CPU to co-processor:
                                        OperatorSpecification copy_op_spec(copy_type,
                                                                            t,
                                                                            mem_id_current_node, //memory id of input data
                                                                            mem_id_parent_node); //memory id of result data

                                        hype::query_optimization::QEP_Node* copy_node = new hype::query_optimization::QEP_Node(copy_op_spec, hype::core::DeviceConstraint());
                                        copy_node->assignAlgorithm(copy_alg_ptr.get());
                                        if(!hype::quiet && hype::verbose && hype::debug)
                                            std::cout << "create Copy Operation: " << copy_node->toString() << std::endl;
                                        
                                        hype::query_optimization::QEP_Node* old_parent=inner_node->parent;
                                        
                                        copy_node->setChilds(&inner_node,1);
                                        copy_node->parent=old_parent;
                                        
                                        for(unsigned int i=0;i<old_parent->number_childs;++i){
                                            if(old_parent->childs[i]==inner_node){
                                                old_parent->childs[i]=copy_node;
                                            }
                                        }
                                        
                                        //InternalPhysicalOperator copy_operator(copy_alg_ptr, t, copy_alg_ptr->getEstimatedExecutionTime(t).getTimeinNanoseconds());
                                        //SchedulingDecision copy_operator( (*(copy_alg_ptr.get())), copy_alg_ptr->getEstimatedExecutionTime(t),t);
                                        //cout << "insert " << copy_operator.getNameofChoosenAlgorithm() << " between " << result_plan->back().getNameofChoosenAlgorithm() << " and " << current_operator.getNameofChoosenAlgorithm()   << endl;
                                        //result_plan.push_back(copy_operator);
                                    }
                                }
                        }
                                       
                        
	};//end namespace query_optimization
}; //end namespace hype
                        
