

#include <iostream>
#include <util/utility_functions.hpp>
#include <cassert>

namespace hype{
	namespace util{
            using namespace core;
            bool satisfiesDeviceConstraint(const DeviceSpecification& dev_spec, const DeviceConstraint& dev_constr){
                 if(dev_constr==hype::ANY_DEVICE 
                || (dev_constr==hype::CPU_ONLY && dev_spec.getDeviceType()==hype::CPU) 
                || (dev_constr==hype::GPU_ONLY && dev_spec.getDeviceType()==hype::GPU)
                || (dev_constr==hype::FPGA_ONLY && dev_spec.getDeviceType()==hype::FPGA)    
                || (dev_constr==hype::XEON_PHI_ONLY && dev_spec.getDeviceType()==hype::XEON_PHI)
                || (dev_constr==hype::NP_ONLY && dev_spec.getDeviceType()==hype::NP)){
                    return true;
                }else{
                    return false;
                }   
            }
            
            bool isCoprocessor(ProcessingDeviceID id){
                return !isCPU(id);
            }
            bool isCPU(ProcessingDeviceID id){
                ProcessingDeviceType pd_type = core::Scheduler::instance().getProcessingDevices().getVirtualProcessingDevice(id)->getDeviceSpecification().getDeviceType();
                if(pd_type==CPU){
                    return true;
                }else{
                    return false;
                }
            }
            std::string getCopyOperationType(ProcessingDeviceID id1, ProcessingDeviceID id2){
                assert(id1!=id2);  
                if(isCPU(id1) && isCPU(id2)){
                    return "";
                }else if (isCPU(id1) && isCoprocessor(id2)){
                    return "COPY_CPU_CP";
                }else if(isCoprocessor(id1) && isCPU(id2)){
                    return "COPY_CP_CPU";
                }else if(isCoprocessor(id1) && isCoprocessor(id2)){
                    return "COPY_CP_CP";
                }else{
                    HYPE_FATAL_ERROR("INVALID COMBINATION of PROCESSING DEVICE TYPES! MAYBE BUG IN hype::util::isCPU() OR hype::util::isCoprocessor() ?",std::cout);
                }
            }
            
            bool isCopyOperation(const core::SchedulingDecision& sched_dec){
                std::string name = sched_dec.getNameofChoosenAlgorithm();
                if(name=="COPY_CPU_CP"
                || name=="COPY_CP_CPU"
                || name=="COPY_CP_CP"){
                    return true;
                }else{
                    return false;
                }
            }
            
            bool isCopyOperation(const std::string& name){
                if(name=="COPY_CPU_CP"
                || name=="COPY_CP_CPU"
                || name=="COPY_CP_CP"){
                    return true;
                }else{
                    return false;
                }
            }
        
	}; //end namespace util
}; //end namespace hype


